/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.categ.ManageDNARNAChars;

import java.util.*;
import java.awt.*;
import java.io.*;
import mesquite.lib.*;
import mesquite.lib.characters.*;
import mesquite.lib.duties.*;
import mesquite.categ.lib.*;

/* ======================================================================== 
Manages DNA and RNA data matrices */
public class ManageDNARNAChars extends CategMatrixManager {
	
	/*.................................................................................................................*/
	public boolean startJob(String arguments, Object condition, CommandRecord commandRec, boolean hiredByName) {
		return true;
  	 }
	public Class getDataClass(){
		return DNAData.class;
	}
	/*.................................................................................................................*/
	public boolean isPrerelease(){
		return false;
	}
	/*.................................................................................................................*/
	public  String getDataClassName(){
		return "DNA Data";
	}
	/*.................................................................................................................*/
	public  CharacterData getNewData(Taxa taxa, int numChars){
		return new DNAData(this, taxa.getNumTaxa(), numChars, taxa);
	}
	/*.................................................................................................................*/
	public boolean readsWritesDataType(Class dataClass){
		return (dataClass == DNAData.class);
	}
	/*.................................................................................................................*/
	public boolean readsWritesDataType(String dataType){
		return dataType.equalsIgnoreCase("DNA") || dataType.equalsIgnoreCase("RNA")  || dataType.equalsIgnoreCase("NUCLEOTIDE");
	}
	/*.................................................................................................................*/
	public CharacterData processFormat(MesquiteFile file, Taxa taxa, String dataType, String formatCommand, MesquiteInteger stringPos, int numChars) {
		if (taxa==null) {
			MesquiteMessage.warnProgrammer("Error: taxa null in processFormat  in " + getName());
			return null;
		}
		MesquiteProject proj=null;
		if (file!=null)
			proj = file.getProject();
		CategoricalData data;
		if (formatCommand.equalsIgnoreCase("RNA")) {
			data= new DNAData(this, taxa.getNumTaxa(), numChars, taxa);
			((DNAData)data).setDisplayAsRNA(true);
		}
		else
			data= new DNAData(this, taxa.getNumTaxa(), numChars, taxa);
		String tok = ParseUtil.getToken(formatCommand, stringPos);
		while (!tok.equals(";")) {
			if (tok.equalsIgnoreCase("TRANSPOSE")) {
				alert("Sorry, Transposed matrices of DNA characters can't yet be read");
				return null;
			}
			else if (tok.equalsIgnoreCase("interleave")) {
				data.interleaved = true;
			}
			else if (tok.equalsIgnoreCase("MISSING")) { 
				ParseUtil.getToken(formatCommand, stringPos); //eating up =
				String t = ParseUtil.getToken(formatCommand, stringPos);
				if (t!=null && t.length()==1)
					data.setUnassignedSymbol(t.charAt(0));
			}
			else if (tok.equalsIgnoreCase("GAP")) { 
				ParseUtil.getToken(formatCommand, stringPos); //eating up =
				String t = ParseUtil.getToken(formatCommand, stringPos);
				if (t!=null && t.length()==1)
					data.setInapplicableSymbol(t.charAt(0));
			}
			else if (tok.equalsIgnoreCase("MATCHCHAR")) { 
				ParseUtil.getToken(formatCommand, stringPos); //eating up =
				String t = ParseUtil.getToken(formatCommand, stringPos);
				if (t!=null && t.length()==1)
					data.setMatchChar(t.charAt(0));
			}
			else if (tok.equalsIgnoreCase("DATATYPE")) { 
				alert("Error in NEXUS file format: DATATYPE subcommand must appear as the first item in the FORMAT command");
				return null;
			}
			else if (tok.equalsIgnoreCase("SYMBOLS")) {  //should this be allowed?
				ParseUtil.getToken(formatCommand, stringPos); //eating up =
				String t = ParseUtil.getToken(formatCommand, stringPos); //eating up "
				if (t!=null && t.equalsIgnoreCase("\"")){
					t = ParseUtil.getToken(formatCommand, stringPos); //getting next token
					int count = 0;
					while (t!=null && !t.equals("\"") &&  !t.equals(";")){
						for (int i = 0; i<t.length(); i++){ //process entire string
							char c = t.charAt(i);
							if (!StringUtil.whitespace(c, null)) {
								if (count<=CategoricalState.maxCategoricalState)
									data.setSymbol(count, c);
								count++;
							}
						}
						t = ParseUtil.getToken(formatCommand, stringPos); //eating up "
					}
				}
			}
			else {
				alert("Unrecognized token (\"" + tok+ "\") in FORMAT statement of DNA matrix; matrix will be stored as foreign, and not processed.");
				return null;
			}
			tok = ParseUtil.getToken(formatCommand, stringPos);
		}
		return data;
	}
	/*.................................................................................................................*
	/*.................................................................................................................*/
	public boolean processCommand(CharacterData data, String commandName, String commandString) {
		if (commandName.equalsIgnoreCase("CHARSTATELABELS")){
			String cN = parser.getFirstToken(commandString); //eating up command name
			cN = parser.getNextToken();
			String charName;
			String stateName;
			while (!StringUtil.blank(cN) && !cN.equals(";") ) {
				int charNumber =MesquiteInteger.fromString(cN);
				charName = parser.getNextToken();  
				if (!charName.equals(",") && !charName.equals(";")) {
					if (!charName.equals("/")) {   // there is a character name
						data.setCharacterName(charNumber-1, charName);
						stateName = parser.getNextToken(); // eating up slash
					}
					else   // there is no character name, only a slash
						stateName = "";
					while (stateName!=null && !stateName.equals(",") && !stateName.equals(";")) {
						stateName = parser.getNextToken();
					}
				}
				if (charName.equals(";"))    // we've encountered the semicolon
					cN = ";";
				else
					cN = parser.getNextToken();
			}
			return true;
		}
		return false;
	}
	/*.................................................................................................................*/
	public String getCharStateLabels(CharacterData data) {
		String csl = "CHARSTATELABELS " + StringUtil.lineEnding();
		boolean found = false;
		String end = " "; //StringUtil.lineEnding()
		for (int i = 0; i<data.getNumChars(); i++) {
			String cslC="";
			if (i>0 && found)
				cslC += "," + end;
			else if (i ==0)
				cslC += "\t\t";
			cslC += Integer.toString(i+1) + " ";    //i+1 to convert to 1 based
			boolean foundInCharacter = false;
			if (data.characterHasName(i)) {
				foundInCharacter = true;
				cslC += StringUtil.tokenize(data.getCharacterName(i));
			}
			if (foundInCharacter) {
				csl += cslC;
				found = true;
			}
		}
		if (found)
			return csl + " ; " + StringUtil.lineEnding();
		else
			return "";
	}
	/*.................................................................................................................*/
	public void writeCharactersBlock(CharacterData data, CharactersBlock cB, MesquiteFile file, ProgressIndicator progIndicator){
		String endLine = ";" + StringUtil.lineEnding();
		StringBuffer blocks = new StringBuffer(5000);
		if (data.getFile()!=null && data.getFile().useDataBlock)
			blocks.append("BEGIN DATA");
		else
			blocks.append("BEGIN CHARACTERS");
		blocks.append(endLine);
		if (data.getName()!=null &&  (getProject().getNumberCharMatrices()>1 || !NexusBlock.suppressTITLE)){
			blocks.append("\tTITLE  ");
			blocks.append( StringUtil.tokenize(data.getName()));
			blocks.append(endLine);
		}
		//if (data.getTaxa().getName()!=null  && getProject().getNumberTaxas(cB.getFile())>1){ //before 13 Dec 01 had been this
		if (data.getTaxa().getName()!=null  && getProject().getNumberTaxas()>1){ // should have an isUntitled method??
			blocks.append("\tLINK TAXA = ");
			blocks.append(StringUtil.tokenize(data.getTaxa().getName()));
			blocks.append(endLine);
		}
		if (data.getAnnotation()!=null) {
			blocks.append("[!" + data.getAnnotation() + "]");
			blocks.append(StringUtil.lineEnding());
		}
		blocks.append("\tDIMENSIONS NCHAR=" + data.getNumChars());
		blocks.append(endLine);
		blocks.append("\tFORMAT");
		if (!((DNAData)data).getDisplayAsRNA())
			blocks.append(" DATATYPE = DNA");
		else
			blocks.append(" DATATYPE = RNA");
		if (data.interleaved)
			blocks.append(" INTERLEAVE");
		blocks.append(" GAP = " + data.getInapplicableSymbol() + " MISSING = " + data.getUnassignedSymbol());
		blocks.append(endLine);
		if (data.isLinked()){
			blocks.append("\tOPTIONS ");
			Vector ds = data.getDataLinkages();
			for (int i = 0; i<ds.size(); i++) {
				blocks.append(" LINKCHARACTERS = ");
				blocks.append(StringUtil.tokenize(((CharacterData)ds.elementAt(i)).getName()));
			}
			blocks.append(endLine);
		}
		blocks.append(getCharStateLabels(data));

		writeNexusMatrix(data, cB, blocks, file, progIndicator);
		
		blocks.append(cB.getUnrecognizedCommands() + StringUtil.lineEnding());
		blocks.append("END;" + StringUtil.lineEnding());
	//	MesquiteModule.mesquiteTrunk.mesquiteMessage("DNA matrix composed", 1, 0);

		file.writeLine( blocks.toString());
	}


	/*.................................................................................................................*/
    	 public String getName() {
		return "Manage DNA/RNA matrices";
   	 }
   	 
	/*.................................................................................................................*/
 	/** returns an explanation of what the module does.*/
 	public String getExplanation() {
 		return "Manages DNA/RNA data matrices (including read/write in NEXUS file)." ;
   	 }
}
	
	
