/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.categ.lib;

import java.awt.*;
import java.util.*;
import java.util.zip.*;
import mesquite.lib.*;
import mesquite.lib.characters.*;
import mesquite.lib.duties.*;

/* ======================================================================== */
/** A subclass of CharacterData for data stored as Categorical sets (e.g, "{0, 2}").  Associated with the CharacterState subclass CategoricalState.*/
public class CategoricalData extends CharacterData {
	private String[][] stateNames;
	private String[][] stateNotes;
	protected char[] symbols;
	protected final static char[] defaultSymbols;

	private long[][] matrix;
	private short[][] matrixShort;
	private static Vector defaultModels;
	protected static char ANDseparator = '&';  //why static?
	protected static char ORseparator = '/';
	protected String puncString;
	private int maxSymbolDefined = -1;
	private static boolean permitShortMatrices = true;
	static {
		defaultModels= new Vector(2);
		defaultSymbols = new char[] {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J', 'K', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'j', 'k', 'm', 'n', 'p' , 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z'};
	}
	
	public CategoricalData(CharMatrixManager manager, int numTaxa, int numChars, Taxa taxa){
		super(manager, numTaxa, numChars, taxa);
		symbols = new char[] {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J', 'K', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'j', 'k', 'm', 'n', 'p' , 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z'};
		try {
	   		puncString = "" + ANDseparator + ORseparator + getUnassignedSymbol() + getInapplicableSymbol();
			if (permitShortMatrices && compressibleToShort(getDefaultState())) {
				matrixShort = new short[ numChars][numTaxa];
				for (int it=0; it<numTaxa; it++)
					for (int ic=0; ic<numChars; ic++)
						matrixShort[ic][it] = compressToShort(getDefaultState());
			}
			else {
				matrix = new long[ numChars][numTaxa];
				for (int it=0; it<numTaxa; it++)
					for (int ic=0; ic<numChars; ic++)
						matrix[ic][it] = getDefaultState();
			}
		}
		catch (OutOfMemoryError e){
			MesquiteMessage.warnProgrammer("Sorry, the Character Matrix could not be created because of lack of memory");
		}
	}
	
	/** The mask that preserves bits 0.. maxCategoricalState*/
	public static final long shortStatesBitsMaskL = (~0L)>>>53; //5 for high bits + 48 for reduction from 64 to 16
	private boolean compressibleToShort(long s){
		return (s & shortStatesBitsMaskL) == (s & CategoricalState.statesBitsMask);
	}
	private short compressToShort(long s){
		return (short)((s & shortStatesBitsMaskL) | ((s & CategoricalState.highBitsMask)>>>48));
	}
	private long expandFromShort(short s){
		long LS = ((long)s);
		return (LS & shortStatesBitsMaskL) | ((LS << 48) & CategoricalState.highBitsMask);
	}
	public boolean usingShortMatrix(){
		return matrix == null;
	}
	private void upgradeToFullMatrix(){
		matrix = new long[ numChars][numTaxa];
		for (int ic=0; ic<numChars && ic<matrixShort.length; ic++)
			for (int it=0; it<numTaxa && it<matrixShort[ic].length; it++)
				matrix[ic][it] = expandFromShort(matrixShort[ic][it]);
		matrixShort = null;
	}
	/*..........................................  CategoricalData  ..................................................*/
	public long getDefaultState(){
  		return CategoricalState.unassigned;
	}
	/*..........................................  CategoricalData  ..................................................*/
	public void dispose(){
  		super.dispose();
  		matrix = null;
  		matrixShort = null;
  		stateNames = null;
  		stateNotes = null;
  		symbols = null;
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** Indicates the type of character stored */ 
	public Class getStateClass(){
		return CategoricalState.class;
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** returns the name of the type of data stored */
	public String getDataTypeName(){
		return "Standard Categorical Data";
	}
	/*..........................................  CategoricalData  ..................................................*/
	/**clone this CharacterData and return new copy */ //TODO: here should use super.setToClone(data) to handle specssets etc.???
	public CharacterData cloneData(){
		CategoricalData data = new CategoricalData(matrixManager, numTaxa, numChars, getTaxa());
		if (symbols!=null)
			for (int i=0; i<=CategoricalState.maxCategoricalState && i<symbols.length; i++) {
				data.setSymbolDirect(i, symbols[i]);
			}
		for (int ic=0; ic<numChars; ic++){
			if (hasStateNames())
				for (int i = 0; i < CategoricalState.maxCategoricalState; i++)
					data.setStateName(ic,i,getStateName(ic,i));
			for (int it=0; it<numTaxa; it++) {
				data.setState(ic, it, getStateRaw(ic, it)); 
			}
		}
		data.resetChangedSinceSave();
		return data;
	}
	public int getNumTaxa(){
		if (disposed)
			return 0;
		int n = super.getNumTaxa();
		if (matrix == null && matrixShort == null)
			MesquiteMessage.warnProgrammer("Categorical data with null internal matrix. getNumTaxa() = " + n + " name " + getName() + " file " + getFile());
		else if (matrix != null && matrix.length>0 && matrix[0] !=null && matrix[0].length != n)
			MesquiteMessage.warnProgrammer("Categorical matrix with incorrect record of number of taxa. getNumTaxa() = " + n + " matrix[0].length " + matrix[0].length + " name " + getName() + " file " + getFile());
		else if (matrixShort != null && matrixShort.length>0 && matrixShort[0] !=null && matrixShort[0].length != n)
			MesquiteMessage.warnProgrammer("Categorical matrix with incorrect record of number of taxa. getNumTaxa() = " + n + " matrixShort[0].length " + matrixShort[0].length + " name " + getName() + " file " + getFile());
		return n;
	}
	public int getNumChars(){
		if (disposed)
			return 0;
		int n = super.getNumChars();
		if (matrix == null && matrixShort == null)
			MesquiteMessage.warnProgrammer("Categorical data with null internal matrix. getNumChar() = " + n  + " nAdd = " + nAdd + " nDel = " + nDel + " nMove = " + nMove + " name " + getName() + " file " + getFile());
		else if (matrix != null && matrix.length !=n)
			MesquiteMessage.warnProgrammer("Categorical matrix with incorrect record of number of characters. getNumChar() = " + n + " matrix.length " + matrix.length + " nAdd = " + nAdd + " nDel = " + nDel + " nMove = " + nMove + " name " + getName() + " file " + getFile());
		else if (matrixShort != null && matrixShort.length !=n)
			MesquiteMessage.warnProgrammer("Categorical matrix with incorrect record of number of characters. getNumChar() = " + n + " matrixShort.length " + matrixShort.length + " nAdd = " + nAdd + " nDel = " + nDel + " nMove = " + nMove + " name " + getName() + " file " + getFile());
		return n;
	}
	/*..........................................  CategoricalData  ..................................................*/
	/**swaps characters first and second*/
	public  boolean swapParts(int first, int second){
		if (first<0 || first >= numChars)
			return false;
		if (second<0 || second >= numChars)
			return false;
		if (!usingShortMatrix())
			Long2DArray.swapColumns(matrix, first, second);
		else
			Short2DArray.swapColumns(matrixShort, first, second);
		StringArray.swapColumns(stateNames, first, second);
		StringArray.swapColumns(stateNotes, first, second);

		return super.swapParts(first,second);
	
	}
	/*..........................................  CategoricalData  ..................................................*/
	/**move num characters starting at first to just after character justAfter */
	public  boolean moveParts(int first, int num, int justAfter){
		if (justAfter<0)
			justAfter = -1;
		else if (justAfter>=numChars)
			justAfter = numChars-1;
		if (first<0 || first >= numChars)
			return false;
		if (num<=0)
			return false;
		if (first + num>numChars)
			return false;
		if (!usingShortMatrix())
			Long2DArray.moveColumns(matrix, first, num, justAfter);
		else
			Short2DArray.moveColumns(matrixShort, first, num, justAfter);
		StringArray.moveColumns(stateNames, first, num, justAfter);
		StringArray.moveColumns(stateNotes, first, num, justAfter);

		return super.moveParts(first,num,justAfter);
	
	}

	/*..........................................  CategoricalData  ..................................................*/
	/**Adds num characters after position "starting"; returns true iff successful.*/
	public  boolean addParts(int starting, int num){
		if (starting<0)
			starting = -1;
		else if (starting>=numChars)
			starting = numChars-1;
		int newNumChars = numChars + num;

		try {
			if (!usingShortMatrix())
				matrix = Long2DArray.addColumns(matrix, numTaxa, getDefaultState(), starting, num);
			else
				matrixShort = Short2DArray.addColumns(matrixShort, numTaxa, compressToShort(getDefaultState()), starting, num);


			//numChars = newNumChars; don't do this since addCharactersAdjust needs to remember old number
			if (stateNames!=null){
				String[][] newStateNames = new String[newNumChars][];
				for (int i=0; i<=starting; i++) {
					newStateNames[i] = stateNames[i];
				}
				for (int i=0; i<num; i++) {
					newStateNames[starting + i + 1] = null;
				}
				for (int i=0; i<numChars-starting-1; i++) {
					newStateNames[i +starting+num+1] = stateNames[starting + i+1];
				}
				stateNames = newStateNames;
			}
			if (stateNotes !=null){
				String[][] newStateNotes = new String[newNumChars][];
				for (int i=0; i<=starting; i++) {
					newStateNotes[i] = stateNotes[i];
				}
				for (int i=0; i<num; i++) {
					newStateNotes[starting + i + 1] = null;
				}
				for (int i=0; i<numChars-starting-1; i++) {
					newStateNotes[i +starting+num+1] = stateNotes[starting + i+1];
				}
				stateNotes = newStateNotes;
			}
			return super.addParts(starting, num);
		}
		catch (OutOfMemoryError e){
			MesquiteMessage.warnProgrammer("Sorry, characters could not be added to the Character Matrix because of lack of memory");
		}
		return false;

	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Deletes num characters from position "starting"; returns true iff successful.*/
	public boolean deleteParts(int starting, int num){
		if (num<=0)
			return false;
		if (starting<0)
			return false;
		else if (starting>numChars)
			return false;
		if (num+starting>numChars)
			num = numChars-starting;
		int newNumChars =numChars-num;
		if (!usingShortMatrix())
			matrix = Long2DArray.deleteColumns(matrix, starting, num);
		else
			matrixShort = Short2DArray.deleteColumns(matrixShort, starting, num);

		//numChars = newNumChars; don't do this since super.deleteCharacters needs to remember old number
		if (stateNames!=null){
			String[][] newStateNames = new String[newNumChars][];
			for (int i=0; i<starting; i++) {
				newStateNames[i] = stateNames[i];
			}
			for (int i=starting + num;i<numChars; i++) {
				newStateNames[i-num] = stateNames[i];
			}
			stateNames = newStateNames;
		}
		if (stateNotes !=null){
			String[][] newStateNotes = new String[newNumChars][];
			for (int i=0; i<starting; i++) {
				newStateNotes[i] = stateNotes[i];
			}
			for (int i=starting + num;i<numChars; i++) {
				newStateNotes[i-num] = stateNotes[i];
			}
			stateNotes = newStateNotes;
		}
		return super.deleteParts(starting, num);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Adds num taxa after position "starting"; returns true iff successful.*/
	public boolean addTaxa(int starting, int num){

		setDirty(true);
		if (starting<0)
			starting = -1;
		else if (starting>=numTaxa)
			starting = numTaxa-1;
		int newNumTaxa = numTaxa + num;
		
		try {
			if (usingShortMatrix()){
				short shortUnassigned = compressToShort(CategoricalState.unassigned);
				for (int ic=0; ic<numChars; ic++) {
					matrixShort[ic] = ShortArray.addParts(matrixShort[ic], starting, num, shortUnassigned);
				}
			}
			else {
				for (int ic=0; ic<numChars; ic++) {
					matrix[ic] = LongArray.addParts(matrix[ic], starting, num, CategoricalState.unassigned);
				}
			}
		}
		catch (OutOfMemoryError e){
			MesquiteMessage.warnProgrammer("Sorry, taxa could not be added to the Character Matrix because of lack of memory");
		}
		numTaxa = newNumTaxa; 
		return super.addTaxa(starting, num);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Deletes num taxa from position "starting"; returns true iff successful.*/
	public boolean deleteTaxa(int starting, int num){
		if (num<=0)
			return false;
		if (starting<0)
			return false;
		else if (starting>=numTaxa)
			return false;
		setDirty(true);
		if (num+starting>numTaxa)
			num = numTaxa-starting;
		int newNumTaxa = numTaxa - num;
		
			
		if (usingShortMatrix()){
			for (int ic=0; ic<numChars; ic++)
				matrixShort[ic] = ShortArray.deleteParts(matrixShort[ic], starting, num);
		}
		else for (int ic=0; ic<numChars; ic++)
			matrix[ic] = LongArray.deleteParts(matrix[ic], starting, num);


		numTaxa = newNumTaxa; 
		return super.deleteTaxa(starting, num);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/**moves num taxa from position "starting" to just after position "justAfter"; returns true iff successful.*/
	public boolean moveTaxa(int starting, int num, int justAfter){
		if (num<=0)
			return false;
		if (starting<0)
			return false;
		else if (starting>=numTaxa)
			return false;
		setDirty(true);
		if (usingShortMatrix()) {
			for (int ic=0; ic<numChars; ic++)
				ShortArray.moveParts(matrixShort[ic], starting, num, justAfter);
		}
		else for (int ic=0; ic<numChars; ic++)
			LongArray.moveParts(matrix[ic], starting, num, justAfter);
		return super.moveTaxa(starting, num, justAfter);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** Get symbol for state st. */
	public char getSymbol(int st) {
		if (st <0 || st>=symbols.length)
			return getUnassignedSymbol();
		return symbols[st];
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** Set symbol for state st.  Does not check for duplicate symbols */
	public void setSymbolDirect(int st, char symbol) {
		if (st <0 || st>=symbols.length)
			return;
		setDirty(true);
		int current = findSymbol(symbol);
		if (maxSymbolDefined<st)
			maxSymbolDefined = st;
		symbols[st] = symbol;
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** Set symbol for state st.  Returns whether setting implied duplicate symbol */
	public boolean setSymbol(int st, char symbol) {
		if (st <0 || st>=symbols.length)
			return false;
		setDirty(true);
		int current = findSymbol(symbol);
		if (maxSymbolDefined<st)
			maxSymbolDefined = st;
		if (current >=0 && current != st){ //duplicate symbol!
			symbols[st] = symbol;
			char unusedSymbol = findUnusedSymbol();
			if (unusedSymbol != (char)0)
				symbols[current] = unusedSymbol;
			return true;
		}
		else {
			symbols[st] = symbol;
			return false;
		}
	}
	
	private int findSymbol(char symbol){
		for (int ist = 0; ist<symbols.length; ist++) {
			if (symbols[ist]==symbol) {
				return ist;
			}
		}
		return -1;
	}
	private char findUnusedSymbol(){
		for (int ist = 0; ist<defaultSymbols.length; ist++) {
			if (findSymbol(defaultSymbols[ist])<0) {
				return defaultSymbols[ist];
			}
		}
		return (char)0;
	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** sets the symbol used for inapplicable character (e.g., gap)*/
   	public void setInapplicableSymbol(char inapp) {
   		super.setInapplicableSymbol(inapp);
   		puncString = "" + ANDseparator + ORseparator + getUnassignedSymbol() + getInapplicableSymbol();
   	}
    	/** sets the symbol used for missing data (unassigned) character*/
   	public void setUnassignedSymbol(char mc) { 
   		super.setUnassignedSymbol(mc);
   		puncString = "" + ANDseparator + ORseparator + getUnassignedSymbol() + getInapplicableSymbol();
  	}
	/*..........................................CategoricalData................*/
   	/** trades the states of character ic and ic2 in taxon it.  Used for reversing sequences (for example).*/
	public void tradeStatesBetweenCharacters(int ic, int ic2, int it) { 
		if (notInStorage(ic, it) || notInStorage(ic2, it))
			return;
		if (usingShortMatrix()){
			short temp = matrixShort[ic][it];
			matrixShort[ic][it] =  matrixShort[ic2][it];
			matrixShort[ic2][it] = temp;
		}
		else {
			long temp = matrix[ic][it];
			matrix[ic][it] =  matrix[ic2][it];
			matrix[ic2][it] = temp;
		}
		setDirty(true, ic, it);
		setDirty(true, ic2, it);
	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** trades the states of character ic between taxa it and it2.  Used for reshuffling.*/
	public void tradeStatesBetweenTaxa(int ic, int it, int it2) {
		if (notInStorage(ic, it) || notInStorage(ic, it2))
			return;
		if (usingShortMatrix()){
			short temp = matrixShort[ic][it];
			matrixShort[ic][it] =  matrixShort[ic][it2];
			matrixShort[ic][it2] = temp;
		}
		else {
			long temp = matrix[ic][it];
			matrix[ic][it] =  matrix[ic][it2];
			matrix[ic][it2] = temp;
		}
		setDirty(true, ic, it);
		setDirty(true, ic, it2);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Dump matrix to log */
	public void logMatrix(){
		MesquiteModule.mesquiteTrunk.logln("Matrix of " + numTaxa + " taxa and " + numChars + " characters.");
		String matrixString="";
		for (int it=0; it<numTaxa; it++) {
			for (int ic=0; ic<numChars; ic++) {
				matrixString+=CategoricalState.toString(getStateRaw(ic, it));
			}
			matrixString+= '\r';
		}
		MesquiteModule.mesquiteTrunk.logln(matrixString);
	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** Fills matrix with data from passed StsOfCharacters object.*/
   	public void setMatrix(MCharactersDistribution states){ //TODO: make this so it will change size of matrix if needed???
   		if (!(states instanceof MCategoricalDistribution))
   			return;
		setDirty(true);
   		MCategoricalDistribution dStates = (MCategoricalDistribution)states;
   		for (int ic = 0; ic< numChars; ic++) {
   			for (int it = 0; it< numTaxa; it++)  {
   				if (ic>dStates.getNumChars())
   					setState(ic, it, getDefaultState());
   				else
   					setState(ic, it, dStates.getState(ic, it));
   				stampHistoryChange(ic, it);
   			}
   		}
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/**get matrix and return as StsOfCharacters */
   	public MCharactersDistribution getMCharactersDistribution(){
   		MCategoricalEmbedded states = new MCategoricalEmbedded(this);
		return states;
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/**return a CharacterDistribution that points to character ic */
   	public CharacterDistribution getCharacterDistribution(int ic){
   		CategoricalEmbedded states = new CategoricalEmbedded(this, ic);
		 return states;
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/**extract character ic and return as an array of long's */
   	public long[] getCharacter(int ic){
   		if (notInStorage(ic, 0)) //illegal check
			return null;
   		if (ic<numChars && ic>=0) {
	  		long[] c = new long[getNumTaxa()];
	  		for (int i=0; i<c.length; i++)
	  			c[i] = getStateRaw(ic, i);
	  		return c;
		}
		 return null;
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Return CharacterDistribution object with same number of taxa as this object (but not with any particular data; just so that of same data type) */
   	public CharacterDistribution makeCharacterDistribution(){
   		CategoricalAdjustable c= new CategoricalAdjustable(getTaxa(), numTaxa);
   		c.setParentData(this);
   		return c;
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** creates an empty CharacterState object of the same data type as CharacterData subclass used.*/
   	public CharacterState makeCharacterState(){
   		return new CategoricalState();
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Return CharacterData object with same number of taxa & characters as this object (but not with any particular data; just so that of same data type) */
   	public CharacterData makeCharacterData() {
   		return new CategoricalData(getMatrixManager(), getNumTaxa(), getNumChars(), getTaxa());
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Return CharacterData object with passed number of taxa & characters (but not with any particular data; just so that of same data type) */
   	public CharacterData makeCharacterData(int ntaxa, int nchars) {
   		return new CategoricalData(getMatrixManager(), ntaxa, nchars, getTaxa());
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	public CharacterData makeCharacterData(CharMatrixManager manager, Taxa taxa){ 
   		return new CategoricalData(getMatrixManager(), taxa.getNumTaxa(), 0, taxa);
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** returns whether a given char is acceptable as the value for a character state */
   	public boolean isAcceptableCharForState(char c){
   		long value = fromChar(c);
   		return (value == CategoricalState.unassigned || value == CategoricalState.inapplicable || CategoricalState.isCombinable(value));
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** returns Color for state of character, scaled to maxState (e.g. for tracing to yield white/black for binary) */
	public Color getColorOfState(int ic, int istate, int maxState){
		return MesquiteColorTable.getDefaultColor(maxState,istate, MesquiteColorTable.COLORS);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Returns the color for the given state of character ic.  Currently only returns default */
	public Color getColorOfState(int ic, int istate){
		return MesquiteColorTable.getDefaultColor(CategoricalState.maxCategoricalState,istate, MesquiteColorTable.COLORS);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** Gets the color representing state(s) of character ic in taxon it */ 
	public Color getColorOfStates(int ic, int it){
   		if (notInStorage(ic, it)) //illegal check
			return ColorDistribution.unassigned;
		long s = getStateRaw(ic, it);
		if (CategoricalState.isCombinable(s)) {
			int colorCount = CategoricalState.cardinality(s);
			if (colorCount>1){
				return Color.lightGray;
			}
			else {
				return getColorOfState(ic, CategoricalState.maximum(s), getMaxState(ic));
			}
		}
		else 
			return ColorDistribution.unassigned;
	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** returns whether the character ic is inapplicable to taxon it*/
   	public  boolean isInapplicable(int ic, int it){
   		if (notInStorage(ic, it)) //illegal check
   			return true;
   		return getStateRaw(ic, it)== CategoricalState.inapplicable;
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** returns whether the state of character ic is missing in taxon it*/
   	public  boolean isUnassigned(int ic, int it){
   		if (notInStorage(ic, it)) //illegal check
   			return true;
   		return getStateRaw(ic, it)== CategoricalState.unassigned;
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** returns the maximum symbol defined*/
   	public int getMaxSymbolDefined(){
   		if (maxSymbolDefined>=0)
   			return maxSymbolDefined;
   		else
   			return getMaxState();
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** returns the maximum state in character ic*/
   	public int getMaxState(int ic){
   		if (notInStorage(ic, 0)) //illegal check
   			return -1;
   		
   		long allstates = 0;
		for (int it=0; it<numTaxa; it++)
			allstates |= getState(ic, it);
   		return CategoricalState.maximum(allstates);
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** returns the set of states in character ic*/
   	public long getAllStates(int ic){
   		if (notInStorage(ic, 0)) //illegal check
   			return 0L;
   		
   		long allstates = 0;
		for (int it=0; it<numTaxa; it++)
			allstates |=getState(ic, it);
   		return allstates & CategoricalState.statesBitsMask;
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** returns the maximum state in the matrix*/
   	public int getMaxState(){
   		if ((matrix == null && matrixShort == null) ) //illegal check
   			return -1;
   		long allstates = 0;
   		for (int ic=0; ic<numChars; ic++)
   			for (int it=0; it<numTaxa; it++)
   				allstates |=getState(ic, it);
   		return CategoricalState.maximum(allstates);
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** returns the state of character ic in taxon it*/
   	public  CharacterState getCharacterState(CharacterState cs, int ic, int it){
   		if (notInStorage(ic, it)) //illegal check
   			return null;
   		if (cs!=null && cs.getClass()==CategoricalState.class) {
   			((CategoricalState)cs).setValue(getState(ic, it));
   			return cs;
		}
   		return new CategoricalState(getState(ic, it));
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	protected boolean notInStorage(int ic, int it) {
   		if (ic>=numChars || it>=numTaxa || ic < 0 || it < 0)
   			return true;
   		if (matrix != null && (ic >= matrix.length || matrix.length<1 || matrix[0] == null || it >= matrix[0].length))
   			return true;
   		if (matrixShort != null && (ic >= matrixShort.length || matrixShort.length<1 || matrixShort[0] == null || it >= matrixShort[0].length))
   			return true;
   		if (matrixShort == null && matrix == null)
   			return true;
   		return false;
   	}
   	MesquiteInteger resultCode = new MesquiteInteger();
	/*..........................................  CategoricalData  ..................................................*/
   	/** sets the state of character ic in taxon it from CharacterState cs*/
   	public  void setState(int ic, int it, CharacterState cs){
   		if (cs !=null && cs instanceof CategoricalState) {
   			setState(ic, it, ((CategoricalState)cs).getValue());
  			setDirty(true, ic, it);
 		}
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** Set stateset in node it and character ic to state indicated by state symbol. */
   	public void setState(int ic, int it, char c){
		setState(ic, it, fromChar(c));
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** Set stateset in node it and character ic to states. */
   	public void setState(int ic, int it, long states){
   		if (notInStorage(ic, it)) //illegal check
   			return;
		if (usingShortMatrix()) {
			if (!compressibleToShort(states)) {
				try {
					upgradeToFullMatrix();
					matrix[ic][it]= states;
				}
				catch (OutOfMemoryError e){
					matrix = null;
					MesquiteMessage.warnProgrammer("Sorry, the character state of character " + (ic+1) + " in taxon " + (it+1) + " could not be set to " + CategoricalState.toString(states) + " because of insufficient memory ");
				}
			}
			else
				matrixShort[ic][it] = compressToShort(states);
		}
		else
			matrix[ic][it]= states;
  		setDirty(true, ic, it);
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Set the state at character ic and taxon it from the string of the parser, beginning at current position of the parser.
	If pos == null, assumes from data editor and looks for state names, treats differently*/
   	public int setState(int ic, int it, Parser parser, boolean fromEditor, MesquiteString result){ 
   		if (notInStorage(ic, it)) {//illegal check
   			if (result!=null)
   				result.setValue("Character or taxon number out of bounds (Set state, CategoricalData)." + ic + "  " + ic);
   			MesquiteMessage.printStackTrace("");
   		 	return OUTOFBOUNDS;
   		 }
   		 if (StringUtil.blank(parser.getString())) {
   			if (result!=null)
   				result.setValue("State string is blank.");
   		 	return ERROR;
   		 }
   		 resultCode.setValue(MesquiteInteger.unassigned);
		long v = fromString(ic, parser, fromEditor, resultCode, result);
		if (v==CategoricalState.impossible) {
			if (resultCode.getValue() == EOL)
				return EOL;
			if (result !=null) {
				if (fromEditor)
					result.setValue("\"" + parser.getString() + "\" cannot be translated into a state for this character");
				else
					result.setValue("Illegal state entry for categorical character");
			}
			return ERROR;
  		}
		else if (v!=getStateRaw(ic, it)) {
			setState(ic, it, v);
  			setDirty(true, ic, it);
  		}
		if (resultCode.isUnassigned())
			return OK;
		else
			return resultCode.getValue();
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** return stateset in node it and character ic. */
   	public long getState(int ic, int it){
   		if (notInStorage(ic, it)) //illegal check
   			return CategoricalState.impossible;
	  	if (usingShortMatrix())
	  		return CategoricalState.dataBitsMask & expandFromShort(matrixShort[ic][it]);
	  	else
	  		return CategoricalState.dataBitsMask & matrix[ic][it];
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** return stateset in node it and character ic, raw, as in matrix. */
   	public long getStateRaw(int ic, int it){
   		if (notInStorage(ic, it)) //illegal check
   			return CategoricalState.impossible;
	  	if (usingShortMatrix())
	  		return expandFromShort(matrixShort[ic][it]);
	  	else
	  		return matrix[ic][it];
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** returns true if state names are available for matrix */
   	public boolean hasStateNames() {
   		return (stateNames!=null);
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** returns true if state names are available for character ic*/
   	public boolean hasStateNames(int ic) {
		return (stateNames!= null && ic>=0 && ic<numChars && stateNames[ic]!=null );
   	}
	/*..........................................  CategoricalData  ..................................................*/
	public long calculateChecksum(CRC32 crc32){
		byte[] bytes = new byte[8];
		crc32.reset();
		for (int ic=0; ic<numChars; ic++)
			for (int it=0; it<numTaxa; it++)  {
				long state = getStateRaw(ic, it);
				if (CategoricalState.isUnassigned(state))
					state = CategoricalState.unassigned;
				else if (CategoricalState.isInapplicable(state))
					state = CategoricalState.inapplicable;
				else if (CategoricalState.isImpossible(state))
					state = 0L;
				else
					state &= CategoricalState.statesBitsMask | (1L<<CategoricalState.uncertainBit);
				
				bytes = MesquiteNumber.longToBytes(state, bytes);
				for (int i = 0; i<8; i++)
					crc32.update(bytes[i]);
			}
		
		return crc32.getValue();
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** returns true if state name for state s is available for character ic*/
   	public boolean hasStateName(int ic, int s) {
		return (stateNames!= null && ic>=0 && ic<numChars && stateNames[ic]!=null && s< stateNames[ic].length && stateNames[ic][s]!=null);
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** returns true if state name for state s is available for character ic*/
   	public boolean hasStateAnnotation(int ic, int s) {
		return (stateNotes!= null && ic>=0 && ic<numChars && stateNotes[ic]!=null && s< stateNotes[ic].length && stateNotes[ic][s]!=null);
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** sets state name of state "state" of character ic to name */
	public void setStateName(int ic, int state, String name) { //todo: check if in bounds
		setDirty(true);
		if (StringUtil.blank(name))
			name = null;
		if (stateNames==null) {
			stateNames = new String[numChars][];
		}
		if (stateNames[ic]==null) {
			if (state ==0)
				stateNames[ic] = new String[2];
			else
				stateNames[ic] = new String[state+1];
		}
		else if (state>= stateNames[ic].length) {
			String[] newStateNames = new String[state+1];
			for (int i=0; i<stateNames[ic].length; i++) {
				newStateNames[i] = stateNames[ic][i];
			}
			stateNames[ic]=newStateNames;
		}
		stateNames[ic][state]=name;
		
		notifyListeners(this, new Notification(NAMES_CHANGED, new int[] {ic}));
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** gets state name of state "state" of character ic */
	public String getStateName(int ic, int state) {
		if (stateNames!= null && state>=0 && stateNames.length> ic && stateNames[ic]!=null && state< stateNames[ic].length && stateNames[ic][state]!=null) {
			return stateNames[ic][state];
		}
		else if (symbols!=null && state < symbols.length)
			return String.valueOf(symbols[state]);
		else return null;
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** sets state name of state "state" of character ic to name */
	public void setStateNote(int ic, int state, String name) { //todo: check if in bounds
		setDirty(true);
		if (StringUtil.blank(name))
			name = null;
		if (stateNotes==null) {
			stateNotes = new String[numChars][];
		}
		if (stateNotes[ic]==null) {
			if (state ==0)
				stateNotes[ic] = new String[2];
			else
				stateNotes[ic] = new String[state+1];
		}
		else if (state>= stateNotes[ic].length) {
			String[] newStateNotes = new String[state+1];
			for (int i=0; i<stateNotes[ic].length; i++) {
				newStateNotes[i] = stateNotes[ic][i];
			}
			stateNotes[ic]=newStateNotes;
		}
		stateNotes[ic][state]=name;
		
		notifyListeners(this, new Notification(ANNOTATION_CHANGED, new int[] {ic}));
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** gets state name of state "state" of character ic */
	public String getStateNote(int ic, int state) {
		if (stateNotes!= null && stateNotes.length> ic && ic >= 0 && stateNotes[ic]!=null && state>=0 && state< stateNotes[ic].length && stateNotes[ic][state]!=null) {
			return stateNotes[ic][state];
		}
		else return null;
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** returns state number for last state with name for character ic*/
   	public int maxStateWithName(int ic) {
		if (stateNames[ic]==null) 
			return 0;
		else
			for (int state=stateNames[ic].length-1; state >= 0; state--) {
				if (!((stateNames[ic][state] == null)||(stateNames[ic][state] == ""))){
					return state;
				}
			}
		return 0;
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** gets state symbol for state "state", returned as string. */
	public static String getDefaultStateSymbol(Class dataClass, int state) {
		if (dataClass==DNAState.class) 
			return DNAData.getDefaultStateSymbol(state);
		else if (dataClass==RNAState.class) 
			return RNAData.getDefaultStateSymbol(state);
		else if (dataClass==ProteinState.class) 
			return ProteinData.getDefaultStateSymbol(state);
		else
			return getDefaultStateSymbol(state);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** gets state symbol for state "state", returned as string. */
	public static String getDefaultStateSymbol( int state) {
		if (state <0 || state>=defaultSymbols.length)
			return String.valueOf(CharacterData.defaultInapplicableChar);
		if (state>=0 &&  state <= CategoricalState.maxCategoricalState && defaultSymbols!=null )
			return String.valueOf(defaultSymbols[state]);
		else 
			return Integer.toString(state);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** gets state symbol for state "state" of character ic, returned as string. */
	public String getStateSymbol(int ic, int state) {
		if (state <0 || state>=symbols.length)
			return String.valueOf(getUnassignedSymbol());
		if (state>=0 && symbols!=null ) {
			return String.valueOf(symbols[state]);
		}
		else if (state >=0 && state <= CategoricalState.maxCategoricalState)
			return String.valueOf(defaultSymbols[state]);
		else 
			return Integer.toString(state);
	}
	/*..........................................CategoricalData................*/
   	/** returns a String summarizing the states of a character (e.g., "2 states", "0.1-0.9").*/
	public String getStatesSummary(int ic){
			CategoricalState state=null;
			long allstates =0L;
			for (int it=0; it<numTaxa; it++) {
				state =(CategoricalState)getCharacterState(state, ic, it);
				if (!CategoricalState.isUnassigned(state.getValue()) && !CategoricalState.isInapplicable(state.getValue()))
					allstates |= state.getValue();
			}
			return Integer.toString(CategoricalState.cardinality(allstates));
	}
	/*..........................................CategoricalData................*/
   	public static DefaultReference findDefaultReference(NameReference paradigm){
   		if (defaultModels == null) {
   			MesquiteMessage.warnProgrammer("findDefaultReference with null default models ");
   			MesquiteMessage.printStackTrace();
   			return null;
   		}
   		for (int i=0; i<defaultModels.size(); i++){
   			DefaultReference dR = (DefaultReference)defaultModels.elementAt(i);
   			
   			if (dR.getParadigm() != null && dR.getParadigm().equals(paradigm))
   				return dR;
   		}
   		return null;
   		
   	}
	/*..........................................CategoricalData................*/
   	public static void registerDefaultModel(String paradigm, String name){
   		if (defaultModels == null)
   			return;
   		NameReference p = NameReference.getNameReference(paradigm);
   		DefaultReference dR = findDefaultReference(p);
		if (dR == null) {
			dR = new DefaultReference(p);
			defaultModels.addElement(dR);
		}
		dR.setDefault(name);
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	public CharacterModel getDefaultModel(String paradigm){
   		NameReference p = NameReference.getNameReference(paradigm);
   		DefaultReference dR = findDefaultReference(p);
   		if (dR==null)
   			return null;
   		else {
   			CharacterModel cm = getProject().getCharacterModel(dR.getDefault());
   			if (cm==null) 
   				MesquiteMessage.println("Default model not found / " + dR.getDefault());
   			return cm;
   		}
   	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** appends to buffer string describing the state(s) of character ic in taxon it. */
	public void statesIntoStringBufferCore(int ic, long s, StringBuffer sb, boolean forDisplay){
		statesIntoStringBufferCore(ic,s,sb,forDisplay, true, true);
	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** appends to buffer string describing the state(s) of character ic in taxon it. */
	public void statesIntoStringBufferCore(int ic, long s, StringBuffer sb, boolean forDisplay, boolean includeInapplicable, boolean includeUnassigned){
		if (s==CategoricalState.inapplicable) {
			if (includeInapplicable)
				sb.append(getInapplicableSymbol());
		}
		else if (s==CategoricalState.unassigned) {
			if (includeUnassigned)
				sb.append(getUnassignedSymbol());
		}
		else {
			boolean first=true;
			char sep;
			if (CategoricalState.isUncertain(s))
				sep = '/';
			else
				sep = '&';
			for (int e=0; e<=CategoricalState.maxCategoricalState; e++) {
				if (CategoricalState.isElement(s, e)) {
					if (!first)
						sb.append(sep);
					if (forDisplay)
						sb.append(getStateName(ic, e));
					else
						sb.append(getStateSymbol(ic, e));
					first=false;
				}
			}
			if (first)
				sb.append('!'); //no state found!
		}
	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** appends to buffer string describing the state(s) of character ic in taxon it.*/
	public void statesIntoStringBuffer(int ic, int it, StringBuffer sb, boolean forDisplay, boolean includeInapplicable, boolean includeUnassigned){
   		if (notInStorage(ic, it)) //illegal check
			return;
		long s = getStateRaw(ic, it);
		statesIntoStringBufferCore(ic,s,sb, forDisplay, includeInapplicable, includeUnassigned);
	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** appends to buffer string describing the state(s) of character ic in taxon it.*/
	public void statesIntoStringBuffer(int ic, int it, StringBuffer sb, boolean forDisplay){
		statesIntoStringBuffer(ic,it,sb, forDisplay, true, true);
	}
	/*..........................................  CategoricalData  ..................................................*/
   	/** appends to buffer string describing the state(s) of character ic in taxon it. */
	public void statesIntoNEXUSStringBuffer(int ic, int it, StringBuffer sb){
   		if (notInStorage(ic, it)) //illegal check
			return;
		boolean first=true;
		long s = getStateRaw(ic, it);
		if (s==CategoricalState.inapplicable)
			sb.append(getInapplicableSymbol());
		else if (s==CategoricalState.unassigned)
			sb.append(getUnassignedSymbol());
		else {
			int card =0;
			int current = sb.length();
			for (int e=0; e<=CategoricalState.maxCategoricalState; e++) {
				if (CategoricalState.isElement(s, e)) {
					card++;
					if (!first) {
						sb.append(',');
					}
					sb.append(getStateSymbol(ic, e));
					first=false;
				}
			}
			if (card>1) {
				if (CategoricalState.isUncertain(s)) {
					sb.insert(current,'{');
					sb.append('}');
				}
				else {
					sb.insert(current,'(');
					sb.append(')');
				}
			}
			if (first){ //nothing written!  Write illegal character so that change will not go unnoticed
				sb.append('!');
				MesquiteMessage.warnProgrammer("ERROR: nothing written for character " + (ic+1) + " taxon " + (it+1));
			}
		}
	}

	/*..........................................  CategoricalData  ..................................................*/
	CategoricalState tempState = (CategoricalState)makeCharacterState(); //a utility CategoricalState for the fromChar method, which isn't static
	
	/*..........................................  CategoricalData  ..................................................*/
	/**Set the state at character ic from the string s, beginning at position pos in the string. */ 
   	public long fromString(int ic, Parser parser, boolean fromEditor, MesquiteInteger resultCode, MesquiteString result){
   		if (ic>=numChars || ic < 0) { //does not allow request for state for character other than those in matrix
   			if (resultCode!=null)
   				resultCode.setValue(OUTOFBOUNDS);
   		 	return CategoricalState.impossible;
   		 }
   		 
   		 if (parser== null || StringUtil.blank(parser.getString())) { //no string passed
   			if (resultCode!=null)
   				resultCode.setValue(ERROR);
   		 	return CategoricalState.unassigned;
   		 }
		long stateSet = 0;
		int multi = 0;
   		if (fromEditor)  { //string is from editor; thus will use /,& instead of {},() for uncertainty,polymorphism
   			parser.setPunctuationString(puncString);
   			parser.setPosition(0);
   			parser.setWhitespaceString("");  // whitespace can be present WITHIN a state name!
   			String t;
   			while (!StringUtil.blank(t = parser.getNextToken())){
   			//todo: what if linefeed comes in because parser so set?
	   			if (t.length()==1) {  //single character token
	   				char c = t.charAt(0) ;
	   				if (c == ANDseparator || c == ',') 
	   					multi = 1;
	   				else if (c == ORseparator) 
	   					multi = 2;
	   				else if (c == getInapplicableSymbol()) {
						multi = 0;
	   					stateSet = CategoricalState.inapplicable;
	   				}
	   				else if (c== getUnassignedSymbol()){
						multi = 0;
	   					stateSet = CategoricalState.unassigned;
					}
					else if (c == getMatchChar()) {
						stateSet = getStateRaw(ic, 0);
					}
	   				else {
	   					long thisState = fromChar(c);
				  		if (thisState == 0L)
				  			stateSet |= CategoricalState.makeSet(MesquiteInteger.fromString(t));
				  		else
				  			stateSet |= thisState;
				  	}
	   			}
	   			else { //multi character token
	   				boolean found=false;
	   				if (hasStateNames(ic)) { //try to interpret as state name
						for (int i=0; i<stateNames[ic].length && !found; i++) {
							if (t.equalsIgnoreCase(getStateName(ic, i))) {
								stateSet |= CategoricalState.makeSet(i);
								found = true;
							}
						}
					}
					if (!found) { //no state name; possibly several symbols without punctuation
						for (int i=0; i<t.length(); i++){
				  			long stt = fromChar(t.charAt(i));
				  			if (stt == CategoricalState.impossible) {
				  				if (result!=null) {
					  				result.setValue("Illegal entry in cell.");
					  				resultCode.setValue(ERROR);
					  			}
				  				return CategoricalState.impossible;
				  			}
				  			stateSet |= stt;
			  			}
			  			if (stateSet == 0L && result!=null) {
			  				result.setValue("Illegal entry in cell.");
			  				resultCode.setValue(ERROR);
			  			}
			  		}
	   			}
   			}
	   	}

		else { //String is not from editor; assume it's from matrix in NEXUS file.  Thus, assume symbols, not names
			boolean done = false;
			int loc = parser.getPosition();
   			parser.setPunctuationString(null);
			String s = parser.getString();
			while (loc<s.length() && !done) {
				char c = s.charAt(loc++); //get next dark character
				while (parser.whitespace(c)  && c!=0 && loc<s.length())
					c = s.charAt(loc++);
					
				if (!parser.whitespace(c)){
					if (Parser.lineEndCharacter(c)) {
						char r = parser.charOnDeck(1);
						//if next character is also lineEndCharacter but of opposite sort (i.e. \n instead of \r) then eat it up
						if (r!=c && Parser.lineEndCharacter(r))
							loc++;
			   			if (resultCode!=null)
			   				resultCode.setValue(EOL);
						stateSet =CategoricalState.impossible;
						done = true;
					}
					else if (c == '(') //polymorphism
						multi = 1;
					else if (c == '{') //uncertainty
						multi = 2;
					else if (c == '}' || c == ')') 
						done = true;
	   				else if (c == getInapplicableSymbol()) {
						stateSet = CategoricalState.inapplicable;
						multi = 0;
						done = true;
	   				}
	   				else if (c== getUnassignedSymbol()) {
						done = true;
						multi = 0;
						stateSet = CategoricalState.unassigned;
	   				}
					else if (c == getMatchChar()) {
						done = true;
						stateSet = getStateRaw(ic, 0);
					}
					else if (multi>0) {
						long state =  fromChar(c);
						if (CategoricalState.isCombinable(state))
							stateSet |=  state;
					}
					else {
						stateSet =  fromChar(c);
						done = true;
					}
				}
			}
			parser.setPosition(loc);
		}
		if (multi == 2)
			stateSet = CategoricalState.setUncertainty(stateSet, true);
		return stateSet;
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/** Given a character, what state set is implied.  Looks up character among current symbols. */
   	public long fromChar(char state){
		if (state == getInapplicableSymbol())
			return CategoricalState.inapplicable;
		else if (state== getUnassignedSymbol())
			return  CategoricalState.unassigned;
		else if (state == getMatchChar()) {
			return CategoricalState.unassigned; //note: this should't happen, as matchchar with no char #
		}
		else {
   			for (int ist = 0; ist<symbols.length; ist++) {
   				if (symbols[ist]==state) {
   					return CategoricalState.makeSet(ist);
   				}
   			}
   			return tempState.fromChar(state);
  		}
 	}
}

