/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.categ.lib;

import java.awt.*;
import java.util.*;
import mesquite.lib.*;
import mesquite.lib.characters.*;
import mesquite.lib.duties.*;

/* ======================================================================== */
public class CategoricalHistory extends CategoricalAdjustable implements CharacterHistory {
	private int watchPoint=-1;
	
	public CategoricalHistory (Taxa taxa, int numNodes) {
		super(taxa, numNodes);
	}
	public CategoricalHistory (Taxa taxa) {
		super(taxa);
	}
	
	/*..........................................  CategoricalHistory  ..................................................*/
	/**clone this; if passed s is not null, place the clone there (so as to avoid creating new object)*/
	public CharacterHistory clone(CharacterHistory s) {
		CategoricalHistory snew;
		if ((s==null) || (s.getNumNodes()!=numNodes) || (!(s instanceof CategoricalHistory)))
			snew = new CategoricalHistory(getTaxa(), numNodes);
		else {
			snew = (CategoricalHistory)s;
		}
		for (int i=0; i<numNodes; i++) {
			snew.setState(i, getState(i));
		}
		if (frequenciesExist())
			copyFrequencies(this, snew);
		else
			snew.disposeFrequencies();
		snew.data = data;
		snew.characterNumber = characterNumber;  //setting parent character
		if (observedStates!=null)
			snew.setObservedStates((CharacterDistribution)observedStates.getAdjustableClone());
		return snew;
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	/** Returns a new object indicating the states at the tips (used whether or not History is reconstruction)*/
	public CharacterDistribution getStatesAtTips(Tree tree){
		if (observedStates !=null)
			return (CharacterDistribution)observedStates.getAdjustableClone();
		else {
			CharacterState c = getCharacterState();
			CategoricalAdjustable d = (CategoricalAdjustable)c.makeAdjustableDistribution(tree.getTaxa(), tree.getTaxa().getNumTaxa());
			fillDistribution(tree, tree.getRoot(), d);
			return d;
		}
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	private void fillDistribution(Tree tree, int node, CategoricalAdjustable dist) {  
		if (tree.nodeIsTerminal(node)){
			int t = tree.taxonNumberOfNode(node);
			dist.setState(t, getState(node));
		}
		else for (int d = tree.firstDaughterOfNode(node); tree.nodeExists(d); d = tree.nextSisterOfNode(d))
				fillDistribution(tree, d, dist);
	}

	CharacterDistribution observedStates;
	/*..........................................  CategoricalHistory  ..................................................*/
	/** Returns the states in the terminal taxa (used if History is reconstruction)*/
	public CharacterDistribution getObservedStates(){
		return observedStates;
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	/** Sets the states in the terminal taxa (used if History is reconstruction)*/
	public void setObservedStates(CharacterDistribution observed){
		this.observedStates = observed;
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	private void surveyStates(Tree tree, int node) {  // establishes min, max, etc. for use in color returning
		if (checkIllegalNode(node, 9))
			return;
		long s = 0L;
		if (frequenciesExist())
			s=getSetFromFrequencies(node);
		else
			s=getState(node);
			
		allStates|=s; 
		if ((!minFound) || (CategoricalState.minimum(s)<minState)) {
			minState=CategoricalState.minimum(s);
			minFound=true;
		}
		if ((!maxFound) || (CategoricalState.maximum(s)>maxState)) {
			maxState=CategoricalState.maximum(s);
			maxFound=true;
		}
		for (int d = tree.firstDaughterOfNode(node); tree.nodeExists(d); d = tree.nextSisterOfNode(d))
				surveyStates(tree, d);
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	private void surveyTerminalStates(Tree tree, int node) {  // establishes min, max, etc. for use in color returning
		if (tree.nodeIsTerminal(node)) {
			checkIllegalNode(tree.taxonNumberOfNode(node), 10);
			long s = 0L;
			if (frequenciesExist()) 
				s = getSetFromFrequencies(tree.taxonNumberOfNode(node));
			else 
				s=getState(tree.taxonNumberOfNode(node));
			allStates|=s; 
			if ((!minFound) || (CategoricalState.minimum(s)<minState)) {
				minState=CategoricalState.minimum(s);
				minFound=true;
			}
			if ((!maxFound) || (CategoricalState.maximum(s)>maxState)) {
				maxState=CategoricalState.maximum(s);
				maxFound=true;
			}
		}
		for (int d = tree.firstDaughterOfNode(node); tree.nodeExists(d); d = tree.nextSisterOfNode(d))
				surveyTerminalStates(tree, d);
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	/** Must be called before colors for states are used (e.g., before a tree is shaded).  Goes through all nodes to find states present, to set
	minima and maxima and record all states present. This allows the color range to be known before
	colors for particular states are requested by a tree shade, for example.*/
	public void prepareColors(Tree tree, int drawnRoot) {  
		minFound=false;
		maxFound=false;
		allStates=0;
		if (tree.getNumNodeSpaces()>getNumNodes()) {
			surveyTerminalStates(tree, drawnRoot);
		}
		else {
			surveyStates(tree, drawnRoot);
		}
		long s = allStates;
		if (data!=null && getParentCharacter()>=0){
			long obsAll = ((CategoricalData)data).getAllStates(getParentCharacter());
			s |= obsAll;
		}
		else if (observedStates!=null && observedStates instanceof CategoricalDistribution){
			long obsAll = ((CategoricalDistribution)observedStates).getAllStates();
			s |= obsAll;
		}
		if (enforcedMaxState == 0) {
			maxState = CategoricalState.maximum(s);
		}
		else
			maxState = enforcedMaxState;
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	/** places into the already instantiated ColorDistribution the colors corresponding to the CharacterState, and returns the number of colors.  Uses default colors. 
		Mode is MesquiteColorTable.GRAYSCALE, COLORS, COLORS_NO_BW, or DEFAULT (default depends on subclass)*/
	public int getColorsOfState(CharacterState state, ColorDistribution colors, int mode) {
		if (colors==null || state == null || ! (state instanceof CategoricalState))
			return 0;
		colors.initialize();
		int colorCount=0;
		long s = ((CategoricalState)state).getValue();
		for (int i=0; i<=maxState; i++) {
			if (CategoricalState.isElement (s,i)) {
				colors.setWeight(colorCount, 1.0);
				if (data==null) 
					colors.setColor(colorCount++, getDefaultColor(maxState,i));
				else 
					colors.setColor(colorCount++, ((CategoricalData)data).getColorOfState(characterNumber, i, maxState));
			}
		}
		return colorCount;
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	/** places into the already instantiated ColorDistribution the colors corresponding to the CharacterState, and returns the number of colors.  Uses default colors. 
		Mode is MesquiteColorTable.GRAYSCALE, COLORS, COLORS_NO_BW, or DEFAULT (default depends on subclass)*/
	public int getColorsAtNode(int node, ColorDistribution colors, boolean showWeights,int mode) {
		if (checkIllegalNode(node, 11))
			return 0;
		return getColorsAtNode(node, colors, null, showWeights, mode);
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	/** Places into ColorDistribution the colors at the node, and returns the number of colors.  Uses passed
	stateColors as color key.  If null is passed, uses color system of parent data (using maxState of this
	StsOfCharcter) or, if there is not parent data object, the default colors. */
	public int getColorsAtNode(int node, ColorDistribution colors,MesquiteColorTable stateColors, boolean showWeights) {
		return getColorsAtNode(node, colors, stateColors, showWeights, MesquiteColorTable.DEFAULT);
	}
	public Color getDefaultColor(int maxState, int istate){
		return MesquiteColorTable.getDefaultColor(maxState,istate, MesquiteColorTable.COLORS);
	}
	/*..........................................CategoricalHistory................*/
	private int getColorsAtNode(int node, ColorDistribution colors, MesquiteColorTable stateColors, boolean showWeights,int mode) {
		if (checkIllegalNode(node, 12))
			return 0;
		if (colors==null)
			return 0;
		colors.initialize();
		int colorCount=0;
		long s;
		if (frequenciesExist() && showWeights)
			s = CategoricalState.compressFromPresence(getFrequencies(node));
		else 
			s = states[node];

		for (int i=0; i<=maxState; i++) {
			if (CategoricalState.isElement (s,i)) {
				if (frequenciesExist() && showWeights)
					colors.setWeight(colorCount, getFrequency(node, i));
				else
					colors.setWeight(colorCount, 1.0);
				if (stateColors==null) {
					if (data==null) {
						colors.setColor(colorCount++, getDefaultColor(maxState,i));
					}
					else {
						colors.setColor(colorCount++, ((CategoricalData)data).getColorOfState(characterNumber, i, maxState));
					}
				}
				else {
					colors.setColor(colorCount++, stateColors.getColor(maxState,i));
				}
			}
		}
		if (colorCount<=0) {
			colors.setColor(0, ColorDistribution.unassigned);
			colorCount = 1;
		}
		return colorCount;
	}
	/*..........................................CategoricalHistory................*/
	/**Get the information about states occurring in this object for a legend.  Colors and state names of states
	are place in the passed array.  The colors for states are taken from the passed MesquiteColorTable.*/
	public int getLegendStates(Color[] cs, String[] stateNames, MesquiteColorTable stateColors) {
		return getLegendStates(cs, stateNames, stateColors, MesquiteColorTable.DEFAULT);
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	protected int getLegendStates(Color[] cs, String[] stateNames, MesquiteColorTable stateColors, int mode) {
		if (cs==null || stateNames == null)
			return 0;
		int colorCount=0;
		long s = allStates;
		int maxStateLegend = maxState;
		
		if (data!=null && getParentCharacter()>=0){
			long obsAll = ((CategoricalData)data).getAllStates(getParentCharacter());
			s |= obsAll;
			int mso = ((CategoricalData)data).getMaxState(getParentCharacter());
			if (maxStateLegend<mso)
				maxStateLegend = mso;
		}
		for (int e=0; e<=maxStateLegend && colorCount<cs.length && colorCount<stateNames.length; e++) {
			if (CategoricalState.isElement (s,e)) {
				if (stateColors==null) {
					if (data == null)
						cs[colorCount]= getDefaultColor(maxStateLegend,e);
					else
						cs[colorCount]= ((CategoricalData)data).getColorOfState(-1,e, maxStateLegend);
				}
				else {
					cs[colorCount]= stateColors.getColor(maxStateLegend,e);
				}
				if (data!=null && characterNumber>=0 && characterNumber<data.getNumChars() && data instanceof CategoricalData) {
					stateNames[colorCount++]=((CategoricalData)data).getStateName(characterNumber, e);
				}
				else
					stateNames[colorCount++]=Integer.toString(e);
			}
		}
		/*if (colorCount<=0)
			MesquiteMessage.println("getLegendStates: no states found");
		*/
		return colorCount;
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	/**Get the information about states occurring in this object for a legend.  Colors and state names of states
	are place in the passed array.*/
	public int getLegendStates(Color[] cs, String[] stateNames, int mode) {
		return getLegendStates(cs, stateNames, null, mode);
	}
	/*..........................................  CategoricalHistory  ..................................................*/
	public CharacterHistory adjustSize(Tree tree) {
		if (tree.getNumNodeSpaces() == this.getNumNodes())
			return this;
		else {
			CategoricalHistory soc = new CategoricalHistory(tree.getTaxa(), tree.getNumNodeSpaces());
			soc.setParentData(getParentData());
			soc.setParentCharacter(getParentCharacter());
			return soc;
		}
	}
	public CharacterHistory makeHistory(Tree tree){
		return new CategoricalHistory(tree.getTaxa(), tree.getNumNodeSpaces());
	}
	/*................................................................................................................*/
	public void polymorphToUncertainties(int N, Tree tree) {
		for (int d = tree.firstDaughterOfNode(N); tree.nodeExists(d); d = tree.nextSisterOfNode(d)) {
			polymorphToUncertainties(d, tree);
		}
		if (tree.nodeIsInternal(N)){
			long s = getState(N);
			if (CategoricalState.cardinality(s)>1)
				setState(N, CategoricalState.setUncertainty(s, true));
		}
	}
}

