/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.categ.lib;

import java.awt.*;
import java.util.*;
import mesquite.lib.*;
import mesquite.lib.characters.*;
import mesquite.lib.duties.*;

/* ======================================================================== */
/** A subclass of CharacterData for DNA data stored as Categorical sets (e.g, "{A, C}" */
public class DNAData extends CategoricalData {
	boolean displayAsRNA = false;
	boolean colorCodPosInColumnNumber = true;
	static Color dnaRed, dnaGreen, dnaYellow, dnaBlue;
	CodonPositionsSet codonPositionsSet=null;
	private static Vector defaultModels;
	static String A, C, G, T;
	static {
		defaultModels= new Vector(2);
		dnaRed = new Color((float)1, (float)0.3, (float)0.3);
		dnaGreen = new Color((float)0.3, (float)1, (float)0.3);
		dnaYellow = new Color((float)1, (float)1, (float)0.3);
		//dnaYellow = Color.yellow;
		dnaBlue = new Color((float)0.4, (float)0.4, (float)1);
		A = "A";
		C = "C";
		G = "G";
		T = "T";
	}
	public DNAData(CharMatrixManager manager, int numTaxa, int numChars, Taxa taxa){
		super(manager, numTaxa, numChars, taxa);
		
		super.setSymbolDirect(0, 'A');
		super.setSymbolDirect(1, 'C');
		super.setSymbolDirect(2, 'G');
		super.setSymbolDirect(3, 'T');
	}
	/*..........................................  DNAData  ..................................................*/
	public void setDisplayAsRNA(boolean displayAsRNA){
  		this.displayAsRNA = displayAsRNA;
	}
	/*..........................................  DNAData  ..................................................*/
	public boolean getDisplayAsRNA(){
  		return displayAsRNA;
	}
	/*..........................................  DNAData  ..................................................*/
	public long getDefaultState(){
  		return CategoricalState.inapplicable;
	}
	/*..........................................  DNAData  ..................................................*/
	/** Indicates the type of character stored */ 
	public Class getStateClass(){
		return DNAState.class;
	}
	/*.................................................................................................................*/
	/** Indicates whether the data are molecular sequence data or not */ 
	public boolean isMolecularSequence() {
		return true;
	}
	/*..........................................  DNAData  ..................................................*/
	public String getDataTypeName(){
		return "DNA Data";
	}
	/*..........................................  DNAData  ..................................................*/
	/**clone this CharacterData and return new copy */
	public CharacterData cloneData(){
		DNAData data = new DNAData(matrixManager, numTaxa, numChars, getTaxa());
		for (int ic=0; ic<numChars; ic++){
			for (int it=0; it<numTaxa; it++) {
				data.setState(ic, it, getStateRaw(ic, it)); 
			}
		}
		data.resetChangedSinceSave();
		return data;
	}
	/*..........................................  DNAData  ..................................................*/
	/**get matrix and return as StsOfCharacters */
   	public MCharactersDistribution getMCharactersDistribution(){
   		MDNAEmbedded states = new MDNAEmbedded(this);
		return states;
   	}
	/*..........................................  DNAData  ..................................................*/
	/**return a CharacterDistribution that points to character ic */
   	public CharacterDistribution getCharacterDistribution(int ic){
   		DNAEmbedded states = new DNAEmbedded(this, ic);
		 return states;
   	}
	/*..........................................  DNAData  ..................................................*/
	/**Return CharacterDistribution object with same number of taxa as this object (but not with any particular data; just so that of same data type) */
   	public CharacterDistribution makeCharacterDistribution(){
   		DNACharacterAdjustable c= new DNACharacterAdjustable(getTaxa(), numTaxa);
   		c.setParentData(this);
   		return c;
   	}
	/*..........................................  DNAData  ..................................................*/
   	/** creates an empty CharacterState object of the same data type as CharacterData subclass used.*/
   	public CharacterState makeCharacterState(){
   		return new DNAState();
   	}
	/*..........................................  DNAData  ..................................................*/
	/**Return CharacterData object with same number of taxa & characters as this object (but not with any particular data; just so that of same data type) */
   	public CharacterData makeCharacterData() {
   		return new DNAData(getMatrixManager(), getNumTaxa(), getNumChars(), getTaxa());
   	}
	/*..........................................  DNAData  ..................................................*/
	/**Return CharacterData object with passed number of taxa & characters (but not with any particular data; just so that of same data type) */
   	public CharacterData makeCharacterData(int ntaxa, int nchars) {
   		return new DNAData(getMatrixManager(), ntaxa, nchars, getTaxa());
   	}
   	public CharacterData makeCharacterData(CharMatrixManager manager, Taxa taxa){ 
   		return new DNAData(getMatrixManager(), taxa.getNumTaxa(), 0, taxa);
   	}
	/*..........................................  DNAData  ..................................................*/
   	/** returns whether the matrix would prefer to have columns sized individually in editors.  Default is true.*/
	public boolean pleaseAutoSizeColumns() {
		return false;
	}
	/*..........................................  DNAData  ..................................................*/
   	public boolean colorCellsByDefault(){ //added 19Jan02
   		return true;
   	}
	/*..........................................  DNAData  ..................................................*/
   	/** returns the state of character ic in taxon it*/
   	public  CharacterState getCharacterState(CharacterState cs, int ic, int it){
   		if (notInStorage(ic, it)) {
   			MesquiteMessage.warnUser("character number (" + ic + ") or taxon number (" + it + ") out of bounds in getCharacterState");
   			return null;
   		}
   		if (cs!=null && cs.getClass()==DNAState.class) {
   			((DNAState)cs).setValue(getState(ic, it));
   			return cs;
		}
   		return new DNAState(getState(ic, it));
   	}
	/*..........................................  DNAData  ..................................................*/
	public Color getColorOfState(int ic, int istate, int maxState){
		return getDNAColorOfState(istate);
	}
	public Color getColorOfState(int ic, int istate){
		return getDNAColorOfState(istate);
	}
	/*..........................................  CategoricalData  ..................................................*/
	/** Gets the color representing state(s) of character ic in taxon it */ 
	public Color getColorOfStates(int ic, int it){
   		if (notInStorage(ic, it)) //illegal check
			return ColorDistribution.unassigned;
		long s = getStateRaw(ic, it);
		if (CategoricalState.isCombinable(s)) {
			int colorCount = CategoricalState.cardinality(s);
			if (colorCount>1){
				return Color.lightGray;
			}
			else {
				return getDNAColorOfState(CategoricalState.maximum(s));
			}
		}
		else 
			return ColorDistribution.unassigned;
	}
	public static Color getDNAColorOfState(int istate){
		if (istate == 0)
			return dnaRed;
		else if (istate == 1)
			return dnaGreen;
		else if (istate == 2)
			return dnaYellow;
		else if (istate ==3)
			return dnaBlue;
		else
			return ColorDistribution.unassigned;
	}
  	/** returns the color of character ic; e.g., to indicate codon positions */
   	public Color getDefaultCharacterColor(int ic){
		if (colorCodPosInColumnNumber) {
			int position = getCodonPosition(ic);
			if (position>=1 && position <=3)
				return ColorDistribution.codPosMedium[position-1];
			else
				return null; //ColorDistribution.codPosMedium[3];
		}
		return super.getDefaultCharacterColor(ic);
   	}
  	/** returns the dark color of character ic; e.g., to indicate codon positions */
   	public Color getDarkDefaultCharacterColor(int ic){
		if (colorCodPosInColumnNumber) {
			int position = getCodonPosition(ic);
			if (position>=1 && position <=3)
				return ColorDistribution.codPosDark[position-1];
			else
				return ColorDistribution.codPosDark[3];
		}
		return null;
   	}
	/*..........................................  DNAData  ..................................................*/
   	/** returns the codon position of character ic in taxon it*/
   	public  int getCodonPosition(int ic){
		if (codonPositionsSet==null)
			codonPositionsSet = (CodonPositionsSet)getCurrentSpecsSet(CodonPositionsSet.class);
		if (codonPositionsSet != null && ic<codonPositionsSet.getNumberOfParts() && ic>=0)
			return codonPositionsSet.getInt(ic);
		return MesquiteInteger.unassigned;
   	}
	/*..........................................  DNAData  ..................................................*/
   	/** returns if there is a current codon position set*/
   	public  boolean isCurrentCodonPositionsSet(){
		if (codonPositionsSet==null)
			codonPositionsSet = (CodonPositionsSet)getCurrentSpecsSet(CodonPositionsSet.class);
		return (codonPositionsSet!=null);
   	}
 	/*.................................................................................................................*/
 	/** gets the explanation of this matrix*/
	public String getExplanation() {
		String extra = "This DNA character matrix for the taxa block \"" + getTaxa().getName() + "\" has " + getNumChars() + " characters for the " + getNumTaxa() + " taxa. Category of data: " + getDataTypeName() + "\n";
		int[] freq = new int[4];
		for (int i = 0; i<4; i++)
			freq[i] = 0;
		int cg = 0;
		for (int it = 0; it<getNumTaxa(); it++) {
			for (int ic = 0; ic<getNumChars(); ic++) {
					long s = getStateRaw(ic,it);
					if (!CategoricalState.isUnassigned(s) && !CategoricalState.isInapplicable(s)) {
						cg++;
						for (int i=0; i<4; i++)
							if (CategoricalState.isElement(s, i))
								freq[i]++;
					}
			}
		}
		if (cg != 0) {
			extra += "Frequency of A: " + MesquiteDouble.toString(((double)freq[0])/cg) + " C: " + MesquiteDouble.toString(((double)freq[1])/cg) + " G: " + MesquiteDouble.toString(((double)freq[2])/cg) + " T: " + MesquiteDouble.toString(((double)freq[3])/cg);
		}
		
		return extra;

	}
	/*..........................................DNAData................*/
   	public static DefaultReference findDefaultReference(NameReference paradigm){
   		if (defaultModels == null) {
   			MesquiteMessage.warnProgrammer("findDefaultReference with null default models ");
   			MesquiteMessage.printStackTrace();
   			return null;
   		}
   		for (int i=0; i<defaultModels.size(); i++){
   			DefaultReference dR = (DefaultReference)defaultModels.elementAt(i);
   			
   			if (dR.getParadigm() != null && dR.getParadigm().equals(paradigm))
   				return dR;
   		}
   		return null;
   		
   	}
	/*..........................................DNAData................*/
   	public static void registerDefaultModel(String paradigm, String name){
   		if (defaultModels == null)
   			return;
   		NameReference p = NameReference.getNameReference(paradigm);
   		DefaultReference dR = findDefaultReference(p);
		if (dR == null) {
			dR = new DefaultReference(p);
			defaultModels.addElement(dR);
		}
		dR.setDefault(name);
   	}
	/*..........................................DNAData................*/
   	public CharacterModel getDefaultModel(String paradigm){
   		NameReference p = NameReference.getNameReference(paradigm);
   		DefaultReference dR = findDefaultReference(p);
   		if (dR==null)
   			return null;
   		else {
   			CharacterModel cm = getProject().getCharacterModel(dR.getDefault());
   			if (cm==null) 
   				MesquiteMessage.println("Default model not found / " + dR.getDefault());
   			return cm;
   		}
   	}
	/*..........................................  CategoricalData  ..................................................*/
	/**Returns the states at character ic and taxon it as a string */
	public String getIUPACSymbol(long s) {
		boolean s0 = CategoricalState.isElement(s,0);
		boolean s1 = CategoricalState.isElement(s,1);
		boolean s2 = CategoricalState.isElement(s,2);
		boolean s3 = CategoricalState.isElement(s,3);
		
		if (s0 && s1 && s2 && s3)
			return "N";
		else if (s0 && s1 && s2)
			return "V";
		else if (s0 && s2 && s3)
			return "D";
		else if (s0 && s1 && s3)
			return "H";
		else if (s1 && s2 && s3)
			return "B";
		else if (s0 && s1)
			return "M";
		else if (s0 && s2)
			return "R";
		else if (s0 && s3)
			return "W";
		else if (s1 && s2)
			return "S";
		else if (s1 && s3)
			return "Y";
		else if (s2 && s3)
			return "K";
		return "X";
	}
	/*..........................................  DNAData  ..................................................*/
   	/** appends to buffer string describing the state(s) of character ic in taxon it.*/
	public void statesIntoStringBufferCore(int ic, long s, StringBuffer sb, boolean forDisplay){
		statesIntoStringBufferCore(ic, s, sb, forDisplay, true, true);
	}
	/*..........................................  DNAData  ..................................................*/
   	/** appends to buffer string describing the state(s) of character ic in taxon it.*/
	public void statesIntoStringBufferCore(int ic, long s, StringBuffer sb, boolean forDisplay, boolean includeInapplicable, boolean includeUnassigned){
		if (CategoricalState.cardinality(s) > 1 && CategoricalState.isUncertain(s))
			sb.append(getIUPACSymbol(s));
		else {
			if (s==CategoricalState.inapplicable) {
				if (includeInapplicable)
					sb.append(getInapplicableSymbol());
			}
			else if (s==CategoricalState.unassigned) {
				if (includeUnassigned)
					sb.append(getUnassignedSymbol());
			}
			else {
				boolean first=true;
				char sep;
				if (CategoricalState.isUncertain(s))
					sep = '/';
				else
					sep = '&';
				boolean lowerCase = CategoricalState.isLowerCase(s);
				for (int e=0; e<=CategoricalState.maxCategoricalState; e++) {
					if (CategoricalState.isElement(s, e)) {
						if (!first)
							sb.append(sep);
						appendStateSymbol(e, lowerCase, sb);
						first=false;
					}
				}
				if (first)
					sb.append('!'); //no state found!
			}
		}
	}
	/*............................................................................................*/
	/** gets default state symbol for state "state", returned as string. */
	public static String getDefaultStateSymbol(int state) {
		if (state==0)
			return "A";
		else if (state==1)
			return "C";
		else if (state==2)
			return "G";
		else if (state==3)
			return "T";
		else
			return Integer.toString(state);
	}
	/*..........................................  DNAData  ..................................................*/
	/* Appends to buffer state symbol for state e */
	protected void appendStateSymbol(int e, boolean lowerCase, StringBuffer sb){
		if (lowerCase){
			if (e==0)
				sb.append('a');
			else if (e==1)
				sb.append('c');
			else if (e==2)
				sb.append('g');
			else if (e==3) {
				if (displayAsRNA)
					sb.append('u');
				else
					sb.append('t');
			}
		}
		else {
			if (e==0)
				sb.append('A');
			else if (e==1)
				sb.append('C');
			else if (e==2)
				sb.append('G');
			else if (e==3) {
				if (displayAsRNA)
					sb.append('U');
				else
					sb.append('T');
			}
		}
	}
	
	/*..........................................  DNAData  ..................................................*/
	/* Fills buffer with string version of state in char ic and taxon it */
	public void statesIntoNEXUSStringBuffer(int ic, int it, StringBuffer sb){
		boolean first=true;
		long s = getStateRaw(ic, it);
		if (s==CategoricalState.inapplicable)
			sb.append(getInapplicableSymbol());
		else if (s==CategoricalState.unassigned)
			sb.append(getUnassignedSymbol());
		else {
			int card =0;
			int current = sb.length();
			for (int e=0; e<=DNAState.maxDNAState; e++) {
				if (CategoricalState.isElement(s, e)) {
					card++;
					if (!first)
						sb.append(',');
					appendStateSymbol(e, CategoricalState.isLowerCase(s), sb);
					first=false;
				}
			}
			if (card>1) {
				if (CategoricalState.isUncertain(s)) {
					sb.insert(current,'{');
					sb.append('}');
				}
				else {
					sb.insert(current,'(');
					sb.append(')');
				}
			}
			if (first){ //nothing written!  Write illegal character so that change will not go unnoticed
				sb.append('!');
				MesquiteMessage.warnProgrammer("ERROR: nothing written for character " + (ic+1) + " taxon " + (it+1));
			}
		}
	}
	/*..........................................  DNAData  ..................................................*/
	/* Returns state set from single state symbol  */
   	public long fromChar(char state){
		if (state == getInapplicableSymbol())
			return CategoricalState.inapplicable;
		else if (state== getUnassignedSymbol() || state == 'N')
			return  CategoricalState.unassigned;
		else if (state == getMatchChar()) {
			return CategoricalState.unassigned;//note: this should't happen, as matchchar with no char #
		}
		else {
   			return DNAState.fromCharStatic(state);
  		}
 	}
}

