/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/

package mesquite.lib;

import java.awt.*;
import java.util.*;
import java.awt.event.*;
import java.io.*;
import mesquite.lib.duties.*;


/* ======================================================================== */
/** An object that sends commands; for use in console input (e.g., ConsoleWindow, ConsoleThread). */
public class CommandCommunicator {
	Puppeteer puppeteer;
	Object objectCommanded = null;
	File currentDirectory = null;
	Object previousObject = null;
	Object result;
	boolean echoToSystemOut;
	boolean done = false;
	public CommandCommunicator(MesquiteModule module, Object initCommanded, boolean echoToSystemOut){
		puppeteer = new Puppeteer(module);
		objectCommanded = initCommanded;
		this.echoToSystemOut = echoToSystemOut;
		if (MesquiteTrunk.suggestedDirectory!=null)
			currentDirectory = new File(MesquiteTrunk.suggestedDirectory);
	}
	
	String objectName(Object obj){
		if (obj == null)
			return "no object";
		else if (obj instanceof Listable) {
			if (obj instanceof MesquiteWindow)
				return "Window: " + ((Listable)obj).getName();
			else
				return ((Listable)obj).getName();
		}
		else
			return obj.getClass().getName();
	}
	private void write(String s){
		if (echoToSystemOut)
			MesquiteTrunk.mesquiteTrunk.log(s);
		else
			MesquiteTrunk.mesquiteTrunk.logNoEcho(s); //here use nonechoversion
	}
	
	//THIS SHOULD NOT BE CALLED ON MAIN EVENT THREAD, as it sleeps until answer returned
	public void enterTyped(String commandLine){
		write("\n");
		if (commandLine == null) {
			showPrompt();
			return;
		}
		MesquiteInteger pos = new MesquiteInteger(0);
		String command = ParseUtil.getToken(commandLine, pos);
		String arguments = commandLine.substring(pos.getValue(), commandLine.length());
		if (arguments !=null){
			arguments = StringUtil.stripTrailingWhitespace(arguments);
			if (arguments.length() == 0)
				arguments = null;
			else if (arguments.charAt(arguments.length()-1) == ';')
				arguments = arguments.substring(0, arguments.length()-1);
			if (arguments !=null && ";".equals(arguments))
				arguments = null;
			if (arguments !=null && arguments.length() == 0)
				arguments = null;
		}
		if ("who".equalsIgnoreCase(command)) {
			write("Object being commanded: " + objectName(objectCommanded) + "\n");
		}
		else if ("help".equalsIgnoreCase(command)) { //console help
//*****put all of these here at this level as asynchronous
//***** make commands to window and module different (window:)
//***** put console window listing on other thread
//***** single line commands send to command, not puppetter
			write("\nConsole mode commands:\n");
			write("who  - indicates what is currently commanded object\n");
			write("?    -show commands of currently commanded object (equivalently, sc)\n");
			write("??    -show commands of currently commanded object, with explanations\n");
			write("? <name of command>    -show explanation for that command of currently commanded object\n");
			write("..    - if MesquiteModule, command employer module instead; if window, command its owner module instead\n");
			write(">    - command the returned object\n");
			write("<    - command the previously commanded object\n");
			write("se    - show immediate employee modules (command to be given to module)\n");
			write("ge  <number>  - shift command to employee module, numbered as from command se\n");
			write("etree    - show employee tree (command to be given to module)\n");
			write("jet  <number>  - jump to employee module, numbered as from command etree\n");
			write("gw  - shift command to window belonging to module (command to be given to module)\n");
			write("sw  - list current windows\n");
			write("jw  <number> - jump to window, numbered as from command sw\n");
			write("menus - show menus of current window (window related to object commanded)\n");
			write("menu <number> - show items of given menu of current window (window related to object commanded)\n");
			write("text  - show text version of window (command to be given to window)\n");
			write("sb  - show buttons of any dialog up front\n");
			write("press <buttonName>  - press buttons of any dialog up front\n");
			write("cd  - show current directory.  The current directory is to explore contents of disk, and as default for openFile command.\n");
			write("cd ..  - change directory to parent\n");
			write("cd <directoryPath>  - change directory to path\n");
			write("ls  - list contents of current directory\n");
			//ALSO needed: "tell"
		} 
		else if ("se".equalsIgnoreCase(command)) {
			if (objectCommanded instanceof MesquiteModule){
				MesquiteModule mb = (MesquiteModule)objectCommanded;
				write("Employee Modules of " + mb.getName() +"\n");
				for (int i=0; i<mb.getEmployeeVector().size(); i++) {
					MesquiteModule mbb = (MesquiteModule)mb.getEmployeeVector().elementAt(i);
					write("  " + i + "  --  " + mbb.getName() +"\n");
				}
			}
			else
				write("The se command can be given only to modules");
		}
		else if ("cd".equalsIgnoreCase(command)) {
			if (MesquiteTrunk.suggestedDirectory!=null && currentDirectory == null)
				currentDirectory = new File(MesquiteTrunk.suggestedDirectory);
			
			if ("..".equals(arguments)) {
				if (currentDirectory !=null)
					currentDirectory = new File(currentDirectory.getParent());
			}
			else if (!StringUtil.blank(arguments)){
				File cd = new File(arguments);
				if (cd.exists() && cd.isDirectory())
					currentDirectory = cd;
				else if (currentDirectory !=null) {
					cd = new File(currentDirectory + MesquiteFile.fileSeparator + arguments);
					if (cd.exists() && cd.isDirectory())
						currentDirectory = cd;
				}
			}
			write("Current directory: " + currentDirectory);
		}
		else if ("ls".equalsIgnoreCase(command)) {
			if (MesquiteTrunk.suggestedDirectory!=null && currentDirectory == null)
				currentDirectory = new File(MesquiteTrunk.suggestedDirectory);
			if (currentDirectory !=null && currentDirectory.exists() && currentDirectory.isDirectory()){
				String[] list = currentDirectory.list();
				for (int i=0; i<list.length; i++) {
					if (i % 3 == 1)
						write(list[i] +"\n");
					else if (i % 3 == 0)
						write("\t" + list[i] +"\n");
					else
						write("\t" + list[i]);
				}
			}
		}
		else if ("sw".equalsIgnoreCase(command)) {
			ListableVector w = MesquiteTrunk.windowVector;
			if (w == null || w.size() == 0)
				write("There are no windows showing");
			else {
				write("Current Windows:\n");
				for (int i=0; i<w.size(); i++) {
					MesquiteWindow mbb = (MesquiteWindow)w.elementAt(i);
					write("  " + i + "  --  " + mbb.getName() +"\n");
				}
			}
		}
		else if ("jw".equalsIgnoreCase(command)) {
			int i = MesquiteInteger.fromString(arguments);
			ListableVector w = MesquiteTrunk.windowVector;
			if (w == null || w.size() == 0)
				write("There are no windows showing");
			else if (i< w.size() && MesquiteInteger.isCombinable(i)){
				previousObject = objectCommanded;
				objectCommanded = w.elementAt(i);
			}
		}
		else if ("menus".equalsIgnoreCase(command)) {
			MenuBar menus = getMenuBar(objectCommanded);
			if (menus !=null){
				for (int i= 0; i<menus.getMenuCount(); i++){
					Menu menu = menus.getMenu(i);
					write("  " + i + "  --  " + menu.getLabel() +"\n");
				}
			}
			
		}
		else if ("menu".equalsIgnoreCase(command)) {
			MenuBar menus = getMenuBar(objectCommanded);
			if (menus !=null){
				int i = MesquiteInteger.fromString(arguments);
				if (i>=0 && i<menus.getMenuCount()){
					Menu menu = menus.getMenu(i);
					write("  Menu: " + menu.getLabel() +"\n");
					for (int j = 0; j<menu.getItemCount(); j++) {
						if (!("-".equals(menu.getItem(j).getLabel())))
							write("      " + j + "  --  " + menu.getItem(j).getLabel() +"\n");
					}
				}
			}
			
		}
		else if ("sb".equalsIgnoreCase(command)) {
			ListableVector d = MesquiteModule.mesquiteTrunk.dialogVector;
			
			if (d!=null && d.size()>0) {
				Object obj = d.elementAt(d.size()-1);
				if (obj !=null && obj instanceof MesquiteDialog){
					MesquiteDialog dlog = (MesquiteDialog)obj;
					Vector v = new Vector();
					dlog.getButtons(dlog,  v);
					if (v.size()==0)
						write("There are no buttons");
					else {
						write("Buttons of dialog (to use type \"press buttonName\"):\n");
						for (int i= 0; i< v.size(); i++){
							write("(" + (String)v.elementAt(i) + ") ");
						}
						write("\n");
					}
				}
			}
			else
				write("There are no dialog boxes being shown at the moment.\n");
		}
		else if ("press".equalsIgnoreCase(command)) {
			ListableVector d = MesquiteModule.mesquiteTrunk.dialogVector;
			
			if (d!=null && d.size()>0) {
				Object obj = d.elementAt(d.size()-1);
				if (obj !=null && obj instanceof MesquiteDialog){
					MesquiteDialog dlog = (MesquiteDialog)obj;
					dlog.selectButton(arguments);
				}
			}
			else
				write("There are no dialog boxes being shown at the moment.\n");
		}
		else if ("etree".equalsIgnoreCase(command)) {
			if (objectCommanded instanceof MesquiteModule){
				MesquiteModule mb = (MesquiteModule)objectCommanded;
				write("Employee Tree of " + mb.getName() + "  use jet <number> to jump to command that employee\n" + mb.listEmployees("  ", new MesquiteInteger(0)) + "\n");
			}
			else
				write("The etree command can be given only to modules");
		}
		else if ("jet".equalsIgnoreCase(command)) {
			if (objectCommanded instanceof MesquiteModule){
				MesquiteModule mb = (MesquiteModule)objectCommanded;
				int i = MesquiteInteger.fromString(arguments);
				if (MesquiteInteger.isCombinable(i)) {
					Object r = mb.getDeepEmployee(i, new MesquiteInteger(0));
					if (r !=null) {
						previousObject = objectCommanded;
						objectCommanded = r;
					}
				}
			}
			else
				write("The jet command can be given only to modules");
		}
		else if ("ge".equalsIgnoreCase(command)) {
			if (objectCommanded instanceof MesquiteModule){
				MesquiteModule mb = (MesquiteModule)objectCommanded;
				int i = MesquiteInteger.fromString(arguments);
				if (MesquiteInteger.isCombinable(i) && i>=0 && i<mb.getEmployeeVector().size()) {
					Object r = mb.getEmployeeVector().elementAt(i);
					if (r !=null) {
						previousObject = objectCommanded;
						objectCommanded = r;
					}
				}
			}
			else
				write("The ge command can be given only to modules");
		}
		else if ("gw".equalsIgnoreCase(command)) {
			if (objectCommanded instanceof MesquiteModule){
				MesquiteModule mb = (MesquiteModule)objectCommanded;
				Object r = mb.getModuleWindow();
				if (r !=null) {
					previousObject = objectCommanded;
					objectCommanded = r;
				}
				
			}
			else
				write("The gw command can be given only to modules");
		}
		else if (">".equalsIgnoreCase(command)) { 
			if (result != null) {
				previousObject = objectCommanded;
				objectCommanded = result;
			}
		}
		else if ("<".equalsIgnoreCase(command)) { 
			if (previousObject != null)
				objectCommanded = previousObject;
		}
		else if ("..".equalsIgnoreCase(command)) { 
			Object r = null;
			if (objectCommanded instanceof MesquiteModule) {
				r = ((MesquiteModule)objectCommanded).getEmployer();
				if (r !=null){
					previousObject = objectCommanded;
					objectCommanded = r;
				}
			}
			else if (objectCommanded instanceof MesquiteWindow){
				r = ((MesquiteWindow)objectCommanded).getOwnerModule();
				if (r !=null){
					previousObject = objectCommanded;
					objectCommanded = r;
				}
			}
		}
		else if (!StringUtil.blank(commandLine)) {
			String c = StringUtil.stripTrailingWhitespace(commandLine);
			if (c.length()>0 && c.charAt(c.length()-1) != ';')
				c += ";";
			done = false;
			//HERE set up Pending command and put it on command queue
			PendingCommand pc = new PendingCommand(this,  puppeteer, objectCommanded, c, false);
			pc.setFromCommandLine(true);
			MainThread.pendingCommands.addElement(pc, false);
			//then sleep inutil Pending command calls back commandDone method of this CommandCommunicator with an object reutrned;
			try {
				while (!done)
					Thread.sleep(20);
			}
			catch (InterruptedException e){
			}
			if (result instanceof Commandable)
				write("    [> " + objectName(result) + " ]");
			else {
				//write("    [ result: " + result + " ]");
				result = null;
			}
		}
		showPrompt();
		
	}
	private MenuBar getMenuBar(Object objectCommanded){
			MesquiteWindow w= null;
			if (objectCommanded instanceof MesquiteWindow)
				w = (MesquiteWindow)objectCommanded;
			else if (objectCommanded instanceof MesquiteModule) {
				Frame f = ((MesquiteModule)objectCommanded).containerOfModule();
				if (f instanceof MesquiteWindow)
					w = (MesquiteWindow)f;
			}
			else if (objectCommanded instanceof OwnedByModule){
				MesquiteModule mb = ((OwnedByModule)objectCommanded).getOwnerModule();
				if (mb !=null) {
					Frame f = mb.containerOfModule();
					if (f instanceof MesquiteWindow)
						w = (MesquiteWindow)f;
				}
			}
			MenuBar menus = w.getMenuBar();
			return menus;
	}
	public void commandDone(Object returned){
		result = returned;
		done = true;
	}
	public void showPrompt(){
			write("\n(" + objectName(objectCommanded) + ") # ");
	}
}
