/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)

Modifications:
28 July 01 (WPM) getCompatibleFileElements added
*/
package mesquite.lib;

import java.awt.*;
import java.util.*;
import java.net.*;
import mesquite.lib.duties.*;
import mesquite.lib.characters.*;

/* ======================================================================== */
/** An object of this class represents a collection of information (data matrices, tree blocks, etc.) that may reside in more than one
file (MesquiteFile, which represents files on disk or at a URL).  More than one project can be open at once.  Information from different projects
generally doesn't interact (e.g., a data matrix from one project can't belong to a set of taxa from another project), but within a single project
information can interact regardless of what file the information belongs to.<p>
Although more than one file can belong to the project, the first file opened is treated as the "home file".  In the future the idea of a "project" may be made stronger,
and a separate sort of file (the project file) may be created.  tHis would eliminate the need for a "home file".
*/
public class MesquiteProject implements Listable, Commandable, HNode {
	LeakFinder leakFinder = MesquiteTrunk.leakFinderObject;
	/** A vector of the sets of taxa within the project.*/
	public ListableVector taxas;  
	/** Vector of data sets (including matrix) within the project. */
	public ListableVector datas;  
	/** Character models belonging to the project.*/
	protected ModelVector charModels; 
	/** Other file elements belonging to the project.*/
	protected ListableVector otherElements; 
	
	/** Nexus blocks belonging to the files in the project.*/
	protected ListableVector nexusBlocks; 
	/** */
	CentralModelListener modelListener;
	
	public MesquiteWindow windowToActivate, activeWindowOfProject;

	/** The files belonging to the project. */
	ListableVector files;
	/** The home file of the project. */
	MesquiteFile homeFile;
	
	/** Commands used with standard Close, Save and Save As...and other menu items*/
	private MesquiteCommand closeCommand, closeFilesCommand, saveCommand, saveFilesCommand, saveAsCommand, saveLinkagesCommand, linkFileCommand, linkURLCommand, includeFileCommand, includeURLCommand;
	public MesquiteCommand exportCommand, newLinkFileCommand, getInfoCommand;
	/** The file coordinator module that owns this object*/
	public FileCoordinator ownerModule=null; 
	
	/** Background color scheme for project's windows*/ 
	public int projectColor;  

	public static int totalCreated = 0;  // a static counter of the number of projects that have been opened
	public static int totalDisposed = 0;  // to catch memory leaks
	public static int totalFinalized = 0;  // to catch memory leaks
	private int idNumber;
	public boolean developing = true;
	public boolean isDoomed = false;
	
	public MesquiteProject(FileCoordinator ownerModule) {
		this.ownerModule = ownerModule;
		datas = new ListableVector();  
		taxas = new ListableVector();  
		charModels = new ModelVector();
		modelListener = new CentralModelListener();
		otherElements = new ListableVector();
		files = new ListableVector();
		nexusBlocks = new ListableVector( 1);
		idNumber = totalCreated;
		totalCreated++;
		
		projectColor = MesquiteTrunk.mesquiteTrunk.projects.requestNextColor();
		
		closeCommand= MesquiteModule.makeCommand("closeFile", ownerModule);
		closeFilesCommand= MesquiteModule.makeCommand("closeProject", ownerModule);
		saveCommand= MesquiteModule.makeCommand("saveFile", ownerModule);
		saveFilesCommand= MesquiteModule.makeCommand("saveFiles", ownerModule);
		saveAsCommand= MesquiteModule.makeCommand("saveFileAs", ownerModule);
		exportCommand= MesquiteModule.makeCommand("export", ownerModule);
		getInfoCommand= MesquiteModule.makeCommand("getInfo", ownerModule);
		saveLinkagesCommand= MesquiteModule.makeCommand("saveLinkagesAs", ownerModule);
 		linkFileCommand = MesquiteModule.makeCommand("linkFile", ownerModule);
 		newLinkFileCommand = MesquiteModule.makeCommand("newLinkedFile", ownerModule);
 		linkURLCommand = MesquiteModule.makeCommand("linkURL", ownerModule);
 		includeFileCommand = MesquiteModule.makeCommand("includeFile", ownerModule);
 		includeURLCommand = MesquiteModule.makeCommand("includeURL", ownerModule);
	}
	/*.................................................................................................................*/
	/** instantiations of projects are numbered sequentially so they can be referred to by number*/
	public long getID(){
		return idNumber;
	}
	/*.................................................................................................................*/
	/** Disposes of the components of the project. */
	public void dispose(){  //TODO: should all the dispose methods be renamed to something  else???
		isDoomed = true;
		for (int i=0; i< files.size(); i++) {
			((MesquiteFile)files.elementAt(i)).projectClosing = true;
			((MesquiteFile)files.elementAt(i)).close();
		}
		windowToActivate = null;
		activeWindowOfProject = null;
		files.dispose();
		taxas.dispose();
		datas.dispose();
		charModels.dispose();
		otherElements.dispose();
		nexusBlocks.dispose();
		files = null;
		taxas = null;
		datas = null;
		charModels = null;
		otherElements = null;
		nexusBlocks = null;
		
		ownerModule = null; // to minimize chance of memory leaks
		closeCommand= null;
		closeFilesCommand= null;
		saveCommand= null;
		saveFilesCommand= null;
		saveAsCommand= null;
		exportCommand= null;
		getInfoCommand= null;
		saveLinkagesCommand= null;
 		linkFileCommand =null;
 		newLinkFileCommand = null;
 		linkURLCommand = null;
 		includeFileCommand = null;
 		includeURLCommand = null;
		totalDisposed++;
		
	}
	public void finalize() throws Throwable {
		totalFinalized++;
		super.finalize();
	}
	/*.................................................................................................................*/
	/** returns whether any files are dirty */
	public boolean isDirty(){  
		Enumeration enum=files.elements();
		MesquiteFile fi;
		while (enum.hasMoreElements()){
			Object obj = enum.nextElement();
			if (obj instanceof MesquiteFile) {
				fi = (MesquiteFile)obj;
				if (fi.isDirty())
					return true;
			}
		}
		return false;
	}
	/*.................................................................................................................*/
	/** Here as a placeholder; doesn't do anything at the moment. */
    	 public Object doCommand(String commandName, String arguments, CommandRecord commandRec, CommandChecker checker) {
		if (checker.compare(this.getClass(), "Returns coordinator module of project", null, commandName, "getCoordinatorModule")) {
			return ownerModule;
    	 	}
		return null;
   	 }
	/*.................................................................................................................*/
	/** Returns the background color scheme for the project's windows.  This is 0, unless multiple projects are open
	simultaneously, in which case other colors are used as well (to distinguish windows belonging to different projects). */
	public int getProjectColor(){
		return projectColor;
	}
	/* ---------------------------- DIRECTORIES/NAMES ------------------------------*/

	/** Returns name of directory of home file. */
	public String getHomeDirectoryName() {
		if (getHomeFile()==null)
			return null;
		return getHomeFile().getDirectoryName();
	}
	/*.................................................................................................................*/
	/** Returns name of home file. */
	public String getHomeFileName() {
		if (getHomeFile()==null)
			return null;
		return getHomeFile().getFileName();
	}
	/*.................................................................................................................*/
	/** Returns the name of the project.  Currently (temporarily?), the name is based upon the name of the home file. */
	public String getName() {  //TODO: should have separate assignable name
		return "Project with home file \"" + getHomeFileName() + "\"";
	}
	/*.................................................................................................................*/
	/** DOCUMENT */
	public URL getHomeURL() {
		if (getHomeFile()==null)
			return null;
		return getHomeFile().getURL();
	}
	/* ---------------------------- FILES/DIRECTORES ------------------------------*/
	/*.................................................................................................................*/
	/** DOCUMENT */
	public void setHomeFile(MesquiteFile file){
		homeFile = file;
	}
	/*.................................................................................................................*/
	/** DOCUMENT */
	public MesquiteFile getHomeFile(){
		return homeFile;
	}
	/*.................................................................................................................*/
	/** add a file to list of currently linked files */
	public void addFile(MesquiteFile file){
		files.addElement(file, true);
		MesquiteTrunk.mesquiteTrunk.refreshBrowser(MesquiteFile.class);
	}
	/*.................................................................................................................*/
	/** remove a file from list of currently linked files */
	public void removeFile(MesquiteFile file){
		removeNexusBlocks(file);
		files.removeElement(file, true);
		MesquiteTrunk.mesquiteTrunk.refreshBrowser(MesquiteFile.class);
	}
	/*.................................................................................................................*/
	/** Puts all file elements belonging to passed file into home file, and gets rid of record of file. */
	public void mergeFile(MesquiteFile file){
		if (file==homeFile)
			return;
		if (files.indexOf(file)<0)
			return;
		//first, all file elements move to other home file
		Enumeration enum = datas.elements();
		while (enum.hasMoreElements()){
			FileElement obj = (FileElement)enum.nextElement();
			if (obj.getFile() == file) {
				obj.setFile(homeFile);
				homeFile.addFileElement(obj);
			}
		}
		enum = taxas.elements();
		while (enum.hasMoreElements()){
			FileElement obj = (FileElement)enum.nextElement();
			if (obj.getFile() == file) {
				obj.setFile(homeFile);
				homeFile.addFileElement(obj);
			}
		}
		enum = charModels.elements();
		while (enum.hasMoreElements()){
			FileElement obj = (FileElement)enum.nextElement();
			if (obj.getFile() == file) {
				obj.setFile(homeFile);
				homeFile.addFileElement(obj);
			}
		}
		enum = otherElements.elements();
		while (enum.hasMoreElements()){
			FileElement obj = (FileElement)enum.nextElement();
			if (obj.getFile() == file) {
				obj.setFile(homeFile);
				homeFile.addFileElement(obj);
			}
		}
		removeFile(file);
	}
	/*.................................................................................................................*/
	/** returns number of currently linked files */
	public int getNumberLinkedFiles(){
		if (files == null)
			return 0;
		return files.size();
	}
	/*.................................................................................................................*/
	/** returns the index of a currently linked file */
	public int getFileNumber(MesquiteFile f){
		if (f == null || files == null)
			return -1;
		return files.indexOf(f);
	}
	/*.................................................................................................................*/
	/** returns a currently linked file */
	public MesquiteFile getFile(String name){
		return (MesquiteFile)files.getElement(name);
	}
	/*.................................................................................................................*/
	/** returns a currently linked file */
	public MesquiteFile getFile(int num){
		if (num<0 || files == null || num>= files.size())
			return null;
		return (MesquiteFile)files.elementAt(num);
	}
	/*.................................................................................................................*/
	/** returns a currently linked file */
	public MesquiteFile getFileByID(int id){  //TODO: FIND BY id number!!!
		Enumeration enum=files.elements();
		MesquiteFile fi;
		while (enum.hasMoreElements()){
			Object obj = enum.nextElement();
			if (obj instanceof MesquiteFile) {
				fi = (MesquiteFile)obj;
				if (fi.getID()==id) {
					return fi;
				}
			}
		}
		return null;
	}
	/*.................................................................................................................*/
	/** returns currently linked files */
	public ListableVector getFiles(){
		return files;
	}

	/*.................................................................................................................*/
	/** Select one of the currently linked files; selects first if only one or if scripting. 12 Jan 02 */
  	public MesquiteFile chooseFile(CommandRecord commandRec, String message){
			if (getNumberLinkedFiles()==1 || commandRec.scripting())
				return getHomeFile();
			else 
    	 		return (MesquiteFile)ListDialog.queryList(MesquiteModule.mesquiteTrunk.containerOfModule(), "Select file", message, MesquiteString.helpString, files, 0);
  	}
	/* ---------------------------- COMMANDS ------------------------------*/
	/*.................................................................................................................*/
	/** returns command to close a file*/
	public MesquiteCommand getCloseCommand(){
		return closeCommand;
	}
	/*.................................................................................................................*/
	/** returns command to close project (linked files)*/
	public MesquiteCommand getCloseFilesCommand(){
		return closeFilesCommand;
	}
	/*.................................................................................................................*/
	/** returns command to save a file*/
	public MesquiteCommand getSaveCommand(){
		return saveCommand;
	}
	/*.................................................................................................................*/
	/** returns command to save project (linked files)*/
	public MesquiteCommand getSaveFilesCommand(){
		return saveFilesCommand;
	}
	/*.................................................................................................................*/
	/** returns command to link file*/
	public MesquiteCommand getLinkFileCommand(){
		return linkFileCommand;
	}
	/*.................................................................................................................*/
	/** returns command to make new linked file*/
	public MesquiteCommand getNewLinkFileCommand(){
		return newLinkFileCommand;
	}
	
	/*.................................................................................................................*/
	/** returns command to link file*/
	public MesquiteCommand getLinkURLCommand(){
		return linkURLCommand;
	}
	/*.................................................................................................................*/
	/** returns command to incorporate file*/
	public MesquiteCommand getIncludeFileCommand(){
		return includeFileCommand;
	}
	/*.................................................................................................................*/
	/** returns command to incorporate file*/
	public MesquiteCommand getIncludeURLCommand(){
		return includeURLCommand;
	}
	/*.................................................................................................................*/
	/** returns command to save as*/
	public MesquiteCommand getSaveAsCommand(){
		return saveAsCommand;
	}
	/*.................................................................................................................*/
	/** returns command to save as project (linked files)*/
	public MesquiteCommand getSaveLinkagesCommand(){
		return saveLinkagesCommand;
	}
	/*.................................................................................................................*/
	/** returns the file coordinator module in charge of the project*/
	public FileCoordinator getCoordinatorModule(){
		return ownerModule;
	}
	/* ---------------------------- FILE ELEMENTS ------------------------------*/
 	/*.................................................................................................................*/
	/** Adds the passed element to the project.  */
  	public void addFileElement(FileElement element) {
  		if (element==null)
  			return;
  			
  		if (element instanceof Taxa) {
  			taxas.addElement(element, true);
  			element.addListener(taxas);
  		}
  		else if (element instanceof CharacterData){
	  		datas.addElement(element, true);
  			element.addListener(datas);
 		}
 		else if (element instanceof CharacterModel) {
 			charModels.addElement(element, true);
 			modelListener.addModel((CharacterModel)element);
 		}
 		else 
 			otherElements.addElement(element, true);
  			
  		broadcastAddFileElement(ownerModule, element);
  	}
	/*.................................................................................................................*/
	/** DOCUMENT */
  	public void removeFileElement(FileElement element) {
  		if (element==null)
  			return;
		if (element instanceof Taxa) {
			if (taxas != null) {
  				taxas.removeElement(element, true);
  				element.removeListener(taxas);
  			}
  			//taxas.notifyListenersOfDisposed(element);
		}
		else if (element instanceof CharacterData) {
			if (datas != null) {
  				datas.removeElement(element, true);
  				element.removeListener(datas);
  			}
  			//datas.notifyListenersOfDisposed(element);
  		}
		else if (element instanceof CharacterModel) {
			if (charModels != null)
 				charModels.removeElement(element, true);
  			//charModels.notifyListenersOfDisposed(element);
			if (modelListener != null)
 				modelListener.removeModel((CharacterModel)element);
 		}
 		else {
			if (otherElements != null)
 				otherElements.removeElement(element, true);
  			//otherElements.notifyListenersOfDisposed(element);
 		}
		//TODO: shouldn't broadcase of deletion be here?
	}
 	/*.................................................................................................................*/
	/** Returns the number of FileElements that are subclasses of the given class.  */
  	public int getNumberOfFileElements(Class c) {
  		if (c ==null)
  			return 0;
  			
  		if (Taxa.class.isAssignableFrom(c)) {
  			return taxas.size();
  		}
  		else if (CharacterData.class.isAssignableFrom(c))
	  		return datas.size();
 		else if (CharacterModel.class.isAssignableFrom(c)) {
 			return charModels.size();
 		}
 		else {
 			return otherElements.size(c);
 		}
  	}
 	/*.................................................................................................................*/
	/** Sets elements of file as not dirty (i.e. saved).  */
  	public void fileSaved(MesquiteFile f) {
  		f.setDirtiedByCommand(false);
  		for (int i=0; i<taxas.size(); i++) {
  			FileElement fe = (FileElement)taxas.elementAt(i);
  			if (fe.getFile() == f)
  				fe.setDirty(false);
  		}
  		for (int i=0; i<datas.size(); i++) {
  			FileElement fe = (FileElement)datas.elementAt(i);
  			if (fe.getFile() == f)
  				fe.setDirty(false);
  		}
  		for (int i=0; i<charModels.size(); i++) {
  			FileElement fe = (FileElement)charModels.elementAt(i);
  			if (fe.getFile() == f)
  				fe.setDirty(false);
  		}
  		for (int i=0; i<otherElements.size(); i++) {
  			FileElement fe = (FileElement)otherElements.elementAt(i);
  			if (fe.getFile() == f)
  				fe.setDirty(false);
  		}
 		
  	}
 	/*.................................................................................................................*/
	/** Returns a vector of FileElements that are of the given class or are of its subclasses.  */
  	public Listable[] getFileElements(Class c) {
  		if (c ==null)
  			return null;
  			
  		if (Taxa.class.isAssignableFrom(c)) {
  			return taxas.getElementArray();
  		}
  		else if (CharacterData.class == c)
	  		return datas.getElementArray();
  		else if (CharacterData.class.isAssignableFrom(c)) {
 			int count =0;
 			for (int i=0; i<datas.size(); i++)
 				if (c.isAssignableFrom(datas.elementAt(i).getClass()))
 					count++;
 			Listable[] v = new Listable[count];
 			count =0;
 			for (int i=0; i<datas.size(); i++)
 				if (c.isAssignableFrom(datas.elementAt(i).getClass())) {
 					v[count] = (Listable)datas.elementAt(i);
 					count++;
 				}
 			return v;
 		}
 		else if (CharacterModel.class == c)
 			return charModels.getElementArray();
 		else if (CharacterModel.class.isAssignableFrom(c)) {
 			int count =0;
 			for (int i=0; i<charModels.size(); i++)
 				if (c.isAssignableFrom(charModels.elementAt(i).getClass()))
 					count++;
 			Listable[] v = new Listable[count];
 			count =0;
 			for (int i=0; i<charModels.size(); i++)
 				if (c.isAssignableFrom(charModels.elementAt(i).getClass())) {
 					v[count] = (Listable)charModels.elementAt(i);
 					count++;
 				}
 			return v;
 		}
 		else {
 			int count =0;
 			for (int i=0; i<otherElements.size(); i++)
 				if (c.isAssignableFrom(otherElements.elementAt(i).getClass())) {
 					count++;
 				}
 			Listable[] v = new Listable[count];
 			count =0;
 			for (int i=0; i<otherElements.size(); i++)
 				if (c.isAssignableFrom(otherElements.elementAt(i).getClass())) {
 					v[count] = (Listable)otherElements.elementAt(i);
 					count++;
 				}
 			return v;
 		}
 		
  	}
 	/*.................................................................................................................*/
	/** Returns a vector of FileElements that are of the given class or are of its subclasses, and which are compatible with the passed object.  */
  	public Listable[] getCompatibleFileElements(Class c, Object condition) {
  		if (c ==null)
  			return null;
  		Listable[] elementArray = getFileElements(c); //first get all elements, then filter for compatibility
  		if (elementArray == null)
  			return null;
  		if (condition == null)
  			return elementArray;
		int count =0;
		for (int i=0; i<elementArray.length; i++) {
			Object element = elementArray[i];
			if (element instanceof CompatibilityChecker){
				if (((CompatibilityChecker)element).isCompatible(condition, this, null)){
					count++;
				}
			}
			else
				count++;
		}
		Listable[] v = new Listable[count];
		count =0;
		for (int i=0; i<elementArray.length; i++) {
			Object element = elementArray[i];
			if (element instanceof CompatibilityChecker){
				if (((CompatibilityChecker)element).isCompatible(condition, this, null)){
					v[count] = (Listable)element;
					count++;
				}
			}
			else {
				v[count] = (Listable)element;
				count++;
			}
		}
		return v;
  	}
 	/*.................................................................................................................*/
	/** Returns the ith FileElement that is a subclass of the given class.  */
  	public FileElement getFileElement(Class c, int i) {
  		if (c ==null || i<0)
  			return null;
  			
  		if (Taxa.class.isAssignableFrom(c)) {
  			if (i< taxas.size())
  				return (FileElement)taxas.elementAt(i);
  		}
  		else if (CharacterData.class.isAssignableFrom(c)) {
  			if (i< datas.size())
	  			return (FileElement)datas.elementAt(i);
	  	}
 		else if (CharacterModel.class.isAssignableFrom(c)) {
  			if (i< charModels.size())
 				return (FileElement)charModels.elementAt(i);
 		}
 		else {
			int count =0;
			Enumeration e = otherElements.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				if (c.isAssignableFrom(obj.getClass())) {
					if (count == i)
						return (FileElement)obj;
					count ++;
				}
			}
 		}
 		return null;
  	}
 	/*.................................................................................................................*/
	/** Returns the index position of the  FileElement among others of its the subclass c.  */
  	public int getFileElementNumber(FileElement f, Class c) {
  		if (f ==null || c==null)
  			return -1;
  			
  		if (Taxa.class.isAssignableFrom(c)) {
  			return getTaxaNumber((Taxa)f);
  		}
  		else if (CharacterData.class.isAssignableFrom(c)) {
  			return getMatrixNumber((CharacterData)f);
	  	}
 		else if (CharacterModel.class.isAssignableFrom(c)) {
			int count =0;
			Enumeration e = charModels.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				if (c.isAssignableFrom(obj.getClass())) {
					if (obj == f)
						return count;
					count ++;
				}
			}
 		}
 		else {
			int count =0;
			Enumeration e = otherElements.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				if (c.isAssignableFrom(obj.getClass())) {
					if (obj == f)
						return count;
					count ++;
				}
			}
 		}
 		return -1;
  	}
	/*.................................................................................................................*/
 	/* A method called to broadcast to all descendent modules that file element has been added.*/
 	private void broadcastAddFileElement(MesquiteModule module, FileElement element) {
		module.fileElementAdded(element);
		EmployeeVector e = module.getEmployeeVector();
		for (int i = 0; i<e.size(); i++) {
			MesquiteModule employee = (MesquiteModule)e.elementAt(i);
			broadcastAddFileElement(employee, element);
		}
 	}
	/*.................................................................................................................*/
	/** Adds the passed NEXUS block to the project. */
	public void addNexusBlock(NexusBlock block) {
		nexusBlocks.addElement(block, true);
	}
	/*.................................................................................................................*/
	/** Removes the passed NEXUS block  from the project. */
	public void removeNexusBlock(NexusBlock block) {
		if (nexusBlocks != null) {
			if (block !=null) {
				if (files.indexOf(block.getFile())<0)
					MesquiteMessage.warnProgrammer("Error: NEXUS block found referring to file that no longer exists " + block.getClass());
				nexusBlocks.removeElement(block, true);
			}
		}
	}
	/*.................................................................................................................*/
	/** Returns those NEXUS blocks belonging to the given file. */
	public void removeNexusBlocks(MesquiteFile f) {
		int count = 0;
		for (int i = 0; i<nexusBlocks.size(); i++) {
			NexusBlock nb = (NexusBlock)nexusBlocks.elementAt(i);
			if (nb.getFile()==f)
				count++;
		}
		if (count==0)
			return;
		NexusBlock[] nbs = new NexusBlock[count];
		count=0;
		for (int i = 0; i<nexusBlocks.size(); i++) {
			NexusBlock nb = (NexusBlock)nexusBlocks.elementAt(i);
			if (nb.getFile()==f) {
				nbs[count]= nb;
				count++;
			}
		}
		
		for (int i = 0; i<count; i++) {
			removeNexusBlock(nbs[i]);
		}
	}
	/*.................................................................................................................*/
	/** Returns an array of NEXUS blocks of the given type and belonging to the given file. */
	public NexusBlock[] getNexusBlocks(Class c, MesquiteFile f) {
		int count = 0;
		for (int i = 0; i<nexusBlocks.size(); i++) {
			NexusBlock nb = (NexusBlock)nexusBlocks.elementAt(i);
			if (c.isAssignableFrom(nb.getClass()) && (nb.getFile()==f || f == null))
				count++;
		}
		if (count==0)
			return null;
		NexusBlock[] nbs = new NexusBlock[count];
		count=0;
		for (int i = 0; i<nexusBlocks.size(); i++) {
			NexusBlock nb = (NexusBlock)nexusBlocks.elementAt(i);
			if (c.isAssignableFrom(nb.getClass()) &&  (nb.getFile()==f || f == null)) {
				nbs[count]= nb;
				count++;
			}
		}
		return nbs;
	}
	/*.................................................................................................................*/
	/** Returns an array of NEXUS blocks of the given type name. 12 Jan 02*/
	public NexusBlock[] getNexusBlocks(String typeName) {
		if (StringUtil.blank(typeName))
			return null;
		int count = 0;
		for (int i = 0; i<nexusBlocks.size(); i++) {
			NexusBlock nb = (NexusBlock)nexusBlocks.elementAt(i);
			if (typeName.equalsIgnoreCase(nb.getBlockName()))
				count++;
		}
		if (count==0)
			return null;
		NexusBlock[] nbs = new NexusBlock[count];
		count=0;
		for (int i = 0; i<nexusBlocks.size(); i++) {
			NexusBlock nb = (NexusBlock)nexusBlocks.elementAt(i);
			if (typeName.equalsIgnoreCase(nb.getBlockName())) {
				nbs[count]= nb;
				count++;
			}
		}
		return nbs;
	}
	/*.................................................................................................................*/
	/** Returns the ListableVector of NEXUS blocks in the project. */
	public ListableVector getNexusBlocks() {
		return nexusBlocks;
	}
	/* ---------------------------- TAXA ------------------------------*/
 	/*.................................................................................................................*/
	/** returns the list of Taxa  */
  	public ListableVector getTaxas() {
  		return taxas;
  	}
 	/*.................................................................................................................*/
	/** returns the Taxa of given number */
  	public Taxa getTaxa(int taxaNumber) {
  		if (taxaNumber>=0 && taxaNumber < taxas.size())
  			return (Taxa)taxas.elementAt(taxaNumber);
  		MesquiteMessage.warnProgrammer("Error: attempt to get taxa set " + taxaNumber + "; maximum  " + taxas.size());
  		MesquiteMessage.printStackTrace();
  		return null;
  	}
  	/*.................................................................................................................*/
	/** returns the Taxa of given name */
 	public Taxa getTaxa(String name) {
  		return getTaxa(null, name);
  	}
  	/*.................................................................................................................*/
	/** returns the Taxa of given reference, name or number (as indicated by string passed) */
 	public Taxa getTaxa(MesquiteFile f, String ref) {
  		if (ref==null)
  			return null;
 		if (ref.length()>0 && ref.charAt(0)=='#'){ //using id number reference
 			String id = ref.substring(1, ref.length());
 			long idN = MesquiteLong.fromString(id);
 			if (MesquiteLong.isCombinable(idN))
		  		for (int i=0; i<taxas.size(); i++) {
		  			Taxa taxa = (Taxa)taxas.elementAt(i);
			  		if (taxa.assignedIDNumber == idN)
			  			return taxa;
		  		}
 		}
 		else { 
	  		for (int i=0; i<taxas.size(); i++) { //check to see if name of taxa is being passed
	  			Taxa t = (Taxa)taxas.elementAt(i);
		  		if (t.getName()!=null && t.getName().equalsIgnoreCase(ref) && (f==null || t.getFile() == f))
		  			return (Taxa)taxas.elementAt(i);
	  		}
	  		
	  		return getTaxa(f, MesquiteInteger.fromString(ref)); //try to see if number
	  		
  		}
  		return null;
  	}
  	/*.................................................................................................................*/
	/** returns the Taxa of given number belonging to given file */
 	public Taxa getTaxa(MesquiteFile f, int taxaNumber) {
  		int count=0;
  		for (int i=0; i<taxas.size(); i++) {
  			Taxa taxa = ((Taxa)taxas.elementAt(i));
	  		if (f== null || taxa.getFile()==f) {
	  			if (count==taxaNumber)
	  				return taxa;
	  			count++;
	  		}
  		}
  		return null;
  	}
  	/*.................................................................................................................*/
	/** returns the number (index position) of the Taxa of given name */
 	public int getTaxaNumber(String name) {
  		for (int i=0; i<taxas.size(); i++) {
	  		if (((Taxa)taxas.elementAt(i)).getName().equalsIgnoreCase(name))
	  			return i;
  		}
  		return -1;
  	}
  	/*.................................................................................................................*/
	/** returns the number (index position) of the Taxa */
 	public int getTaxaNumber(MesquiteFile file, Taxa taxa) {
 		if (file == null)
 			return getTaxaNumber(taxa);
 		if (taxa==null)
 			return -1;
 		int count =0;
  		for (int i=0; i<taxas.size(); i++) {
  			Taxa t = ((Taxa)taxas.elementAt(i));
	  		if (t== taxa) 
	  			return count;
  			if (t.getFile()==file)
  				count++;
  		}
  		return -1;
  	}
  	/*.................................................................................................................*/
	/** returns the number (index position) of the Taxa */
 	public int getTaxaNumber(Taxa taxa) {
 		if (taxa==null)
 			return -1;
 		if (taxa.isDuplicate()){ //This subverts the taxa if it's a duplicate, instead giving number of original copy
	  		for (int i=0; i<taxas.size(); i++) {
	  			Taxa t =(Taxa)taxas.elementAt(i);
		  		if (taxa.equals(t, false) && !t.isDuplicate()) {
		  			return i;
		  		}
	  		}
 		}
  		for (int i=0; i<taxas.size(); i++) {
	  		if (((Taxa)taxas.elementAt(i)) == taxa) {
	  			return i;
	  		}
  		}
  		return -1;
  	}
  	/*.................................................................................................................*/
	/** returns a string that can be used to refer to  taxa block */
 	public String getTaxaReference(Taxa taxa) {
 		if (taxa == null)
 			return null;
 		return "#" + taxa.assignedIDNumber;
 		/*
  		for (int i=0; i<taxas.size(); i++) {
	  		if (((Taxa)taxas.elementAt(i)).getName().equalsIgnoreCase(name))
	  			return i;
  		}
  		return -1;
  		*/
  	}
 	/*.................................................................................................................*/
	public Taxa chooseTaxa(Frame frame, String message, CommandRecord commandRec){
		int num = getNumberTaxas();
		if (num==0)
			return null;
		else if (num==1 || (commandRec.scripting() && num>=1))
			return getTaxa(0);
 		String[] taxaNames = new String[num];
 		for (int i=0; i<num;i++) 
 			taxaNames[i] = getTaxa(i).getName();
		int whichTaxa = ListDialog.queryList(frame, "Select taxa", message, MesquiteString.helpString, taxaNames, 0);
		if (!MesquiteInteger.isCombinable(whichTaxa)) 
			return null;
		else if (whichTaxa>=0) 
			return getTaxa(whichTaxa);
		else
			return getTaxa(0);
	}
 	/*.................................................................................................................*/
 	/*.................................................................................................................*/
 	//TODO:  have general get number of elements, passing project, or file; counts
	/** returns the number of Taxa objects stored in project belonging to a file*/
  	public int  getNumberTaxas(MesquiteFile f) {
  		int count = 0;
  		for (int i=0; i<taxas.size(); i++) {
	  		if (f==null || ((Taxa)taxas.elementAt(i)).getFile()==f)
	  			count++;
  		}
  		return count;
  	}
 	/*.................................................................................................................*/
	/** returns the number of Taxa objects stored in project */
  	public int  getNumberTaxas() {
  		if (taxas == null)
  			return 0;
  		return taxas.size();
  	}
	/* ---------------------------- CHARACTERDATA ------------------------------*/
 	/*.................................................................................................................*/
	/** returns jth data set in file*/
  	public ListableVector getCharacterMatrices() {
  		return datas;
  	}
 	/*.................................................................................................................*/
  	private boolean compatibleMatrix(Object dataClass, CharacterData data){
  		if (dataClass ==null)
  			return true;
  		if (dataClass instanceof Class){
  			return ((Class)dataClass).isAssignableFrom(data.getStateClass());
  		}
  		else if (dataClass instanceof Class[]){
  			Class[] dataClasses = ((Class[])dataClass);
  			for (int i=0; i<dataClasses.length; i++){
  				if (dataClasses[i].isAssignableFrom(data.getStateClass()))
  					return true;
  			}
	  		return false;
  		}
  		return false;
  	}
  	/*.................................................................................................................*/
	/** returns jth data set in file*/
  	public CharacterData getCharacterMatrix(int j) {
  		int count=0;
  		CharacterData data;
  		for (int i=0; i<datas.size(); i++) {
  			data = (CharacterData)datas.elementAt(i);
	  		if (!data.isDoomed()) {
	  			if (count==j)
	  				return data;
	  			count++;
	  		}
  		}
  		MesquiteMessage.printStackTrace("Error: attempt to get data set beyond number (0) [" + j + "] {" + datas.size() + " }"  + getName());
  		return null;
  	}
  	/*.................................................................................................................*/
	/** returns the jth data set belonging to given taxa.*/
  	public CharacterData getCharacterMatrix(Taxa taxa, int j) {
  		return getCharacterMatrix(null, taxa, null, j);
  	}
	/*.................................................................................................................*/
	/** returns the jth data set for given CharacterState subclass.*/
  	public CharacterData getCharacterMatrix(int j, Object dataClass) {
  		return getCharacterMatrix(null, null, dataClass, j);
  	}
 	/*.................................................................................................................*/
	/** returns the jth data set belonging to given taxa for given CharacterState subclass.*/
  	public CharacterData getCharacterMatrix(Taxa taxa, int j, Object dataClass) {
  		return getCharacterMatrix(null, taxa, dataClass, j);
  	}
 	/*.................................................................................................................*/
	/** returns the jth of data sets belonging to a given file*/
  	public CharacterData getCharacterMatrix(MesquiteFile f, int j) {
  		return getCharacterMatrix(f, null, null, j);
  	}
 	/*.................................................................................................................*/
	/** returns the jth of data sets belonging to a given file*/
  	public CharacterData getCharacterMatrix(MesquiteFile f, Taxa taxa, Object dataClass, int j) {
  		int count=0;
  		for (int i=0; i<datas.size(); i++) { 
  			CharacterData data = (CharacterData)datas.elementAt(i);

	  		if (!data.isDoomed() && (f== null || data.getFile()==f) && (taxa == null || taxa.equals(data.getTaxa(), true)) && (dataClass == null || compatibleMatrix(dataClass, data))) {
//	  		if (!data.isDoomed() && (f== null || data.getFile()==f) && (taxa == null || taxa == data.getTaxa()) && (dataClass == null || compatibleMatrix(dataClass, data))) {
	  			if (count==j) {
	  				return data;
	  			}
	  			count++;
	  		}
  		}
  		return null;
  	}
  	/*.................................................................................................................*/
	/** gets the data set of given reference string (ref. number, name, number) from the project */
 	public CharacterData getCharacterData(String name) {
  		return getCharacterData((MesquiteFile)null, (Taxa)null, null, name);
  	}
  	/*.................................................................................................................*/
	/** gets the data set of given reference string (ref. number, name, number) from the project */
 	public CharacterData getCharacterData(MesquiteFile f, String ref) {
 		return getCharacterData(f, null, null, ref);
  	}
  	/*.................................................................................................................*/
	/** gets the data set of given reference string (ref. number, name, number) from the project */
 	public CharacterData getCharacterData(MesquiteFile f, Taxa taxa, String ref) {
 		return getCharacterData(f, taxa, null, ref);
  	}
  	/*.................................................................................................................*/
	/** gets the data set of given reference string (ref. number, name, number) from the project */
 	public CharacterData getCharacterData(MesquiteFile f, Taxa taxa, Object dataClass, String ref) {
 		if (ref !=null && ref.length()>0 && ref.charAt(0)=='#'){ //using id number reference
 			String id = ref.substring(1, ref.length());
 			long idN = MesquiteLong.fromString(id);
 			if (MesquiteLong.isCombinable(idN))
		  		for (int i=0; i<datas.size(); i++) {
		  			CharacterData data = (CharacterData)datas.elementAt(i);
			  		if (!data.isDoomed() && data.assignedIDNumber == idN && (taxa==null || data.getTaxa() == taxa) && (dataClass == null || compatibleMatrix(dataClass, data))) {
			  			return data;
			  		}
		  		}
 		}
 		else { //using name or number
	  		for (int i=0; i<datas.size(); i++) {
	  			CharacterData data = (CharacterData)datas.elementAt(i);
	 
		  		if (!data.isDoomed() && data.getName().equalsIgnoreCase(ref) && (f==null || data.getFile() == f) && (taxa==null || data.getTaxa() == taxa)&& (dataClass == null || compatibleMatrix(dataClass, data)))
		  			return data;
	  		}
  			return getCharacterMatrix(f, taxa, dataClass, MesquiteInteger.fromString(ref)); //try to see if number
 		}
  		return null;  //CHECK FOR COMPATIBLE
  	}
 	/*.................................................................................................................*/
	/** returns a data matrix that is of the same class as that passed, but is a different one.*/
  	public CharacterData getDifferentButCompatibleMatrix(CharacterData data, CommandRecord commandRec, String messageIfNoneFound, String messageForChoice) {
  		if (data == null)
  			return null;
  		else {
			Taxa taxa = data.getTaxa();
			//taxa.prepareEqualTaxaMessage();
			int numDatasTotal = getNumberCharMatrices(taxa);
			//String equalTaxaMessage = taxa.getEqualTaxaMessage();
			//taxa.turnOffEqualTaxaMessage();
			
			String s ="";
			boolean allDataIsCurrentData = true;
			boolean allDataIsWrongClass = true;
			boolean someDataAreLinked = false;
			if (numDatasTotal==0)
				s += " There are no matrices for these taxa.";
			int numDiffComparable = numDatasTotal;
			
			//this seems to retain remnants of a more relaxed system that would allow matrices connected to other but equivalent Taxa
			for (int i = 0; i<numDatasTotal; i++) {
				CharacterData pData =getCharacterMatrix(taxa, i);
				if (pData != null) {
					if (pData== data) //same matrix; don't count
						numDiffComparable--;
					else {
						allDataIsCurrentData = false;   // at least one data matrix with matching taxa is not the current data matrix
						if (pData.getClass() != data.getClass()) //different type of data; don't count
							numDiffComparable--;
						else { 
							allDataIsWrongClass = false;   // at least one novel data matrix with matching taxa is of the correct dataClass
							if (pData.isLinked(data)){ //is linked; don't count
								numDiffComparable--;
								someDataAreLinked = true; 
							}
						}
					}
				}
			}
			if (numDiffComparable<=0) {
				if (!commandRec.scripting() && messageIfNoneFound!=null) {
					if (numDatasTotal>=1) {
						if (allDataIsCurrentData) {
							if (numDatasTotal==1) {
								s += " There are no other matrices for these taxa. "; //+ equalTaxaMessage;
							}
							else
								s += " All matrices are the same as the current one."; //one shouldn't get here!
						}
						else if (allDataIsWrongClass)
							s += " All other matrices are of a different class of data.";
						else if (someDataAreLinked)
							s += " The only comparable matrices are linked.";
					}
					MesquiteTrunk.mesquiteTrunk.alert(messageIfNoneFound + "\n\nProblem: "+s);
				}
				return null;
			}
			else {
				Listable[] matrices = new Listable[numDiffComparable];
				int count=0;
				for (int i = 0; i<numDatasTotal; i++) {
					CharacterData pData =getCharacterMatrix(taxa, i);
					if (pData != null && pData!= data && (pData.getClass() == data.getClass()) && !pData.isLinked(data)) {
						matrices[count]=pData;
						count++;
					}
				}
				boolean differenceFound=false;
				CharacterData oData = (CharacterData)ListDialog.queryList(MesquiteTrunk.mesquiteTrunk.containerOfModule(), "Choose matrix", messageForChoice, MesquiteString.helpString, matrices, 0);
				return oData;
			}
  		}
  	}
  	/*.................................................................................................................*/
	/** gets the number (index position) of the data set of given name. */
 	public int getMatrixNumber(String name) {
  		for (int i=0; i<datas.size(); i++) {
  			CharacterData data = (CharacterData)datas.elementAt(i);
	  		if (!data.isDoomed() && data.getName().equalsIgnoreCase(name))
	  			return i;
  		}
  		return -1;
  	}
  	/*.................................................................................................................*/
	/** gets the number (index position) of the data set. */
 	public int getMatrixNumber(CharacterData data) {
  		for (int i=0; i<datas.size(); i++) {
	  		if ((CharacterData)datas.elementAt(i) == data)
	  			return i;
  		}
  		return -1;
  	}
  	/*.................................................................................................................*/
	/** gets the number (index position) of the data set. */
 	public int getMatrixNumber(MesquiteFile file, CharacterData data) {
  		if (file == null)
  			return getMatrixNumber(data);
  		int count = 0;
  		for (int i=0; i<datas.size(); i++) {
	  		CharacterData dataI =(CharacterData)datas.elementAt(i);
	  		if (dataI == data)
	  			return count;
	  		if (dataI.getFile() == file)
	  			count++;
  		}
  		return -1;
  	}
	/*.................................................................................................................*/
	/** returns a string that can be used to refer to  taxa block */
 	public String getCharMatrixReference(CharacterData data) {
 		if (data == null)
 			return null;
 		return "#" + data.assignedIDNumber;
 		/*
  		for (int i=0; i<taxas.size(); i++) {
	  		if (((Taxa)taxas.elementAt(i)).getName().equalsIgnoreCase(name))
	  			return i;
  		}
  		return -1;
  		*/
  	}
 	/*.................................................................................................................*/
	/** returns number of data sets in file*/
  	public int  getNumberCharMatrices() { //TODO: these should be getNumberOfDataSets
  		return getNumberCharMatrices(null, null, null);
  	}
 	/*.................................................................................................................*/
	/** returns number of data sets belonging to given taxa*/
  	public int  getNumberCharMatrices(Taxa taxa) { //TODO: these should be getNumberOfDataSets
  		return getNumberCharMatrices(null, taxa, null);
  	}
 	/*.................................................................................................................*/
	/** returns number of data sets of a given data class (CharacterState subclass is passed) belonging to given taxa*/
  	public int  getNumberCharMatrices(Taxa taxa, Object dataClass) { //TODO: these should be getNumberOfDataSets
  		return getNumberCharMatrices(null, taxa, dataClass);
  	}
 	/*.................................................................................................................*/
	/** returns number of data sets of a given data class (CharacterState subclass is passed)*/
  	public int  getNumberCharMatrices(Object dataClass) { //TODO: these should be getNumberOfDataSets
  		return getNumberCharMatrices(null, null, dataClass);
  	}
 	/*.................................................................................................................*/
	/** returns number of data sets belonging to a given file*/
  	public int  getNumberCharMatrices(MesquiteFile f) { //TODO: these should be getNumberOfDataSets
  		return getNumberCharMatrices(f, null, null);
  	}
 	/*.................................................................................................................*/
	/** returns number of data sets of a given data class (CharacterState subclass is passed) belonging to given taxa*/
  	public int  getNumberCharMatrices(MesquiteFile file, Taxa taxa, Object dataClass) { //TODO: these should be getNumberOfDataSets
  		int count=0;
  		for (int i=0; i<datas.size(); i++) { 
  			CharacterData data = (CharacterData)datas.elementAt(i);
	  		if (!data.isDoomed() && (file == null || data.getFile()==file) && (taxa == null || taxa.equals(data.getTaxa(), true)) && (dataClass==null || compatibleMatrix(dataClass, data)))
	  			count++;
  		}
  		return count;
  	}
	
 	/*.................................................................................................................*/
	public CharacterData chooseData(Frame frame, Taxa taxa, Object dataClass, CommandRecord commandRec){
		return chooseData(frame, null, taxa, dataClass, "Data matrix", commandRec);
 	}
 	/*.................................................................................................................*/
	public CharacterData chooseData(Frame frame, Taxa taxa, Object dataClass, String message, CommandRecord commandRec){
		return chooseData(frame, null, taxa, dataClass, message, commandRec);
	}
 	/*.................................................................................................................*/
	public CharacterData chooseData(Frame frame, MesquiteFile file, Taxa taxa, Object dataClass, String message, CommandRecord commandRec){
		int num = getNumberCharMatrices(file, taxa, dataClass);
		if (num==0)
			return null;
		else if (num==1 || (commandRec.scripting() && num>=1))
			return getCharacterMatrix(file, taxa, dataClass, 0);
 		String[] dataNames = new String[num];
 		for (int i=0; i<num;i++) 
 			dataNames[i] = getCharacterMatrix(file, taxa, dataClass, i).getName();
		int whichData = ListDialog.queryList(frame, "Select data matrix", message, MesquiteString.helpString, dataNames, 0);
		if (!MesquiteInteger.isCombinable(whichData)) 
			return null;
		else if (whichData>=0) 
			return getCharacterMatrix(file, taxa, dataClass, whichData);
		else
			return getCharacterMatrix(file, taxa, dataClass, 0);
	}
	/* ---------------------------- CHARACTER MODELS ------------------------------*/
	/*.................................................................................................................*/
	/** adds the character model to those stored in the project 
 	public void addCharacterModel(CharacterModel model) {
 		charModels.addElement(model);
 	}
 	/*.................................................................................................................*/
	/** returns the vector of stored character models */
 	public ModelVector getCharacterModels() {
 		return charModels;
 	}
 	/*.................................................................................................................*/
	/** returns the number of character models stored */
 	public int getNumModels() {
 		return charModels.size();
 	}
 	/*.................................................................................................................*/
	/** returns a vector of stored character models that are subclasses of the given class*/
 	public Listable[] getCharacterModels(Class c, Class stateClass) {
 		if (charModels == null || c==null)
 			return null;
 		int numModels = 0;
 		for (int i=0; i<getNumModels(); i++) {
 			CharacterModel mod = getCharacterModel(i);
 			if (c.isAssignableFrom(mod.getClass()) && (stateClass==null || mod.getStateClass() == stateClass))
 				numModels++;
 		}
 		Listable[] mV = new Listable[numModels];
 		int count = 0;
 		for (int i=0; i<getNumModels(); i++) {
 			CharacterModel mod = getCharacterModel(i);
 			if (c.isAssignableFrom(mod.getClass()) && (stateClass==null || mod.getStateClass() == stateClass))
 				mV[count++] = mod;
 		}
 		return mV;
 	}
 	/*.................................................................................................................*/
	/** returns a vector of stored character models that are compatible with the passed information*/
 	public Listable[] getCharacterModels(ModelCompatibilityInfo mci, Class stateClass) {
 		if (charModels == null || mci==null)
 			return null;
 		int numModels = 0;
 		for (int i=0; i<getNumModels(); i++) {
 			CharacterModel mod = getCharacterModel(i);
 			if (mod.isCompatible(mci, null, null))
 				numModels++;
 		}
 		Listable[] mV = new Listable[numModels];
 		int count = 0;
 		for (int i=0; i<getNumModels(); i++) {
 			CharacterModel mod = getCharacterModel(i);
 			if (mod.isCompatible(mci, null, null))
 				mV[count++] = mod;
 		}
 		return mV;
 	}
 	/*.................................................................................................................*/
	public CentralModelListener getCentralModelListener(){
		return modelListener;
	}
 	/*.................................................................................................................*/
	/** returns the index'th character model. */
 	public CharacterModel getCharacterModel(int index) {
 		return (CharacterModel)charModels.elementAt(index);
 	}
 	/*.................................................................................................................*/
	/** returns the index'th character model that is a subclass of c. */
 	public CharacterModel getCharacterModel(ModelCompatibilityInfo mci, int index) {
 		if (charModels == null || mci==null)
 			return null;
 		int count = 0;
 		for (int i=0; i<getNumModels(); i++) {
 			CharacterModel mod = getCharacterModel(i);
 			if (mod.isCompatible(mci, null, null)) {
 				if (index== count)
	  				return mod;
				count++;
 			}
 		}
 		return null;
 	}
 	/*.................................................................................................................*/
	/** returns the index of the passed character model among compatible models. Added Apr 02*/
 	public int getWhichCharacterModel(ModelCompatibilityInfo mci, CharacterModel model) {
 		if (charModels == null || model == null)
 			return -1;
 		if (mci == null){
	 		int count = 0;
	 		for (int i=0; i<getNumModels(); i++) {
	 			CharacterModel mod = getCharacterModel(i);
 				if (model == mod || model.getMother()== mod)  
	  				return count;
				count++;
	 		}
	 		return -1;
 		}
 		int count = 0;
 		for (int i=0; i<getNumModels(); i++) {
 			CharacterModel mod = getCharacterModel(i);
 			if (mod.isCompatible(mci, null, null)) {
 				if (model == mod || model.getMother()== mod)  
	  				return count;
				count++;
 			}
 		}
 		return -1;
 	}
 	/*.................................................................................................................*/
	/** returns the index'th character model that is a subclass of c. */
 	public CharacterModel getCharacterModel(Class c, boolean allowBuiltIn, int index) {
 		if (charModels == null || c==null)
 			return null;
 		int count = 0;
 		for (int i=0; i<getNumModels(); i++) {
 			CharacterModel mod = getCharacterModel(i);
 			if (c.isAssignableFrom(mod.getClass()) && (allowBuiltIn || !mod.isBuiltIn())) {
 				if (index== count)
	  				return mod;
				count++;
 			}
 		}
 		return null;
 	}
 	/*.................................................................................................................*/
	/** returns the index'th character model that is a subclass of c. */
 	public CharacterModel getCharacterModel(Class c, int index) {
 		return getCharacterModel(c, true, index);
 	}
 	/*.................................................................................................................*/
	/** returns the character model with the given name */
 	public CharacterModel getCharacterModel(String name) {
 		return charModels.getCharacterModel(name);
 	}
	/* ---------------------------- HNODE ------------------------------*/
	/* ---------------- for HNode interface ----------------------*/
	/** for HNode interface */
	public HNode[] getHDaughters(){
		ListableVector files = getFiles();
		if (files == null || files.size()== 0)
			return null;
		HNode[] daughters = new HNode[files.size()];
		for (int i = 0; i < files.size(); i++)
			daughters[i] = (HNode)files.elementAt(i);
		return daughters;
	}
	/* ---------------- for HNode interface ----------------------*/
	/** for HNode interface */
	public HNode getHMother(){
		return null; //TODO: should return a class that contains links to all open projects
	}
	/* ---------------- for HNode interface ----------------------*/
	/** for HNode interface */
	public int getNumSupplements(){
		return 0;
	}
	/* ---------------- for HNode interface ----------------------*/
	/** for HNode interface */
	public String getSupplementName(int index){
		return null;
	}
	/* ---------------- for HNode interface ----------------------*/
	/** for HNode interface */
	public void hNodeAction(Container c, int x, int y, int action){
		if (c==null)
			return;
		if (action == HNode.MOUSEDOWN){
			MesquitePopup popup = new MesquitePopup(c);
			popup.addItem(getName(), ownerModule, null);
			popup.addItem("Bring All Windows To Front", ownerModule, ownerModule.makeCommand("allToFront", ownerModule));
			popup.addItem("Close Project", ownerModule, ownerModule.makeCommand("closeProject", ownerModule));
			popup.showPopup( x,y);
		}
		else if (action == HNode.MOUSEMOVE){
			Frame f = MesquiteWindow.frameOfComponent(c);
			if (f != null && f instanceof MesquiteWindow){
				String st = "This represents a project, which currently represents information from ";
				if (getNumberLinkedFiles() == 1) {
					st += " a single file.";
					if (!StringUtil.blank(getHomeDirectoryName()))
						st += " The file is located at " + getHomeDirectoryName() + ".";
				}
				else {
					st += getNumberLinkedFiles() + ". The home file of the project is " + getHomeFileName();
					if (!StringUtil.blank(getHomeDirectoryName()))
						st +=  " and is located at " + getHomeDirectoryName() + ".";
				}
				((MesquiteWindow)f).setExplanation(st);
			}
		}
		else if (action == HNode.MOUSEEXIT){
			/*
			Frame f = MesquiteTrunk.mesquiteTrunk.containerOfModule();
			if (f != null && f instanceof MesquiteWindow){
				((MesquiteWindow)f).setExplanation("");
			}
			*/
		}
	}
	/* ---------------- for HNode interface ----------------------*/
	/** for HNode interface */
	public String getTypeName(){
		return null; //"Project"; not needed since name of project includes "Project"
	}
	/* ---------------- for HNode interface ----------------------*/
	/** for HNode interface */
	public void hSupplementTouched(int index){
	}
	/* ---------------- for HNode interface ----------------------*/
	public Image getHImage(){
		return null;
	}
	/* ---------------- for HNode interface ----------------------*/
	public Color getHColor(){
		return ColorDistribution.project[getProjectColor()];  //project color
	}
	/* ---------------- for HNode interface ----------------------*/
	public boolean getHShow(){
		return true; 
	}
}

