/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.lib;

import java.awt.*;
import java.util.*;
import mesquite.lib.duties.*;

/* ======================================================================== */
/** The panel in which a tree is drawn.  This is used within the main tree window, and can be used by other
modules.  
If multiple trees need to be drawn, multiple TreeDisplays should be created.  A TreeDisplay is created
and supervised by a tree draw coordinator module.  A reference to a TreeDisplay is needed 
for shading and decorating the tree. <P>

Within the graphics Panel which is the TreeDisplay, a special object of class TreeDrawing is
actually the one responsible for drawing the tree itself.  The TreeDrawing is created by the
chosen tree drawing module (e.g. diagonal draw tree).  Modules can contribute to the drawing of the
TreeDisplay, whether shading or node pictures or a simple tree legend, by creating an object
of class TreeDisplayExtra and adding it to the vector of extras.  The TreeDisplay calls all of 
the extras after the tree is drawn, in case those extras need to contribute to the TreeDisplay drawing.
<p> This needs to be fixed up (e.g., fields made private with set and get methods).
*/
public class TreeDisplay extends MesquitePanel  {
	/** The tree being drawn */
	protected Tree tree;
	/** A tree which is to be drawn, but which is being held until drawing of the current tree is finished. */
	protected Tree holdingTree;
	/** Orientation of the tree unspecified */
	public static final int FREEFORM = -1;
	/** Orientation of the tree with tips up */
	public static final int UP = 0;
	/** Orientation of the tree with tips to right */
	public static final int RIGHT = 1;
	/** Orientation of the tree with tips to left */
	public static final int LEFT = 2;
	/** Orientation of the tree with tips down */
	public static final int DOWN = 3;
	/** Tree is drawn in circular arrangement of nodes */
	public static final int CIRCULAR = 4;
	
	/**  TreeDisplayExtras attached to this TreeDisplay*/
	private ListableVector extras;
	/**  This is part of system to record whether the tree needs redrawing, used so as to prevent multiple updates*/
	protected int repaintsPending =0;
	/**  Records whether tree drawing is to be suppressed*/
	protected boolean suppress = true;
	/**  Records whether or not the taxon names should be centered, if drawn in one plane, about axis of longest name*/
	public boolean centerNames = false;
	/**  Records whether or not the taxon names are to be drawn*/
	public boolean suppressNames = false;
	/**  The margin from the tips to the edge of the drawing field*/
	public int tipsMargin = -1;
	/**  If true, then branches are drawn proportional to their lengths*/
	public boolean showBranchLengths = false;
	/**  Scaling of the tree drawing*/
	public double scaling = 1.0;
	/**  If tree drawn with fixed depth, this is the depth.*/
	public double fixedDepthScale = 1.0;
	/**  Records whether fixed depth scale is in use.*/
	public boolean fixedScalingOn = false;
	/**  If true, then in text version draw the extra information directly on the tree; otherwise use node lists*/
	public boolean textVersionDrawOnTree = false;
	private int dist=8;
	private int minDist=8;
	int minForTerminalBoxes = 0;
	
	
	/**  The color of the branches*/
	public Color branchColor;
	/**  The color of a dimmed branch*/
	public Color branchColorDimmed;
	/**  The panel in which the tree is actually drawn.*/
	private TreeDrawing treeDrawing;
	/**  The width of the drawn edges (branches)*/
	private int edgewidth;

	private boolean paintedOnce = false;
	/**  Spacing in pixels between taxa*/
	private int taxonSpacing;
	/**  Orientaton of the tree*/
	private int treeOrientation = UP;
	/**  Is the orientation fixed, or can reorientation be done?*/
	private boolean allowReorient = true;
	/**  A notice that is drawn in the display.*/
	private String notice;
	/**  The taxa belonging to the tree*/
	private Taxa taxa;
	/**  Is a frame to be drawn in the TreeDisplay?*/
	private boolean frameTree=false;
	/**  The area in which the tree is drawn*/
	Rectangle field;
	/**  */
	private int taxonNameBuffer = 30;
	/**  */
	private boolean inProgress = false;
	/**  */
	protected MesquiteModule ownerModule;
	/**  */
	public static long totalCreated;
	/**  */
	private boolean isInvalid = false;
	/**  */
	
	public double[] nodeLocsParameters = new double[8]; //parameters for use by node locs module
	public double[] drawParameters = new double[8];//parameters for use by draw tree module
	DrawTaxonNames namesTask;
	private MesquiteInteger highlightedBranch  = new MesquiteInteger(0);
	
	static {
		totalCreated = 0;
	}
	
	public TreeDisplay (MesquiteModule ownerModule, Taxa taxa) { 
		super();
		setDontDuplicateCommands(true);
		if (MesquiteTrunk.isMacOSX())
			setMoveFrequency(4);
		else
			setMoveFrequency(0);
		totalCreated++;
		this.ownerModule=ownerModule;
		field = new Rectangle();
		extras = new ListableVector();
		this.taxa = taxa;
		setLayout(null);
		branchColor = Color.black;
		branchColorDimmed = Color.gray;
	}
	
	public Color getBranchColor(int N){
		long c = tree.getAssociatedLong(ColorDistribution.colorNameReference, N);
		Color col=null;
		if (!tree.anySelected() || tree.getSelected(N)) {
			if (MesquiteLong.isCombinable(c) && (col = ColorDistribution.getStandardColor((int)c))!=null)
				return col;
			else
				return branchColor;
		}
		else {
			if (MesquiteLong.isCombinable(c) && (col = ColorDistribution.getStandardColorDimmed((int)c))!=null)
				return col;
			else
				return branchColorDimmed;
		}
	}
	public void setTreeDrawing(TreeDrawing td, CommandRecord commandRec) {
		treeDrawing = td;
		if (treeDrawing!=null && tree!=null)
			setTree(tree, commandRec);
	}
	public TreeDrawing getTreeDrawing() {
		return treeDrawing;
	}
	public boolean getDrawingInProcess(){
		return inProgress;
	}
	   
	public void setDrawingInProcess(boolean inProgress, CommandRecord commandRec){
		this.inProgress= inProgress;
		if (!inProgress && holdingTree != null) {
			setTree(holdingTree, commandRec);
			//repaint();
		}
	}
	public DrawTaxonNames getDrawTaxonNames(){
		return namesTask;
	}
	public void setDrawTaxonNames(DrawTaxonNames dtn){
		namesTask = dtn;
	}
	public Taxa getTaxa() {
		return taxa;
	}
	public void setTaxa(Taxa taxa) {
		this.taxa = taxa;
	}
	public Tree getTree() {
		return tree;
	}
	
	public void setTree(Tree tree, CommandRecord commandRec) {
		if (tree!=null && tree.getTaxa() != taxa)
			setTaxa(tree.getTaxa());

	 	if (inProgress)
	 		holdingTree = tree;
	 	else {
			this.tree = tree;
			if (treeDrawing !=null) {
				if (tree !=null)
					treeDrawing.setDrawnRoot(tree.getRoot());
				else
					treeDrawing.setDrawnRoot(0);
			}
			redoCalculations(1, commandRec);
	 		holdingTree = null;
	 	}
	}
	public void recalculatePositions(CommandRecord commandRec) {
		if (!inProgress)
			redoCalculations(12, commandRec);
	}
	boolean tndExplicitlySet = false;
	public void setMinimumTaxonNameDistance(int minForTerminalBoxes, int min) {
		this.minForTerminalBoxes = minForTerminalBoxes;
		this.minDist = min;
		if (!tndExplicitlySet || dist<minDist)
			dist = minDist;
	}
	public void setTaxonNameDistance(int newDist) { 
		if (newDist>=minDist) {
			this.dist = newDist;
			tndExplicitlySet = true;
		}
	}
	public int getTaxonNameDistance() {
		if (treeDrawing != null && treeDrawing.terminalBoxesRequested())
			return dist + minForTerminalBoxes;
		else
			return dist;
	}
	public void setTaxonNameBuffer(int b) {
		taxonNameBuffer = b;
	}
	public int getTaxonNameBuffer() {
		return taxonNameBuffer;
	}
	public void setTipsMargin(int margin) {
		tipsMargin = margin;
	}
	public int getTipsMargin() {
		return tipsMargin;
	}
	public void setInvalid(boolean iv) {  //this does not cause a redraw; simply posting a note.  Flag gets should be set to false in  drawing
		this.isInvalid = iv;
	}
	public boolean getInvalid() { 
		return isInvalid;
	}
	private void updateComponents(Container c){
		Component[] cc = c.getComponents();
		for (int i=0; i<cc.length; i++) {
			cc[i].repaint();
			if (cc[i] instanceof Container)
				updateComponents((Container)cc[i]);
		}
	}
	
	public void repaintAll() {  //TODO: this whole system needs revamping.  
		if (isVisible()) {
			repaint();
			updateComponents(this);
		}
	}
	public int getHighlightedBranch() {  
		return highlightedBranch.getValue();
	}
	public void setHighlightedBranch(int value) {  
		highlightedBranch.setValue(value);
	}
	public void pleaseUpdate() {  //TODO: this whole system needs revamping.  
		repaint(false, null);
	}
	public void pleaseUpdate(boolean resetTree, CommandRecord commandRec) {  //TODO: this whole system needs revamping.  
		if (commandRec == null)
			repaint(resetTree, CommandRecord.getRecNSIfNull());
		else
			repaint(resetTree, commandRec);
		MesquiteWindow mw = MesquiteWindow.windowOfItem(this);
		if (mw !=null && mw.getMode()>0)
			mw.updateTextPage();
	}
	public void repaint(boolean resetTree, CommandRecord commandRec) {  //TODO: this whole system needs revamping.  
		repaint();
	}
	public void repaint() {  //TODO: this whole system needs revamping.  
		if (!isVisible())
			return;
		repaintsPending++;
		if (repaintsPending<=1)
			super.repaint();
	}
	public void setFrame(boolean doFrame) {
		frameTree = doFrame;
	}
	public void setSize(){
		
	}
	private void redoCalculations(int code, CommandRecord commandRec){
		if (treeDrawing!=null && tree !=null)
			treeDrawing.recalculatePositions(tree, commandRec); //to force node locs recalc
		
	}
	public void setFieldSize(int x, int y) {
		boolean ch = (field.width!=x || field.height !=y);
		field.width=x;
		field.height=y;
		if (ch)
			redoCalculations(4, CommandRecord.getRecNSIfNull()); 
	}
	public void setFieldWidth(int x) {
		boolean ch = (field.width!=x);
		field.width=x;
		if (ch)
			redoCalculations(5, CommandRecord.getRecNSIfNull()); 
	}
	public void setFieldHeight(int y) {
		boolean ch = (field.height !=y);
		field.height=y;
		if (ch)
			redoCalculations(6, CommandRecord.getRecNSIfNull()); 
	}
	public int getFieldWidth() {
		//return getBounds().width;
		return field.width;
	}
	public int getFieldHeight() {
		//return getBounds().width;
		return field.height;
	}
	public Rectangle getField() { 
		//return getBounds();
		return field;
	}
	public void setEdgeWidth(int sp) {
		this.edgewidth = sp;
	}
	public int getEdgeWidth() {
		return edgewidth;
	}
	public void setTaxonSpacing(int sp) {
		this.taxonSpacing = sp;
	}
	public int getTaxonSpacing() {
		return taxonSpacing;
	}
	public void setNotice(String s) {
		notice = s;
	}
	public ListableVector getExtras() {
		return extras;
	}
	public void addExtra(TreeDisplayExtra extra) {
		if (extras != null)
		extras.addElement(extra, false);
	}
	public void removeExtra(TreeDisplayExtra extra) {
		if (extras != null)
			extras.removeElement(extra, false);
	}
	public boolean findExtra(TreeDisplayExtra extra) {
		if (extras == null)
			return false;
		return (extras.indexOf(extra) >= 0);
	}
	public TreeDisplayExtra[] getMyExtras(MesquiteModule mb) {
		if (extras != null) {
			int count = 0;
			Enumeration e = extras.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				TreeDisplayExtra ex = (TreeDisplayExtra)obj;
			   	if (ex.getOwnerModule() == mb) 
			   		 count++;
		 	}
		 	if (count == 0)
		 		return null;
		 	TreeDisplayExtra[] ee = new TreeDisplayExtra[count];
			e = extras.elements();
			count = 0;
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				TreeDisplayExtra ex = (TreeDisplayExtra)obj;
			   	if (ex.getOwnerModule() == mb) 
			   		ee[count++] = ex;
		 	}
		 	return ee;
		}
		return null;
	}
	public void setTreeAllExtras(Tree tree, CommandRecord commandRec) {
		if (extras != null) {
			Enumeration e = extras.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				TreeDisplayExtra ex = (TreeDisplayExtra)obj;
			   	if (ownerModule==null || ownerModule.isDoomed()) 
			   		return;
	 			ex.setTree(tree, commandRec);
		 	}
		}
	}
	public void drawAllBackgroundExtrasOfPlacement(Tree tree, int drawnRoot, Graphics g, int placement) {
		if (tree == null || tree.getTaxa().isDoomed())
			return;
		if (extras != null) {
			Enumeration e = extras.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				TreeDisplayExtra ex = (TreeDisplayExtra)obj;
	 			if (ex instanceof TreeDisplayBkgdExtra && ex.getPlacement()==placement) {
				   	if (ownerModule==null || ownerModule.isDoomed()) 
				   		return;
	 				ex.drawOnTree(tree, drawnRoot, g);
	 			}
	 		}
		}
	}
	public void drawAllBackgroundExtras(Tree tree, int drawnRoot, Graphics g) {
		drawAllBackgroundExtrasOfPlacement(tree,drawnRoot,g,TreeDisplayExtra.BELOW);
		drawAllBackgroundExtrasOfPlacement(tree,drawnRoot,g,TreeDisplayExtra.NORMAL);
		drawAllBackgroundExtrasOfPlacement(tree,drawnRoot,g,TreeDisplayExtra.ABOVE);
	}
	public void drawAllExtras(Tree tree, int drawnRoot, Graphics g) {
		if (tree == null || tree.getTaxa().isDoomed())
			return;
		if (extras != null) {
			Enumeration e = extras.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				TreeDisplayExtra ex = (TreeDisplayExtra)obj;
	 			if (!(ex instanceof TreeDisplayBkgdExtra)) {
				   	if (ownerModule==null || ownerModule.isDoomed()) 
				   		return;
	 				ex.drawOnTree(tree, drawnRoot, g);
	 			}
	 		}
		}
		if (notice!=null)
			g.drawString(notice, 6, getBounds().height-6);
		if (frameTree)
			g.drawRoundRect(0, 0, getBounds().width, getBounds().height, 10, 10);
	}
	public void printAllBackgroundExtras(Tree tree, int drawnRoot, Graphics g) {
		if (tree == null || tree.getTaxa().isDoomed())
			return;
		if (extras != null) {
			Enumeration e = extras.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				TreeDisplayExtra ex = (TreeDisplayExtra)obj;
	 			if (ex instanceof TreeDisplayBkgdExtra) {
				   	if (ownerModule==null || ownerModule.isDoomed()) 
				   		return;
	 				ex.printOnTree(tree, drawnRoot, g);
	 			}
	 		}
		}
	}
	public void printAllExtras(Tree tree, int drawnRoot, Graphics g) {
		if (tree == null || tree.getTaxa().isDoomed())
			return;
		if (extras != null) {
			Enumeration e = extras.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				TreeDisplayExtra ex = (TreeDisplayExtra)obj;
	 			if (!(ex instanceof TreeDisplayBkgdExtra)) {
				   	if (ownerModule==null || ownerModule.isDoomed()) 
				   		return;
		 			ex.printOnTree(tree, drawnRoot, g);
	 			}
	 		}
		}
	}
	
	public void suppressDrawing(boolean toSuppress) {
		suppress=toSuppress;
	}
	
	/*.................................................................................................................*/
	public void setOrientation(int orient) {
		if (allowReorient)
			treeOrientation = orient;
	}
	/*.................................................................................................................*/
	public int getOrientation() {
		return treeOrientation;
	}
	/*.................................................................................................................*/
	public void setAllowReorientation(boolean allow) {
		allowReorient = allow;
	}
	/*.................................................................................................................*/
	public boolean getAllowReorientation() {
		return allowReorient;
	}
	public void paint(Graphics g) {
	   	if (MesquiteWindow.checkDoomed(this))
	   		return;
	   	paintedOnce = true;
		MesquiteWindow.uncheckDoomed(this);
	}
	public String getTextVersion() {
		if (tree==null || treeDrawing == null)
			return "";
		String s = "Tree Description:  " +  tree.writeTreeByNames(true);
		TextTree tt = new TextTree(tree);
		String[] nodeNumbers = new String[tree.getNumNodeSpaces()];
		for (int i=0; i< nodeNumbers.length; i++)
			nodeNumbers[i] = Integer.toString(i);
		StringBuffer buff = new StringBuffer(50);
		tt.drawTreeAsText(tree, buff, nodeNumbers);
		if (textVersionDrawOnTree)
			s+= "\n" + buff.toString();
		else
			s+= "\nTree with node numbers:\n" + buff.toString();
		if (extras != null) {
			Enumeration e = extras.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				TreeDisplayExtra ex = (TreeDisplayExtra)obj;
				if (ex!=null){
					String sEx =null;
					if (textVersionDrawOnTree)
						sEx =ex.writeOnTree(tree, treeDrawing.getDrawnRoot());
					else
						sEx = ex.infoAtNodes(tree, treeDrawing.getDrawnRoot());
		 			if (!StringUtil.blank(sEx)) {
						String owner = "";
						if (ex.getOwnerModule()!=null)
							owner = ex.getOwnerModule().getName();
						s+= "\n\n--------------- " + owner + " ---------------";
		 				s+= "\n\n"+ sEx + "\n";
		 			}
	 			}
	 		}
		}
		return s;
	}
	public void fillTaxon(Graphics g, int m){}
	public void addPanelPlease(Panel p){
		if (p==null)
			return;
		add(p);
		MesquiteWindow window = MesquiteWindow.windowOfItem(this);
		if (window !=null)
			window.componentAdded(this, p);
	}
	public void removePanelPlease(Panel p){
		if (p==null)
			return;
		remove(p);
	}
	public void addComponentPlease(Component p){
		if (p==null)
			return;
		add(p);
	}
	public void removeComponentPlease(Component p){
		if (p==null)
			return;
		remove(p);
	}
	public void moveExtraToFront(TreeDisplayExtra extra){
		if (extra == null)
			return;
		Vector panels = extra.getPanels();
		for (int i=0; panels !=null && i<panels.size(); i++) {
			Panel p = (Panel)panels.elementAt(i);
			remove(p);
			add(p);
		}
		getExtras().removeElement(extra, false);
		getExtras().addElement(extra, false);
		pleaseUpdate(false, CommandRecord.scriptingRecord);
	}
	public void destroyExtras(){
		if (extras!=null)
			extras.removeAllElements(true);
	}
	public void dispose(){
		MesquiteWindow w = MesquiteWindow.windowOfItem(this);
		if (w!=null)
			w.waitUntilDisposable();
		
		ownerModule =null;
		if (treeDrawing !=null)
			treeDrawing.dispose();
		treeDrawing = null;
		if (extras != null) {
			Enumeration e = extras.elements();
			while (e.hasMoreElements()) {
				Object obj = e.nextElement();
				TreeDisplayExtra ex = (TreeDisplayExtra)obj;
	 			ex.dispose();
	 		}
		}
		destroyExtras();
		extras = null;
		/*
		protected Tree tree, holdingTree;
		public Vector extras;
		private TreeDrawing treeDrawing;
		protected MesquiteModule ownerModule;
		*/
		super.dispose();
	}

}

