/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.lib;

import java.awt.*;
import mesquite.lib.duties.*;
/*===  Mesquite Basic Class Library:  Trees    ===*/
/*===  the basic classes used by the trunk of Mesquite and available to the modules    ===*/
/* This sublibrary includes taxa and tree classes */



/*  trees  */
/* ======================================================================== */
/**This is the base class for the drawing of the tree.  It contains the information about
the number of nodes, and the x and y positions of the nodes.  It should be subclassed by tree drawing modules, 
so that they can add their own needed field (like polygons for the
tree branches.  TreeDrawings are typically used within TreeDisplays.*/

public abstract class TreeDrawing  {
	public int[] x; //x positions of nodes
	public int[] y; //y positions of nodes
	public double[] z; //z positions of nodes (closeness to viewer, smaller numbers closer)
	public int[] lineBaseX; //base of line on which to draw labels etc.
	public int[] lineBaseY; 
	public int[] lineTipX; //tip of line on which to draw labels etc.
	public int[] lineTipY; 
	/**labelOrientation indicates where label is to be drawn w.r.t. node, in degrees. 0 = normal horizontal 
	writing to right of node, as would be done for a tree with orientation RIGHT.
	This does not represent simple rotation, i.e. 180 is on left side, but the writing is not upside down.  Thus
	0 would be appropriate for tree oriented RIGHT, 90 for tree DOWN, 180 for tree LEFT, 270 for tree UP */
	public int[] labelOrientation; 
	
	/** namesFollowLines indicates whether orientation of names follows the base to tip line of the branch.
	Supercedes labelOrientation for terminals and permits Java2D rotation of taxon names in circular tree */
	public boolean namesFollowLines = false;

	/** namesAngle indicates rotation of names from default in radians  (useful only for Java 1.2 or higher)*/
	public double namesAngle = MesquiteDouble.unassigned;

	public int numNodes;
	public TreeDisplay treeDisplay;//TODO: many of these should be private fields!!!
	private int enableTerminalBoxesRequests = 0;
	private int drawnRoot=-1;
	public static long totalCreated = 0;
	public Polygon[] namePolys;
	
	public TreeDrawing (TreeDisplay treeDisplay, int numNodes) {
		this.treeDisplay = treeDisplay;
		totalCreated++;
		resetNumNodes(numNodes);
	}
	public void resetNumNodes(int numNodes){
		if (this.numNodes == numNodes && x != null)
			return;
		this.numNodes=numNodes;
		totalCreated++;
		x = new int[numNodes];
		y = new int[numNodes];
		z = new double[numNodes];
		labelOrientation = new int[numNodes];
		lineBaseX = new int[numNodes];
		lineBaseY = new int[numNodes];
		lineTipX = new int[numNodes];
		lineTipY = new int[numNodes];
		for (int i=0; i<numNodes; i++) {
			x[i]=0;
			y[i]=0;
			z[i]=MesquiteInteger.unassigned;
			lineBaseX[i]=0;
			lineBaseY[i]=0;
			lineTipX[i]=0;
			lineTipY[i]=0;
			labelOrientation[i] = MesquiteInteger.unassigned;
		}
	}

	public int getDrawnRoot(){
		return drawnRoot;
	}
	public void setDrawnRoot(int node){
		drawnRoot = node;
	}
	

	/** Sets the tree.  This is done outside of a paint() call, and is the place that any complex calculations should be performed! */
	public abstract void recalculatePositions(Tree tree, CommandRecord commandRec) ;

	/** Draw tree in graphics context */
	public abstract void drawTree(Tree tree, int drawnRoot, Graphics g) ;
	
	/** Fill branch N with current color of graphics context */
	public abstract void fillBranch(Tree tree, int N, Graphics g);
	/*_________________________________________________*/
	/** Does the basic inverting of the color of a branch **/
	public  void fillBranchInverted (Tree tree, int N, Graphics g) {
		g.setColor(Color.black);
		g.setXORMode(Color.white);  //for some reason color makes no difference in MacOS, but is inversion color in Win95 
		fillBranch(tree, N, g);
		g.setPaintMode();
		g.setColor(Color.black);
	   }
	
	/** Fill branch N with indicated set of colors */
	public abstract void fillBranchWithColors(Tree tree, int N, ColorDistribution colors, Graphics g);
	
	/** Find which branch is at x,y (-1 if none) */
	public abstract int findBranch(Tree tree, int drawnRoot, int x, int y);
	
	/** Put in a request to enable terminal boxes (this allows taxon names to be drawn to leave room)*/
	public void incrementEnableTerminalBoxes(){
		enableTerminalBoxesRequests++;
	}
	/** Withdraw request to enable terminal boxes (this allows taxon names to be drawn to leave room)*/
	public void decrementEnableTerminalBoxes(){
		enableTerminalBoxesRequests--;
		if (enableTerminalBoxesRequests<0)
			enableTerminalBoxesRequests=0;
	}
	/** Returns if terminal boxes requested*/
	public boolean terminalBoxesRequested(){
		return (enableTerminalBoxesRequests>0);
	}
	/** Fill terminal box with current color. */
	public abstract void fillTerminalBox(Tree tree, int node, Graphics g);
	
	/** Fill terminal box of node "node" with indicated set of colors */
	public abstract void fillTerminalBoxWithColors(Tree tree, int node, ColorDistribution colors, Graphics g);
	
	/** Find which terminal box is at x,y */
	public abstract int findTerminalBox(Tree tree, int drawnRoot, int x, int y);
	
	/** Draw highlight for branch N with current color of graphics context */
	public void drawHighlight(Tree tree, int N, Graphics g, boolean flip){}
	

	
	/** project point x,y onto the line between N's lineBase to lineTip. */
	public Point projectionOnLine(int N, int x, int y) {
		Point thePoint= new Point(x, y);
		try {
			if (lineTipX[N] == lineBaseX[N]) {   //slope infinite; projection uses same y and sets x to be lineBaseX
				thePoint.x =lineBaseX[N];
			}
			else if (lineTipY[N] == lineBaseY[N]) {  //slope zero; projection uses same x and sets y to be lineBaseY
				thePoint.y =lineBaseY[N];
			}
			else {
				double m = (lineTipY[N]-lineBaseY[N])*1.0/(lineTipX[N] - lineBaseX[N]);
				//MesquiteModule.mesquiteTrunk.logln(" m " + MesquiteDouble.toString(m));
				double msquare = m*m;
				double newY = (msquare)/(msquare+1)*((x-lineTipX[N])/m + (lineTipY[N]/msquare) + (y));
				double newX = m* (y-newY) + x;
				//MesquiteModule.mesquiteTrunk.logln(" y " + MesquiteDouble.toString(newY));
				//MesquiteModule.mesquiteTrunk.logln(" x " + MesquiteDouble.toString(newX));
				thePoint.y = (int)newY;
				thePoint.x = (int)newX;
			}
		}
		catch (ArrayIndexOutOfBoundsException e) {
		}
		catch (NullPointerException e) {
		}
		return thePoint;
		
	}
	
	/** Choose point on line from lineBase to lineTip that corresponds to the i'th of "total" units along line. */
	public Point placeOnLine(int N, int i, int total) {
		Point thePoint= new Point();
		try {
			if (lineTipX[N] == lineBaseX[N]) {   //slope infinite; projection uses same y and sets x to be lineBaseX
				thePoint.x =lineBaseX[N];
			}
			else if (lineTipY[N] == lineBaseY[N]) {  //slope zero; projection uses same x and sets y to be lineBaseY
				thePoint.y =lineBaseY[N];
			}
			else {
				double m = (lineTipY[N]-lineBaseY[N])*1.0/(lineTipX[N] - lineBaseX[N]);
				double msquare = m*m;
				//double newY = (msquare)/(msquare-1)*((x-lineTipX[N])/m + (lineTipY[N]/msquare) + (y));
				//double newX = m* (y-newY) + x;
				//thePoint.y =( int)newY;
				//thePoint.x =(int) newX;
			}
		}
		catch (ArrayIndexOutOfBoundsException e) {
		}
		catch (NullPointerException e) {
		}
		return thePoint;
	}
	public void setHighlightsOn(boolean on){
		/*
		if (on) {
			if (highlightThread==null) 
				highlightThread = new HighlightThread(treeDisplay);
			if (!highlightThread.isOn())
				highlightThread.start();
		}
		else {
			if (highlightThread!=null) {
				highlightThread.stop();
				highlightThread.dispose();
			}
			highlightThread = null;
		}
		*/
		
	}
	public boolean getHighlightsOn(){
		return false;
		//return (highlightThread!=null);
	}
	public void dispose(){
		treeDisplay = null;
		setHighlightsOn(false);
	}
	public void finalize() throws Throwable {
		setHighlightsOn(false);
		super.finalize();
	}
}

