/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.lib;

import java.awt.*;
import java.util.*;
import mesquite.lib.duties.*;
/* ======================================================================== */
/**A tree block.  Trees are added to it when trees are read from file or stored.  Many methods could be built here, for 
instance to go through all trees and prune taxa deleted.
Translation table is stored with each TreeVector.*/
public class TreeVector extends ListableVector implements Commandable, CompatibilityChecker, MesquiteListener {
	Taxa taxa = null;
	TranslationTable translationTable;
	public static long totalCreated = 0;
	public static long totalDisposed = 0;
	long id;
	
	public TreeVector  (Taxa taxa) {
		super();
		setTaxa(taxa);
		totalCreated++;
		id = totalCreated;
	}
	
	public void addElement(Listable obj, boolean notify) {
		super.addElement(obj, false);
		if (obj instanceof MesquiteTree)
			((MesquiteTree)obj).setTreeVector(this);
		if (obj instanceof Listened){
			((Listened)obj).addListener(this);
		}
		if (notify)
			notifyListeners(this, new Notification(MesquiteListener.PARTS_ADDED));
	}
	public void insertElementAt(Listable obj, int index, boolean notify) {
		super.insertElementAt(obj, index, false);
		if (obj instanceof MesquiteTree)
			((MesquiteTree)obj).setTreeVector(this);
		if (obj instanceof Listened){
			((Listened)obj).addListener(this);
		}
		if (notify)
			notifyListeners(this, new Notification(MesquiteListener.PARTS_ADDED));
	}
	public void removeElement(Listable obj, boolean notify) {
		if (obj == null)
			return;
		super.removeElement(obj, false);
		if (obj instanceof MesquiteTree)
			((MesquiteTree)obj).setTreeVector(null);
		if (obj instanceof Listened){
			((Listened)obj).removeListener(this);
		}
		if (notify)
			notifyListeners(this, new Notification(MesquiteListener.PARTS_DELETED));
	}
	public void removeElementAt(int index, boolean notify) {
		Object obj = elementAt(index);
		if (obj == null)
			return;
		super.removeElementAt(index, false);
		if (obj instanceof MesquiteTree)
			((MesquiteTree)obj).setTreeVector(null);
		if (obj instanceof Listened){
			((Listened)obj).removeListener(this);
		}
		if (notify)
			notifyListeners(this, new Notification(MesquiteListener.PARTS_DELETED));
	}
	public void removeAllElements(boolean notify){
		for (int i=0; i<size(); i++){
			Object obj = elementAt(i);
			if (obj instanceof Listened){
				((Listened)obj).removeListener(this);
		}

		}
		super.removeAllElements(notify);
	}
	public void replaceElement(Listable old, Listable replacement, boolean notify) {
		if (old instanceof Listened){
			((Listened)old).removeListener(this);
		}
		if (old instanceof MesquiteTree)
			((MesquiteTree)old).setTreeVector(null);
		super.replaceElement(old, replacement, false);
		if (replacement instanceof MesquiteTree)
			((MesquiteTree)replacement).setTreeVector(null);
		if (replacement instanceof Listened){
			((Listened)replacement).removeListener(this);
		}
		if (notify)
			notifyListeners(this, new Notification(MesquiteListener.PARTS_CHANGED));
	}
 	/*.................................................................................................................*/
 	/** gets the explanation (footnote) of this tree vector*/
	public String getExplanation() {
		
		String extra = null;
		if (taxa == null){
			extra = "This block of trees has " + size() + " trees for an unspecified taxa block.\n";
		}
		else 
			extra = "This block of trees has " + size() + " trees for the taxa \"" + taxa.getName() + " \".\n";
		return extra;
	}
	/** returns true if object is set of taxa on which tree vector is based */
 	public boolean isCompatible(Object obj, MesquiteProject project, EmployerEmployee prospectiveEmployer){
 		return (obj == taxa);
 	}
	/**Returns the unique id number of this tree block*/
	public long getID(){
		return id;
	}
	/**Translates internal numbering system to external (currently, 0 based to 1 based*/
	public static int toExternal(int i){
		if (!MesquiteInteger.isCombinable(i))
			return i;
		else
			return i+1;
	}
	/**Translates external numbering system to internal (currently, 1 based to 0 based*/
	public static int toInternal(int i){
		if (!MesquiteInteger.isCombinable(i))
			return i;
		else
			return i-1;
	}
	public void dispose() {
		for (int i=0; i<size(); i++)
			((MesquiteTree)elementAt(i)).dispose();
		removeAllElements(false);
		if (taxa != null)
			taxa.removeListener(this);
		totalDisposed++;
		super.dispose();
	}
 	public Object doCommand(String commandName, String arguments, CommandRecord commandRec, CommandChecker checker){
	    	if (checker.compare(this.getClass(), "Returns identification number of this tree block", null, commandName, "getID")) {
			return new MesquiteInteger((int)id);
		}
	    	else if (checker.compare(this.getClass(), "Prepends to all tree names the given string", "[string to prepend]", commandName, "prefixNames")) {
			String prefix = ParseUtil.getFirstToken(arguments, new MesquiteInteger(0));
			if (StringUtil.blank(prefix))
				return null;
			for (int i=0; i<size(); i++) {
				MesquiteTree t = getTree(i);
				if (t!=null) {
					String n = t.getName();
					if (n == null)
						t.setName(prefix);
					else
						t.setName(prefix + n);
				}
			}
			return null;
		}
    	 	else
    	 		return super.doCommand(commandName, arguments, commandRec, checker);
 	}
	/** Set the taxa to which the tree block applies */
	public void setTaxa(Taxa taxa) {
		if (this.taxa!=taxa && taxa !=null){
			if (this.taxa!=null)
				this.taxa.removeListener(this);
			this.taxa = taxa; //need to redo translation table
			taxa.addListenerHighPriority(this);
			translationTable = new TranslationTable(taxa);
		}
	}
	/** Get the taxa to which the tree block applies */
	public Taxa getTaxa(){
		return taxa;
	}
	
	public MesquiteTree getTree(int i) {
		if (i>=0 && i<size())
			return (MesquiteTree)elementAt(i);
		else
			return null;
	}
	/** Set the label of the taxon with given name to be the label passed*/
	public void setTranslationLabel(String label, String taxonName, boolean checkDuplicates){
		if (translationTable==null)
			return;
		if (taxa==null)
			return;
		Taxon t = taxa.getTaxon(taxonName);
		if (t == null) {
			MesquiteMessage.warnProgrammer("Taxon name in translation table doesn't correspond to name of known taxon (\"" + taxonName + "\" [a])");
		}
		
		translationTable.setLabel(t, label, checkDuplicates);
 		setDirty(true);
	}
	/** DOCUMENT */
	public void checkTranslationTable(){
		if (translationTable==null)
			return;
		translationTable.checkDuplicates();
	}
	
	/** Get the label of the taxon i.  If no label stored, return external taxon number as string*/
	public String getTranslationLabel(String iName) {
		if (StringUtil.blank(iName))
			return null;
		Taxon t = taxa.getTaxon(iName);
		if (translationTable!=null) {
			String label = translationTable.getLabel(t);
			if (!StringUtil.blank(label))
				return label;
		}
		if (t != null)
			return Integer.toString(Taxon.toExternal(taxa.whichTaxonNumber(t))); 
		return null;
	}
	/** Get the label of the taxon i.  If no label stored, return external taxon number as string*/
	public String getTranslationLabel(int taxonNumber) {
		if (taxa == null || translationTable==null || translationTable.getLabel(taxonNumber)==null)
			return Integer.toString(Taxon.toExternal(taxonNumber)); 
		return translationTable.getLabel(taxonNumber);
	}

	MesquiteInteger pos = new MesquiteInteger(0);
	
	/** Get the translation table as a string. */
	public String getTranslationTable() {
		if (translationTable==null || taxa == null)
			return null;
		String temp = "";
		for(int i=0; i<taxa.getNumTaxa(); i++) {
			if (i>0)
				temp += ","+ StringUtil.lineEnding();
			String thisLabel = getTranslationLabel(i);
			temp += "\t\t" + thisLabel + " " + StringUtil.tokenize(taxa.getTaxonName(i)) ;
		}
		return temp;
	}
	/** Sets the translation table to match the passed table. */
	public void setTranslationTable(Vector table) {
		if (table==null)
			return;
		for(int i=0; i<table.size(); i++) {
			String taxonName = (ParseUtil.getFirstToken((String)table.elementAt(i),pos));
			String label = (ParseUtil.getToken((String)table.elementAt(i),pos));
			Taxon t = taxa.getTaxon(taxonName);
			translationTable.setLabel(t, label, false); //note label set even if taxon not found; signal later to return null for whichtaxon number
			if (t==null)
				MesquiteMessage.warnProgrammer("Taxon name in translation table doesn't correspond to name of known taxon (\"" + taxonName + "\" [b])");
		}
		checkTranslationTable();
	}
	/** returns true if stored translation table matches the passed taxa. */
	public boolean tableMatchesTaxa(Taxa taxa, Vector table) {
		if (table==null)
			return false;
		for(int i=0; i<table.size(); i++) {
			String taxonName = (ParseUtil.getFirstToken((String)table.elementAt(i),pos));
			if (taxa.whichTaxonNumber(taxonName) <0)
				return false;
		}
		return true;
	}
	/** Returns which taxon number corresponds to the passed taxon label.  Uses the translation table. */
	public int whichTaxonNumber(String label) {
		if (taxa==null)
			return -1;
		if (translationTable==null)
			return taxa.whichTaxonNumber(label);
			
		int i = taxa.whichTaxonNumber(translationTable.getTaxon(label));
		if (i<0)
			return taxa.whichTaxonNumber(label); //added 1 Oct 2001
		else
			return i;
	}
	
	/** Get the name of the tree block. */
	public String getName(){
		if (name==null)
			return "Untitled Trees Block";
		return name;
	}
	/** Set the name of the tree block. */
	public void setName(String name){
		this.name = name;
		if (getHShow()) {
			MesquiteTrunk.mesquiteTrunk.refreshBrowser(FileElement.class);
		}
	}
	
	/** Get the name & details of the tree block. */
	public String toString(){
		return "TreeVector: " + getName() + " [id " + id + " size " + size() + "]";
	}
	/** Get the list of tree id's. */
	public String listIDs(){
		String s = "Trees: ";
		for (int i=0; i<size(); i++)
			s += " " + getTree(i).getID();
		return s;
	}
	/*-----------------------------------------*/
	long previous = -1;
	/** For MesquiteListener interface.  Passes which object changed, along with optional integer (e.g. for character)*/
	public void changed(Object caller, Object obj, Notification notification, CommandRecord commandRec){
		if (notification != null && notification.getID() == previous)
			return;
		if (notification != null)
			previous = notification.getNotificationNumber();
		if (obj == taxa){
			if (Notification.appearsCosmetic(notification))
				return;
			int[] parameters = Notification.getParameters(notification);
			if (parameters!=null && parameters.length>=2 && translationTable!=null)
				translationTable.taxaModified(Notification.getCode(notification), parameters[0], parameters[1]);
			for (int i=0; i<size(); i++) {//modified 19 Nov 01
				Object o = elementAt(i);
				if (o !=null && o instanceof MesquiteTree) {
					MesquiteTree tree = (MesquiteTree)o;
					tree.changed(caller, obj, notification, commandRec);
				}
			}
		}
		else if (obj instanceof Tree && caller != this && Notification.getCode(notification) != MesquiteListener.ANNOTATION_CHANGED)
			notifyListeners(this, (new Notification(MesquiteListener.PARTS_CHANGED).setNotificationNumber(notification.getNotificationNumber()))); 
	}
	/*-----------------------------------------*/
	/** For MesquiteListener interface.  Passes which object was disposed*/
	public void disposing(Object obj){
			if (obj instanceof Taxa) //modified 19 Nov 01
			for (int i=0; i<size(); i++) {
				Object o = elementAt(i);
				if (o !=null && o instanceof MesquiteTree) {
					MesquiteTree tree = (MesquiteTree)o;
					tree.disposing(obj);
				}
			}
	}
	/*-----------------------------------------*/
	/** For MesquiteListener interface.  Asks whether it's ok to delete the object as far as the listener is concerned (e.g., is it in use?)*/
	public boolean okToDispose(Object obj, int queryUser){
		return true;
	}
	
	/*-----------------------------------------*/
	/** For HNode interface.  todo: this should be renamed getHNodeShow*/
	public boolean getHShow(){
		return true;
	}
	/*-----------------------------------------*/
	/** For HNode interface.*/
	public String getTypeName(){
		return "Tree block";
	}

	/* ---------------- for use with touched from HNode interface ----------------------*/
	public void addToBrowserPopup(MesquitePopup popup){
		super.addToBrowserPopup(popup);
		ElementManager manager = getManager();
		if (manager!=null && manager instanceof MesquiteModule) {
			MesquiteFile file = getFile();
			if (file!=null && file.getProject()!=null) {
				popup.add(new MenuItem("-"));
				popup.add(new MesquiteMenuItem("Show list of trees \"" + getName() + "\"", MesquiteTrunk.mesquiteTrunk, MesquiteTrunk.mesquiteTrunk.makeCommand("showMe", this)));
				if (getFile()!=null)
					popup.add(new MesquiteMenuItem("Show trees \"" + getName() + "\" in tree window", MesquiteTrunk.mesquiteTrunk, MesquiteTrunk.mesquiteTrunk.makeCommand("showTreesInWindow" , (MesquiteModule)manager),Integer.toString(getFile().getProject().getFileElementNumber(this, TreeVector.class))));
			}
		}
	}
}



