/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.lib.characters; 

import java.awt.*;
import java.util.*;
import java.util.zip.*;
import mesquite.lib.duties.*;
import mesquite.lib.*;

/* last documented: April 2003 */
/* ======================================================================== */
/**An object of this class represents a Characters block within a NEXUS file.  It is basically a data matrix
plus additional information about the characters, symbols, etc. Includes methods to add, delete and move characters. 
Note: methods to adjust taxa are in Taxa, but they need to appear here also so that matrices may be adjusted
 as needed (to delete rows).<p>
This is subclassed separately for categorical and continuous data; it is there that the matrices appear. Also, methods to set and get character
states occur in generic form using CharacterState objects, but also datatype-specific methods appear in the subclasses to set or return directly 
the character state in its specific form (e.g. long getState(ic, it). 
<p>
CharacterData is a subclass of Associable, which is considered to consist of a series of parts to each of which information may be attached.
The "parts" of a CharacterData are the characters.  Thus, when addParts, deleteParts, moveParts and so on are called, the response is
to add, delete and move characters.  To change the rows (taxa) of a matrix, the methods referring explicitly to taxa (addTaxa, etc.) 
are to be used.
<p>
Attached to the CharacterData object are the typesets, weightsets, character sets and so on.  Each type of set is grouped into a SpecsSetVector of SpecsSets.
These are handled by the superclass AssociableWithSpecs.  These different types of sets are not predefined within
the CharacterData object, but are established by their managing modules and stored here.<p>
  See general discussion of character storage classes under CharacterState
 */
 
public abstract class CharacterData extends FileElement implements MesquiteListener, StringLister, Identifiable, CompatibilityChecker  {
	protected  int numTaxa; //number of taxa (rows): also determinable by taxa.getNumTaxa()
	protected  int numChars; //number of characters (columns)
	
	private Taxa taxa; //taxa to which this matrix belongs
	private long[] taxaIDs; //the remembered id's of the taxa; to use to reconcile changed Taxa with last used here
	
	private String[] characterNames; //names of characters.  State names are needed only for categorical characters, and are to be found in CategoricalData
	private Image[] characterIllustrations; 
	private String[] characterIllustrationPath; 
	
	//information attached to individual cells of the matrix
	private String[][] footnotes; //strings attached to each cell
	private Vector cellObjects; //Vector of arrays of objects (Object2DArray) that are attached to cells.  A courtesy to modules, so that they can attach and maintain info at the cells
	private boolean[][] cellObjectsDisplay; //indicates whether there exist cell objects at a cell that need to be displayed in any way
	private boolean[][] changedSinceSave; //records whether changed since last save.  
	
	//the following four fields are for the experimental facility to record history of changes to cells
	public static boolean defaultSaveChangeHistory = false;
	public boolean saveChangeHistory = defaultSaveChangeHistory;//todo: make, set
	public static boolean defaultRequireChangeAuthority = false;
	public boolean requireChangeAuthority = defaultRequireChangeAuthority;//todo: make, set
	ChangeAuthority currentChangeAuthority;
	
	private boolean checksumValid = false;
	private long checksum = 0;
	private CRC32 crc32 = new CRC32();
	
	public static final char defaultInapplicableChar = '-';
	public static final char defaultMissingChar = '?';
	public static final char defaultMatchChar = '.';
	private char inapplicableChar = defaultInapplicableChar;
	private char missingChar = defaultMissingChar;
	private char matchChar = defaultMatchChar;
	
	protected CharMatrixManager matrixManager; //module which will supervise this (e.g., writing Characters block to file)
	
	Tree basisTree; //the tree that was the basis of this matrix (i.e. on which it was simulated), if any
	protected int nAdd = 0, nDel = 0, nMove = 0; //for debugging
	
	
	//boolean locked = false; LOCKING SYSTEM not yet in place
	
	boolean columnWidthAutoSize=true;
	public boolean interleaved = false; //todo: used as a flag in reading.  Better to be part of general system of info storage for options and format issues
	public int interleavedLength = 100;
	
	
	public final static int OK = 1;
	public final static int ERROR = 2;
	public final static int OUTOFBOUNDS = 3;
	public final static int EOL = 4;
	public final static int COMMENT = 5;
	
	
	/** Vector of other CharacterData linked to this one.  Linkage requires a 1-to-1 correspondence of characters, so that for each character in one matrix there's
	a corresponding charcter in another.  This allows compound data, e.g. a ProteinData (when it exists!) could be linked to a ContinuousData holding landmark
	positions of amino acid so that both the sequence and 3D conformation of protein are stored. */
	protected Vector linkedDatas = null;
	
	private long id;
	
	public static long totalCreated = 0;
	public static long totalDisposed = 0;

	
	public CharacterData(CharMatrixManager matrixManager, int numTaxa, int numChars, Taxa taxa){
		super(numChars); // for Associable
		NameReference sN = makeAssociatedBits("selected"); //this won't make new Bits if not needed, just return reference
		selected = getWhichAssociatedBits(sN);
		this.matrixManager = matrixManager;
		this.numChars=numChars;
		this.numTaxa = numTaxa;
		this.taxa = taxa;
		linkedDatas = new Vector();
		taxaIDs = taxa.getTaxaIDs();
		taxa.addListenerHighPriority(this);  //TODO: this is dangerous for simulation matrices; if not disposed will be source of memory leak

		/* the following should probably be done only on demand (e.g., so that simulation matrices don't need to remake this)*/
		characterNames = new String[numChars];
		cellObjects = new Vector();
		cellObjectsDisplay = new boolean[numChars][numTaxa];
		changedSinceSave = new boolean[numChars][numTaxa];
		
		totalCreated++;
		id = totalCreated;
	}
	public void dispose(){
		if (taxa!=null)
			taxa.removeListener(this);
		if (getProject()!=null && getProject().getCentralModelListener() !=null)
			getProject().getCentralModelListener().removeListener(this);
		else if (getFile() !=null && getFile().getProject() != null && getFile().getProject().getCentralModelListener() != null)
			getFile().getProject().getCentralModelListener().removeListener(this);
		if (basisTree !=null) {
			basisTree.dispose();
		}
		resignFromLinkageGroup();
		linkedDatas=null;
		taxa = null;
		characterNames = null; 
		characterIllustrations = null; 
		characterIllustrationPath = null; 
		CharacterData.totalDisposed++;
		super.dispose();
	}
	public long getID(){
		return id;
	}
	public NexusBlock addToFile(MesquiteFile f, MesquiteProject proj, ElementManager manager){
		if (proj!=null && proj.getCentralModelListener() !=null) {
			proj.getCentralModelListener().addListener(this);
		}
		else if (f !=null && f.getProject() != null && f.getProject().getCentralModelListener() != null){
			f.getProject().getCentralModelListener().addListener(this);
		}
		return super.addToFile(f, proj, manager);
	}
 	public boolean isCompatible(Object obj, MesquiteProject project, EmployerEmployee prospectiveEmployer){
 		if (obj ==null)
 			return true;
 		else if (obj instanceof Taxa){
 			return taxa == obj;
 		}
 		else if (obj instanceof Class){
 			Class c = (Class)obj;
 			return c.isAssignableFrom(getClass()) || c.isAssignableFrom(getStateClass());
 		}
 		return true;
 	}
	/** sets name of data matrix */
	public void setName(String name) {
		this.name= name;
		notifyListeners(this, new Notification(NAMES_CHANGED));
		if (getHShow()) {
			MesquiteTrunk.mesquiteTrunk.refreshBrowser(FileElement.class);
		}
	}
 	/** returns title of data if it has a title, or "Untitled Character Matrix" if has no title*/
	public String getName() {
		if (name==null)
			return "Untitled Character Matrix";
		else
			return name;
	}
 	/** returns true if data has a title*/
	public boolean hasTitle() {
		return  (name!=null);
	}
	public String getTypeName(){
		return "Character matrix";
	}
	
	/**Returns the type of data stored. */
	public abstract String getDataTypeName();
	
	/** clones the data set*/
	public abstract CharacterData cloneData(); //TODO: SHOULD HERE PASS boolean to say whether to DEAL WITH CHARACTER SPEC SETS, character names, etc.
	//TODO: also need setToClone(data) method to set specsets and names etc. including super.setToClone()
	
	/** returns the module in charge of managing the data matrix, including reading and writing the corresponding block in a NEXUS file*/
	public CharMatrixManager getMatrixManager() {
		return matrixManager;
	}
	/** sets the module in charge of managing the data matrix, including reading and writing the corresponding block in a NEXUS file*/
	public void setMatrixManager(CharMatrixManager matrixManager) {
		this.matrixManager= matrixManager;
	}
	/** returns number of taxa in data matrix*/
	public int getNumTaxa() {
//		checkTaxaIDs(null);
		return numTaxa;
	}
	/** returns number of characters in data matrix*/
	public int getNumChars() {
// 		checkTaxaIDs(null);
		return numChars;
	}

	/** returns the Taxa object to which it applies*/
	public Taxa getTaxa() {
		return taxa;
	}
	/*-----------------------------------------------------------*/
	/** A equivalent of addParts but with notification added */
	public final boolean addCharacters(int starting, int num, boolean notify){
		boolean added = addParts(starting, num);
		if (added && notify)
			notifyListeners(this, new Notification(MesquiteListener.PARTS_ADDED, new int[] {starting, num}));
		return added;
	}
	
	/*-----------------------------------------------------------*/
	/**Adds num characters after position "starting".  If "starting" = -1, then inserts at start.  If "starting" >
	number of characters, adds to end.  Any linked CharacterDatas are
	to be adjusted separately.  Returns true iff successful.*/
	public boolean addParts(int starting, int num){
		if (num<=0)
			return false;
		nAdd++;
		if (starting<0)
			starting = -1;
		else if (starting>=numChars)
			starting = numChars-1;
		int newNumChars = numChars + num;
		if (characterNames!=null){
			characterNames = StringArray.addParts(characterNames, starting, num);
			//newCharacterNames;
		}
		if (footnotes!=null){
			String[][] newFootnotes = new String[newNumChars][numTaxa];
			for (int j = 0; j<numTaxa; j++){
				for (int i=0; i<=starting; i++)
					newFootnotes[i][j] = footnotes[i][j];
				for (int i=0; i<num; i++)
					newFootnotes[starting + i + 1][j] = null;
				for (int i=0; i<numChars-starting-1; i++) 
					newFootnotes[i + starting+num+1][j] = footnotes[starting + i+1][j];
			}
			footnotes = newFootnotes;
		}
		if (cellObjects.size()>0){
			for (int k =0; k<cellObjects.size(); k++){
				Object2DArray objArray = (Object2DArray)cellObjects.elementAt(k);
				Object[][] oldObjects = objArray.getMatrix();
				Object[][] newObjects = new Object[newNumChars][numTaxa];
				for (int j = 0; j<numTaxa; j++){
					for (int i=0; i<=starting; i++)
						newObjects[i][j] = oldObjects[i][j];
					for (int i=0; i<num; i++)
						newObjects[starting + i + 1][j] = null;
					for (int i=0; i<numChars-starting-1; i++) 
						newObjects[i +starting+num+1][j] = oldObjects[starting + i+1][j];
				}
				objArray.setMatrix(newObjects);
			}
		}
		if (cellObjectsDisplay!=null){
			boolean[][] newCOD = new boolean[newNumChars][numTaxa];
			for (int j = 0; j<numTaxa; j++){
				for (int i=0; i<=starting; i++)
					newCOD[i][j] = cellObjectsDisplay[i][j];
				for (int i=0; i<num; i++)
					newCOD[starting + i + 1][j] = false;
				for (int i=0; i<numChars-starting-1; i++) 
					newCOD[i + starting+num+1][j] = cellObjectsDisplay[starting + i+1][j];
			}
			cellObjectsDisplay = newCOD;
		}
		if (changedSinceSave!=null){
			boolean[][] newCOD = new boolean[newNumChars][numTaxa];
			for (int j = 0; j<numTaxa; j++){
				for (int i=0; i<=starting; i++)
					newCOD[i][j] = changedSinceSave[i][j];
				for (int i=0; i<num; i++)
					newCOD[starting + i + 1][j] = false;
				for (int i=0; i<numChars-starting-1; i++) 
					newCOD[i + starting+num+1][j] = changedSinceSave[starting + i+1][j];
			}
			changedSinceSave = newCOD;
		}
		if (characterIllustrations!=null){
			Image[] newCharacterIllustrations = new Image[newNumChars];
			for (int i=0; i<=starting; i++) {
				newCharacterIllustrations[i] = characterIllustrations[i];
			}
			for (int i=0; i<num; i++) {
				newCharacterIllustrations[starting + i + 1] = null;
			}
			for (int i=0; i<numChars-starting-1; i++) {
				newCharacterIllustrations[i +starting+num+1] = characterIllustrations[starting + i+1];
			}
			characterIllustrations = newCharacterIllustrations;
		}
 		numChars = newNumChars;
 		super.addParts(starting, num); //for specssets
		return true;
	}
	/*-----------------------------------------------------------*/
	/** Adds characters in linked data files. */
	public final void addInLinked(int starting, int num, boolean notify){
		if (linkedDatas.size()>0){
			for (int i=0; i<linkedDatas.size(); i++){
				CharacterData d = (CharacterData)linkedDatas.elementAt(i);
				d.addParts(starting, num);
				if (notify)
					d.notifyListeners(this, new Notification(MesquiteListener.PARTS_ADDED, new int[] {starting, num}));
			}
		}
	}
	/*-----------------------------------------------------------*/
	/** An equivalent to deleteParts but with notification added. Final because overriding should be done of the Parts method instead*/
	public final boolean deleteCharacters(int starting, int num, boolean notify){
		boolean deleted = deleteParts(starting, num);
		if (deleted && notify)
			notifyListeners(this, new Notification(MesquiteListener.PARTS_DELETED, new int[] {starting, num}));
		return deleted;
	}
	/*-----------------------------------------------------------*/
	/** deletes num characters from (and including) position "starting"; returns true iff successful.  Should be overridden by particular subclasses, but this called via super so it can clean up.*/
	public boolean deleteParts(int starting, int num){
		if (num<=0)
			return false;
		if (starting<0)
			return false;
		else if (starting>numChars)
			return false;
		nDel++;
		if (num+starting>numChars)
			num = numChars-starting;
		int newNumChars = numChars - num;
		if (characterNames!=null){
			characterNames = StringArray.deleteParts(characterNames, starting, num);
		}
		if (footnotes!=null){
			String[][] newFootnotes = new String[newNumChars][numTaxa];
			for (int j = 0; j<numTaxa; j++){
				for (int i=0; i<starting; i++) {
					newFootnotes[i][j] = footnotes[i][j];
				}
				for (int i=starting+num; i<footnotes.length; i++) {
					newFootnotes[i -num][j] = footnotes[i][j];
				}
			}
			footnotes = newFootnotes;
		}
		if (cellObjects.size()>0){
			for (int k =0; k<cellObjects.size(); k++){
				Object2DArray objArray = (Object2DArray)cellObjects.elementAt(k);
				Object[][] oldObjects = objArray.getMatrix();
				Object[][] newObjects = new Object[newNumChars][numTaxa];
				for (int j = 0; j<numTaxa; j++){
					for (int i=0; i<starting; i++) {
						newObjects[i][j] = oldObjects[i][j];
					}
					for (int i=starting+num; i<footnotes.length; i++) {
						newObjects[i -num][j] = oldObjects[i][j];
					}
				}
				objArray.setMatrix(newObjects);
			}
		}
		if (cellObjectsDisplay!=null){
			boolean[][] newCOD = new boolean[newNumChars][numTaxa];
			for (int j = 0; j<numTaxa; j++){
				for (int i=0; i<starting; i++) {
					newCOD[i][j] = cellObjectsDisplay[i][j];
				}
				for (int i=starting+num; i<cellObjectsDisplay.length; i++) {
					newCOD[i -num][j] = cellObjectsDisplay[i][j];
				}
			}
			cellObjectsDisplay = newCOD;
		}
		if (changedSinceSave!=null){
			boolean[][] newCOD = new boolean[newNumChars][numTaxa];
			for (int j = 0; j<numTaxa; j++){
				for (int i=0; i<starting; i++) {
					newCOD[i][j] = changedSinceSave[i][j];
				}
				for (int i=starting+num; i<changedSinceSave.length; i++) {
					newCOD[i -num][j] = changedSinceSave[i][j];
				}
			}
			changedSinceSave = newCOD;
		}
		if (characterIllustrations!=null){
			Image[] newCharacterIllustrations = new Image[newNumChars];
			for (int i=0; i<starting; i++) {
				newCharacterIllustrations[i] = characterIllustrations[i];
			}
			for (int i=starting+num; i<characterIllustrations.length; i++) {
				newCharacterIllustrations[i-num ] = characterIllustrations[i];
			}
			characterIllustrations = newCharacterIllustrations;
		}
		numChars = newNumChars;
		super.deleteParts(starting, num); //for specs sets
		return true;
	}
	/*-----------------------------------------------------------*/
	/** Deletes characters in linked data matrices. */
	public final void deleteInLinked(int starting, int num, boolean notify){
		if (linkedDatas.size()>0){
			for (int i=0; i<linkedDatas.size(); i++){
				CharacterData d = (CharacterData)linkedDatas.elementAt(i);
				d.deleteParts(starting, num);
				if (notify)
					d.notifyListeners(this, new Notification(MesquiteListener.PARTS_DELETED, new int[] {starting, num}));
			}
		}
	}
	/*-----------------------------------------------------------*/
	/** An equivalent to moveParts but with notification added.*/
	public final boolean moveCharacters(int starting, int num, int justAfter, boolean notify){
		boolean moved = moveParts(starting, num, justAfter);
		if (moved && notify)
			notifyListeners(this, new Notification(MesquiteListener.PARTS_MOVED, new int[] {starting, num, justAfter}));
		return moved;
	}
	/*-----------------------------------------------------------*/
	/**Moves num characters from position "first" to just after position "justAfter"; returns true iff successful.*/
	public boolean moveParts(int starting, int num, int justAfter){
		StringArray.moveParts(characterNames, starting, num, justAfter); 
		StringArray.moveColumns(footnotes, starting, num, justAfter); 
		StringArray.moveParts(characterIllustrationPath, starting, num, justAfter); 
		if (cellObjects.size()>0){
			for (int k =0; k<cellObjects.size(); k++){
				Object2DArray objArray = (Object2DArray)cellObjects.elementAt(k);
				Object[][] oldObjects = objArray.getMatrix();
				Object2DArray.moveColumns(oldObjects, starting, num, justAfter);
			}
		}
		Bits.moveColumns(cellObjectsDisplay, starting, num, justAfter);
		Bits.moveColumns(changedSinceSave, starting, num, justAfter);
		MesquiteImage.moveParts( characterIllustrations, starting, num, justAfter); 
		nMove++;
		return super.moveParts(starting, num, justAfter);
	}
	/*-----------------------------------------------------------*/
	/** Move characters in linked data matrices. */
	public final void moveInLinked(int starting, int num, int justAfter, boolean notify){
		if (linkedDatas.size()>0){
			for (int i=0; i<linkedDatas.size(); i++){
				CharacterData d = (CharacterData)linkedDatas.elementAt(i);
				d.moveParts(starting, num, justAfter);
				if (notify)
					d.notifyListeners(this, new Notification(MesquiteListener.PARTS_MOVED, new int[] {starting, num, justAfter}));
			}
		}
	}
	/*-----------------------------------------------------------*/
	/**Swaps characters first and second.*/
	public boolean swapParts(int first, int second){
		StringArray.swapParts(characterNames, first, second); 
		StringArray.swapColumns(footnotes,  first, second); 
		StringArray.swapParts(characterIllustrationPath,  first, second); 
		if (cellObjects.size()>0){
			for (int k =0; k<cellObjects.size(); k++){
				Object2DArray objArray = (Object2DArray)cellObjects.elementAt(k);
				Object[][] oldObjects = objArray.getMatrix();
				Object2DArray.swapColumns(oldObjects,  first, second);
			}
		}
		Bits.swapColumns(cellObjectsDisplay,  first, second);
		Bits.swapColumns(changedSinceSave,  first, second);
		MesquiteImage.swapParts( characterIllustrations,  first, second); 
		nMove++;
		return super.swapParts( first, second);
	}

	/*-----------------------------------------------------------*/
	/** Swap characters in linked data matrices. */
	public final void swapInLinked(int first, int second, boolean notify){
		if (linkedDatas.size()>0){
			for (int i=0; i<linkedDatas.size(); i++){
				CharacterData d = (CharacterData)linkedDatas.elementAt(i);
				d.swapParts(first, second);
				if (notify)
					d.notifyListeners(this, new Notification(MesquiteListener.PARTS_MOVED));
			}
		}
	}
	/*-----------------------------------------------------------*/
	/** Has linked data matrices send out notifications of change. */
	public final void notifyInLinked(Notification notification){
		if (linkedDatas.size()>0){
			for (int i=0; i<linkedDatas.size(); i++){
				CharacterData d = (CharacterData)linkedDatas.elementAt(i);
				d.notifyListeners(this, notification);
			}
		}
	}
	/*-----------------------------------------------------------*/
	/**Adds num taxa after position "starting"; returns true iff successful.  Assumes details already handled in subclasses, and numTaxa reset there.*/
	public boolean addTaxa(int starting, int num){
		taxaIDs = taxa.getTaxaIDs();
		if (footnotes !=null) {
			footnotes = StringArray.addRows(footnotes, starting, num);
		}
		if (cellObjectsDisplay !=null) {
			cellObjectsDisplay = Bits.addRows(cellObjectsDisplay, starting, num);
		}
		if (changedSinceSave !=null) {
			changedSinceSave = Bits.addRows(changedSinceSave, starting, num);
		}
	
		if (cellObjects != null && cellObjects.size()>0){//Vector of arrays of objects that are attached to cells
			for (int k =0; k<cellObjects.size(); k++){
				Object2DArray objArray = (Object2DArray)cellObjects.elementAt(k);
				Object[][] objects = objArray.getMatrix();
				objects = Object2DArray.addRows(objects, starting, num);
				objArray.setMatrix(objects);
			}
		}
		return true;
	}
	
	/*-----------------------------------------------------------*/
	/**Deletes num taxa from position "starting"; returns true iff successful.  Assumes details already handled in subclasses, and numTaxa reset there.*/
	public boolean deleteTaxa(int starting, int num){
		taxaIDs = taxa.getTaxaIDs();
		if (footnotes !=null) {
			footnotes = StringArray.deleteRows(footnotes, starting, num);
		}
		if (cellObjectsDisplay !=null) {
			cellObjectsDisplay = Bits.deleteRows(cellObjectsDisplay, starting, num);
		}
		if (changedSinceSave !=null) {
			changedSinceSave = Bits.deleteRows(changedSinceSave, starting, num);
		}
	
		if (cellObjects != null && cellObjects.size()>0){//Vector of arrays of objects that are attached to cells
			for (int k =0; k<cellObjects.size(); k++){
				Object2DArray objArray = (Object2DArray)cellObjects.elementAt(k);
				Object[][] objects = objArray.getMatrix();
				objects = Object2DArray.deleteRows(objects, starting, num);
				objArray.setMatrix(objects);
			}
		}
		return true;
	}

	/**moves num taxa from position "starting" to just after position "justAfter"; returns true iff successful.*/
	public boolean moveTaxa(int starting, int num, int justAfter){
		if (footnotes !=null) {
			StringArray.moveRows(footnotes, starting, num, justAfter);
		}
		if (cellObjectsDisplay !=null) {
			Bits.moveRows(cellObjectsDisplay, starting, num, justAfter);
		}
		if (changedSinceSave !=null) {
			Bits.moveRows(changedSinceSave, starting, num, justAfter);
		}
	
		if (cellObjects != null && cellObjects.size()>0){//Vector of arrays of objects that are attached to cells
			for (int k =0; k<cellObjects.size(); k++){
				Object2DArray objArray = (Object2DArray)cellObjects.elementAt(k);
				Object[][] objects = objArray.getMatrix();
				Object2DArray.moveRows(objects, starting, num, justAfter);
				objArray.setMatrix(objects);
			}
		}
		LongArray.moveParts(taxaIDs, starting, num, justAfter);
		return true;
	}
	/** For MesquiteListener interface; passes which object changed, along with optional integer (e.g. for character)*/
	public void changed(Object caller, Object obj, Notification notification, CommandRecord commandRec){
		if (obj == taxa) {
			if (Notification.appearsCosmetic(notification))
				return;
			int code = Notification.getCode(notification);
			int[] parameters = Notification.getParameters(notification);
			if (parameters==null)
				reconcileTaxa(code, commandRec);
			else {
				try {
					if (code== MesquiteListener.PARTS_ADDED)
						addTaxa(parameters[0],parameters[1]);
					else if (code== MesquiteListener.PARTS_DELETED)
						deleteTaxa(parameters[0],parameters[1]);
					else if (code== MesquiteListener.PARTS_MOVED)
						moveTaxa(parameters[0],parameters[1], parameters[2]);
					notifyListeners(this, new Notification(MesquiteListener.PARTS_CHANGED), commandRec);
				}
				catch (ArrayIndexOutOfBoundsException e){}
			}
			if (basisTree!=null && basisTree instanceof MesquiteListener)
				((MesquiteListener)basisTree).changed(caller, obj, notification, commandRec);
		}
	}
	/** For MesquiteListener interface*
	public void disposing(Object obj){
		super.disposing(obj);
	}
	/** For MesquiteListener interface*/
	public boolean okToDispose(Object obj, int queryUser){
		return true;
	}
	public boolean checkTaxaIDs(CommandRecord commandRec){
		String warning = null;
		if (numTaxa != taxa.getNumTaxa()) 
			warning = "Error in CharacterData: numTaxa (" + numTaxa + ") != taxa.getNumTaxa() (" + taxa.getNumTaxa() + ") ";
		if (taxaIDs.length !=numTaxa) 
			warning = "Error in CharacterData: numTaxa (" + numTaxa + ") != taxaIDs.length (" + taxaIDs.length + ") ";
		for (int i = 0; i<taxa.getNumTaxa() && warning == null; i++)
			if (i>= taxaIDs.length || taxa.getTaxon(i).getID() != taxaIDs[i])
				warning = "Error in CharacterData: id of taxon " + i +" in Taxa doesn't match id recorded in CharacterData";
		if (warning == null)
			return true;
		MesquiteMessage.warnProgrammer(warning + " (" + this + ")");
		return false;
	}
	
	/*
	private void dumpIDs(){
		MesquiteMessage.println("local " + LongArray.toString(taxaIDs));
		String s = "[ ";
		for (int i = 0; i<taxa.getNumTaxa(); i++) //go through list of taxa
			s +=  taxa.getTaxon(i).getID() + "  ";
		MesquiteMessage.println("in taxa " + s + "]");
	}
	*/
	private void reconcileTaxa(int code, CommandRecord commandRec){
		//check id list of taxa to see that it matches; otherwise adjust matrix to match taxa sequence and presence/absence
		int newNumTaxa = taxa.getNumTaxa();
		if (newNumTaxa == numTaxa) {
			if (code== MesquiteListener.PARTS_CHANGED || code== MesquiteListener.PARTS_MOVED) {
				/*go through list of taxa.  If any taxon is not in sequence expected from Taxa then find where it is in the list of taxaID's
				and move it into place*/
				for (int i = 0; i<taxa.getNumTaxa(); i++){ //go through list of taxa
					if (taxa.getTaxon(i).getID() != taxaIDs[i]){ //taxon i is not in sequence expected from Taxa
						int loc = LongArray.indexOf(taxaIDs, taxa.getTaxon(i).getID());
						if (loc <0) {
							MesquiteTrunk.mesquiteTrunk.discreetAlert(commandRec, "Error in CharacterData: taxaID's cannot be reconciled with current Taxa");
							return;
						}
						else {
							//move taxon that should be here into this place
							moveTaxa(loc, 1, i-1);
						}
					}
				}

				//MesquiteTrunk.mesquiteTrunk.discreetAlert(commandRec, "ERROR in CharacterData: MesquiteListener.PARTS_MOVED not yet handled");
				notifyListeners(this, new Notification(MesquiteListener.PARTS_CHANGED), commandRec);
			}
			else
				checkTaxaIDs(commandRec);
		}
		else {
			long[] oldTaxaIDs = taxaIDs;
			if (code== MesquiteListener.PARTS_ADDED) {
				//cycle through finding which taxa in Taxa are not in matrix, and adding them
				for (int i=0; i<newNumTaxa; i++) {
					Taxon t = taxa.getTaxon(i);
					long tid = t.getID();
					if (LongArray.indexOf(oldTaxaIDs, tid)<0){
						int tN = taxa.whichTaxonNumber(t);
						addTaxa(tN-1, 1);
						//should instead find contiguous block and add all at once!
					}
				}
				
			}
			else if (code== MesquiteListener.PARTS_DELETED) {
				//cycle through finding which taxa in matrix have been deleted && deleting them from matrix
				for (int i=numTaxa-1; i>=0; i--) {
					Taxon t = taxa.getTaxonByID(oldTaxaIDs[i]);
					if (t==null) {
						deleteTaxa(i, 1);
						//should instead find contiguous block and delete all at once!
					}
				}
			}
			else {
				//cycle through finding which taxa deleted && deleting them from matrix
				for (int i=numTaxa-1; i>=0; i--) {
					Taxon t = taxa.getTaxonByID(oldTaxaIDs[i]);
					if (t==null) {
						deleteTaxa(i, 1);
						//should instead find contiguous block and delete all at once!
					}
				}
				//cycle through finding which taxa in Taxa are not in matrix, and adding them
				for (int i=0; i<newNumTaxa; i++) {
					Taxon t = taxa.getTaxon(i);
					long tid = t.getID();
					if (LongArray.indexOf(oldTaxaIDs, tid)<0){
						int tN = taxa.whichTaxonNumber(t);
						addTaxa(tN-1, 1);
						//should instead find contiguous block and add all at once!
					}
				}
			}
			taxaIDs = taxa.getTaxaIDs();
			checkTaxaIDs(commandRec);
			notifyListeners(this, new Notification(MesquiteListener.PARTS_CHANGED), commandRec);
		}
	}
	
	
   	/** extracts data from matrix for character ic, and returns it in a CharacterDistribution object.*/
   	public abstract CharacterDistribution getCharacterDistribution(int ic); //TODO: should pass a CharacterDistribution to save the instantiation
   	/** extracts data from matrix, and returns it in a MCharactersDistribution object.*/
   	public abstract MCharactersDistribution getMCharactersDistribution();
   	/** Fills matrix with data from passed MCharactersDistribution object.*/
   	public abstract void setMatrix(MCharactersDistribution states);
   	/** creates an empty CharacterState object of the same data type as CharacterData subclass used.*/
   	public abstract CharacterState makeCharacterState();
   	/** creates an empty CharacterDistribution object of the same data type as CharacterData subclass used.*/
   	public abstract CharacterDistribution makeCharacterDistribution();
   	/** creates an empty CharacterData object of the same data type as CharacterData subclass used.*/
   	public abstract CharacterData makeCharacterData();
   	/** creates an empty CharacterData object of the same data type as CharacterData subclass used, based on CharMatrixManager & Taxa passed.*/
   	public abstract CharacterData makeCharacterData(CharMatrixManager manager, Taxa taxa); 
   	/** creates a CharacterData object of the same data type as CharacterData subclass used, with ntaxa and nchars.*/
   	public abstract CharacterData makeCharacterData(int ntaxa, int nchars);
   	/** trades the states of character ic and ic2 in taxon it.  Used for reversing sequences (for example).*/
	public abstract void tradeStatesBetweenCharacters(int ic, int ic2, int it);
   	/** trades the states of character ic between taxa it and it2.  Used for reshuffling (for example).*/
	public abstract void tradeStatesBetweenTaxa(int ic, int it, int it2);
	
   	/** returns whether the matrix would prefer to have columns sized individually in editors.  Default is true.*/
	public boolean pleaseAutoSizeColumns() {
		return columnWidthAutoSize;
	}
   	/** Sets whether to autoSize columns, if this is setable.*/
	public void setAutoSizeColumns(boolean autoSize) {
		columnWidthAutoSize = autoSize;
	}
   	/** returns default column width in editors.  Default is 16.*/
	public int getDefaultColumnWidth() {
		return 16;
	}

   	/** appends to buffer string describing the state(s) of character ic in taxon it.*/
	public abstract void statesIntoStringBuffer(int ic, int it, StringBuffer sb, boolean forDisplay, boolean includeInapplicable, boolean includeUnassigned);
   	/** appends to buffer string describing the state(s) of character ic in taxon it.*/
	public abstract void statesIntoStringBuffer(int ic, int it, StringBuffer sb, boolean forDisplay);
   	/** appends to buffer string describing the state(s) of character ic in taxon it.*/
	public abstract void statesIntoNEXUSStringBuffer(int ic, int it, StringBuffer sb);
	/**Set the state at character ic and taxon it from the string in the parser, beginning at current parser position in the string. 
	Updates current position in string.  If fromEditor is true, should assume whole string is state. Returns a result code (OK, ERROR, EOL, COMMENT).
	EOL is returned if data are interleaved and end of line found.  If there is an error or a comment, an error message or the comment,
	respectively, will be returned in the result MesquiteString.*/
   	public abstract int setState(int ic, int it, Parser parser, boolean fromEditor, MesquiteString result);
   	/** sets the state of character ic in taxon it from CharacterState cs*/
   	public abstract void setState(int ic, int it, CharacterState cs);
   	/** returns whether the character ic is inapplicable to taxon it*/
   	public abstract boolean isInapplicable(int ic, int it);
   	/** returns whether the state of character ic is missing in taxon it*/
   	public abstract boolean isUnassigned(int ic, int it);
   	/** returns whether the character ic is included (i.e. not currently excluded)*/
   	public boolean isCurrentlyIncluded(int ic){
   		CharInclusionSet incl = (CharInclusionSet)getCurrentSpecsSet(CharInclusionSet.class);
   		return (incl==null || incl.isSelected(ic));
   			
   	}
   	/** returns the state of character ic in taxon it*/
   	public abstract CharacterState getCharacterState(CharacterState cs, int ic, int it);
   	/** sets the symbol used for inapplicable character (e.g., gap)*/
   	public void setInapplicableSymbol(char inapp) {
		setDirty(true);
   		inapplicableChar = inapp;
   	}
   	/** returns symbol used for inapplicable character (e.g., gap)*/
   	public char getInapplicableSymbol() {
   		return inapplicableChar;
   	}
    	/** sets the symbol used for missing data (unassigned) character*/
   	public void setUnassignedSymbol(char mc) { 
		setDirty(true);
   		missingChar = mc;
   	}
  	/** returns symbol used for missing data (unassigned) character*/
   	public char getUnassignedSymbol() { 
   		return missingChar;
   	}
    	/** sets the symbol used for matchChar*/
   	public void setMatchChar(char mc) { 
   		matchChar = mc;
   	}
  	/** returns symbol used for matchChar*/
   	public char getMatchChar() { 
   		return matchChar;
   	}
   	
  	/** returns whether cells are to be colored by default*/
   	public boolean colorCellsByDefault(){
   		return false;
   	}
  	/** returns the color of character ic; e.g., to indicate codon positions */
   	public Color getDefaultCharacterColor(int ic){
		return null;
   	}
 	/** returns the dark color of character ic; e.g., to indicate codon positions */
   	public Color getDarkDefaultCharacterColor(int ic){
   	 	return null;
   	}
    	/** returns a String summarizing the states of a character (e.g., "2 states", "0.1-0.9").*/
	public abstract String getStatesSummary(int ic);
	
	/*.................................................................................................................*/
	/** Indicates whether the data are molecular sequence data or not */ 
	public boolean isMolecularSequence() {
		return false;
	}
	
	/** Indicates the type of character stored */ 
	public abstract Class getStateClass();
	/** Gets the color representing state(s) of character ic in taxon it */ 
	public abstract Color getColorOfStates(int ic, int it);
	
 	/*.................................................................................................................*/
	/** returns the default character model for the paradigm (e.g., "parsimony") given in the String */ 
   	public abstract CharacterModel getDefaultModel(String paradigm);
 	/*.................................................................................................................*/
 	/** returns a list of names of the characters (StringLister interface)*/
	public String[] getStrings(){
		String[] s = new String[numChars];
		for (int i=0; i<numChars; i++){
			if (characterNames == null ||  characterNames[i]==null)
				s[i]= "Character " + CharacterStates.toExternal(i); 
			else
				s[i]= Integer.toString(CharacterStates.toExternal(i)) + ". " + characterNames[i]; 
		}
		return s;
	}
 	/*.................................................................................................................*/
 	/** returns whether whether any characters have names*/
	public boolean characterNamesExist() {
		if (characterNames == null)
			return false;
		for (int ic=0; ic<numChars; ic++) {
			if (characterNames[ic]!=null)
				return true; 
		}
		return false;
	}
 	/*.................................................................................................................*/
 	/** returns whether character ic has a name*/
	public boolean characterHasName(int ic) {
		if (ic<0 || ic>=numChars) {
			return false; 
		}
		else if (characterNames == null ||  characterNames[ic]==null)
			return false; 
		else
			return true;
	}
 	/*.................................................................................................................*/
 	/** returns the name of character ic*/
	public String getCharacterName(int ic) {
		if (ic<0 || ic>=numChars) {
			MesquiteMessage.println("Error: character number out of bounds (getCharacterName) " + ic);
			return ""; 
		}
		else if (characterNames == null ||  characterNames[ic]==null)
			return "Character " + CharacterStates.toExternal(ic); 
		else
			return characterNames[ic];
	}
 	/*.................................................................................................................*/
 	/** sets the name of character ic*/
	public void setCharacterName(int ic, String name) {
		if (ic<0 || ic>=numChars) {
			MesquiteMessage.println("Error: character number out of bounds (setCharacterName) " + ic);
		}
		else if (characterNames == null)
			;
		else {
			try {
				int i = Integer.parseInt(name);
				name = "#" + name;  //autoadjust to add character
			}
			catch (NumberFormatException e){
			}
			if (StringUtil.blank(name))
				name = null;
			characterNames[ic]= name;
			notifyListeners(this, new Notification(NAMES_CHANGED, new int[] {ic}));
		}
	}
	/*.................................................................................................................*/
	public String checkNameLegality(int it, String s){
		if (s==null) {
			return null; //characters can have null names
		}
		try {
			int i = Integer.parseInt(s);
			return "The character name \"" + s + "\" is illegal because it consists only of numbers";
		}
		catch (NumberFormatException e){
		}
		return null;
	}
 	/*.................................................................................................................*/
 	/** sets the annotation (footnote) of character ic, taxon it*/
	public void setAnnotation(int ic, int it, String expl) {
		if (ic<0 || ic>=numChars || it<0 || it>=numTaxa) {
			MesquiteMessage.println("Error: character or taxon number out of bounds (setAnnotation) " + ic);
		}
		else {
			if (footnotes == null){
				footnotes = new String[numChars][numTaxa];
			}
			if (!(ic<0 || ic>=footnotes.length || it<0 || footnotes[ic]==null || it>=footnotes[ic].length)){
				if (expl == null)
					footnotes[ic][it]= null;
				else
					footnotes[ic][it]= new String(expl);
				notifyListeners(this, new Notification(ANNOTATION_CHANGED));
			}
		}
	}
 	/*.................................................................................................................*/
 	/** gets the annotation (footnote) of character ic, taxon it*/
	public String getAnnotation(int ic, int it) {
		if (ic<0 || ic>=numChars || it<0 || it>=numTaxa) {
			return null;
		}
		else {
			if (footnotes == null)
				return null;
			if (ic<0 || ic>=footnotes.length || it<0 || footnotes[ic]==null || it>=footnotes[ic].length){
				return null;
			}
			return footnotes[ic][it];
		}
	}
 	/*.................................................................................................................*/
 	/** returns whether the matrix cell has been changed since the last file save*/
	public boolean getChangedSinceSave(int ic, int it) {
		if (ic<0 || ic>=numChars || it<0 || it>=numTaxa) {
			return false;
		}
		else if (changedSinceSave!=null && !(ic<0 || ic>=changedSinceSave.length || it<0 || changedSinceSave[ic]==null || it>=changedSinceSave[ic].length))
			return changedSinceSave[ic][it];
		return false;
	}
 	/*.................................................................................................................*/
 	/** resets to false for all cells whether the matrix cell has been changed since the last file save*/
	public void resetChangedSinceSave() {
		boolean c = false;
		if (changedSinceSave!=null)
			for (int ic = 0; ic<changedSinceSave.length; ic++)
				for (int it=0; it<changedSinceSave[ic].length; it++){
					if (changedSinceSave[ic][it]) {
						changedSinceSave[ic][it] = false;
						c = true;
					}
				}
		watchForChange = true;	
		checksumValid = false;
		getChecksum();
		if (c)
			notifyListeners(this, new Notification(MesquiteListener.ANNOTATION_CHANGED));
	}
 	/*.................................................................................................................*/
	public Vector getCellObjectsVector(){
		return cellObjects;
	}
 	/*.................................................................................................................*/
	public Object2DArray getWhichCellObjects(NameReference nRef){
		if (cellObjects!=null && nRef!=null) {
			for (int i=0; i<cellObjects.size(); i++) {
				Object2DArray b = (Object2DArray)cellObjects.elementAt(i);
				if (b !=null && nRef.equals(b.getNameReference())) {
					return b; 
				}
			}
		}
		return null;
	}
 	/*.................................................................................................................*/
 	/** sets whether something drawing the matrix should check for special symbols for cell ic, it*/
	public void setCellObjectDisplay(int ic, int it) {
		if (ic<0 || ic>=numChars || it<0 || it>=numTaxa) {
			return;
		}
		else if (cellObjectsDisplay!=null && !(ic<0 || ic>=cellObjectsDisplay.length || it<0 || cellObjectsDisplay[ic]==null || it>=cellObjectsDisplay[ic].length)){
			cellObjectsDisplay[ic][it] = true;
		}
	}
 	/*.................................................................................................................*/
 	/** returns whether something drawing the matrix should check for special symbols for cell ic, it*/
	public boolean getCellObjectDisplay(int ic, int it) {
		if (ic<0 || ic>=numChars || it<0 || it>=numTaxa) {
			return false;
		}
		else if (cellObjectsDisplay!=null && !(ic<0 || ic>=cellObjectsDisplay.length || it<0 || cellObjectsDisplay[ic]==null || it>=cellObjectsDisplay[ic].length))
			return cellObjectsDisplay[ic][it];
		return false;
	}
 	/*.................................................................................................................*/
 	/** gets the object of type "name" attached to character ic, taxon it*/
	public void setCellObject(NameReference nr, int ic, int it, Object obj) {
		if (ic<0 || ic>=numChars || it<0 || it>=numTaxa) {
			return;
		}
		else {
			Object2DArray array = getOrMakeCellObjects(nr);
			if (array !=null) {
				array.setValue(ic, it, obj);
			}
		}
	}
 	/*.................................................................................................................*/
 	/** gets the object of type "name" attached to character ic, taxon it*/
	public Object getCellObject(NameReference nr, int ic, int it) {
		if (ic<0 || ic>=numChars || it<0 || it>=numTaxa) {
			return null;
		}
		else {
			Object2DArray array = getWhichCellObjects(nr);
			if (array!=null)
				return array.getValue(ic, it);
		}
		return null;
	}
 	/*.................................................................................................................*/
 	/** gets the object of type "name" attached to character ic, taxon it*/
	public Object2DArray getOrMakeCellObjects(NameReference nr) {
		Object2DArray array = getWhichCellObjects(nr);
		if (array==null) {
			array = new Object2DArray(numChars, numTaxa);
			array.setNameReference(nr);
			cellObjects.addElement(array);
		}
		return array;
	}
	
	boolean watchForChange = false;
 	/*.................................................................................................................*/
 	NameReference historyRef = NameReference.getNameReference("ChangeHistory");
 	/** Marks the data for character ic, taxon it as having changed*/
	public void stampHistoryChange(int ic, int it) {
		if (ic<0 || ic>=numChars || it<0 || it>=numTaxa) { //|| MesquiteTrunk.currentAuthor == null
			return;
		}
		else {
			if (changedSinceSave!=null && !(ic<0 || ic>=changedSinceSave.length || it<0 || changedSinceSave[ic]==null || it>=changedSinceSave[ic].length))
				changedSinceSave[ic][it] = true;
			 checksumValid = false;
			//if (watchForChange)
				//MesquiteMessage.warnProgrammer("SHC " + ic + " " + it);
			watchForChange = false;
			if (saveChangeHistory){ //|| MesquiteTrunk.currentAuthor == null
				ChangeHistory history = (ChangeHistory)getCellObject(historyRef, ic, it);
				ChangeEvent ce = null;
				StringBuffer sb = new StringBuffer();
				statesIntoStringBuffer(ic, it, sb, false);
				if (history != null) {
					if (!history.timeExists(MesquiteTrunk.startupTime)) {
						ce = history.addEvent(0, MesquiteTrunk.startupTime, sb.toString()); //MesquiteTrunk.currentAuthor
					}
					else {
						ce = history.updateEvent(0, MesquiteTrunk.startupTime, sb.toString()); //MesquiteTrunk.currentAuthor
					}
				}
				else {
					ChangeHistory cH = new ChangeHistory();
					ce = cH.addEvent(0, MesquiteTrunk.startupTime, sb.toString());
					setCellObject(historyRef, ic, it, cH);
				}
				if (ce !=null && requireChangeAuthority){ 
					if (currentChangeAuthority == null){//no ChangeAuthority on record
						String s = MesquiteString.queryString(MesquiteWindow.activeWindow, "Authority", "On what authority do you change these data?", "I said so");
						ChangeAuthority ca = new ChangeAuthority();
						if (s == null)
							s = "I said so";
						ca.set(s);
						currentChangeAuthority = ca; //this is to be reset to null before an independent change is made.
									//NOTE:  it is assumed this change will be fully done when notifyListeners is called
									//(Such a convention is a bit delicate, but seems best at moment)
					}
					ce.setAuthority(currentChangeAuthority);
				}
			}
		}
	}
	public void stampHistoryChange() {
		if (changedSinceSave!=null)
			for (int ic = 0; ic<getNumChars(); ic++)
				for (int it = 0; it<getNumTaxa(); it++)
					changedSinceSave[ic][it] = true;
	}
	
 	/*.................................................................................................................*/
	public void doAfterNotify(Notification notification, CommandRecord commandRec){
		currentChangeAuthority = null;
	}
 	/*.................................................................................................................*/
	public long getChecksum(){
		if (checksumValid)
			return checksum;
		checksum = calculateChecksum(crc32);
		checksumValid = true;
		return checksum;
	}
 	/*.................................................................................................................*/
	public abstract long calculateChecksum(CRC32 crc32);
 	/*.................................................................................................................*/
	protected void setDirty(boolean d, int ic, int it){
		setDirty(d); 
		stampHistoryChange(ic, it);
	}
 	/*.................................................................................................................*/
 	/** gets the explanation of this matrix*/
	public String getExplanation() {
		if (taxa == null)
			return null;
		String extra = "This character matrix for the taxa block \"" + taxa.getName() + "\" has " + getNumChars() + " characters for the " + getNumTaxa() + " taxa. Category of data: " + getDataTypeName() + "\n";
		return extra;

	}
 	/*.................................................................................................................*/
 	/** sets the illustration tied to character ic to the given Image at the given path (URL path or file path)*/
	public void setIllustration(int ic, Image illustration, String path) {
		if (ic<0 || ic>=numChars) {
			MesquiteMessage.println("Error: character number out of bounds (setIllustration) " + ic);
		}
		else {
			if (characterIllustrations==null) {
				characterIllustrations = new Image[numChars];
				characterIllustrationPath = new String[numChars];
			}
			characterIllustrations[ic] = illustration;
			characterIllustrationPath[ic] = path;
		}
	}
 	/*.................................................................................................................*/
 	/** returns the path of the illustration tied to character ic*/
	public String getIllustrationPath(int ic) {
		if (ic<0 || ic>=numChars) {
			MesquiteMessage.println("Error: character number out of bounds (getIllustrationPath) " + ic);
			return null;
		}
		else if (characterIllustrationPath != null)
			return characterIllustrationPath[ic] ;
		else
			return null;
	}
 	/*.................................................................................................................*/
 	/** returns the illustration tied to character ic*/
	public Image getIllustration(int ic) {
		if (ic<0 || ic>=numChars) {
			MesquiteMessage.println("Error: character number out of bounds (getIllustration) " + ic);
			return null;
		}
		else if (characterIllustrations != null)
			return characterIllustrations[ic] ;
		else
			return null;
	}
 	/*.................................................................................................................*/
 	/** Returns a string describing the contents of the cells (by default, null, but may be overridden to convey
 	special information, such as what are the items in a continuous data matrix)*/
	public String getCellContentsDescription(){
		return null;
	}
 	/*.................................................................................................................*/
 	/** outputs to log file the matrix*/
	public abstract void logMatrix();
 	/*.................................................................................................................*/
 	/*(LOCKING SYSTEM NOT READY)
 	 // locks data
	public void lockData() {
		locked = true;
	}
 	//unlocks data
	public void unlockData() {
		locked = false;
	}
 	/*.................................................................................................................*/
	private boolean addDataLinkage(CharacterData other){
		if (other==null)
			return false;
		if (other.getTaxa()!=getTaxa())
			return false;
		if (other.getNumChars() != numChars)
			return false;
		if (other == this)
			return false;
		if (linkedDatas.indexOf(other)>=0)
			return false;
		linkedDatas.addElement(other);
		return true;
	}
	private void removeDataLinkage(CharacterData other){
		if (other==null)
			return;
		if (linkedDatas == null)
			return;
		linkedDatas.removeElement(other);
	}
	/** Adds the passed Character data into this's linkage group */
	public void addToLinkageGroup(CharacterData other){
		if (other==null)
			return;
		if (linkedDatas == null)
			return;
		other.getDataLinkages().removeAllElements();
		//mutually link other to all of linked ones
		for (int i=0; i<linkedDatas.size(); i++){
			CharacterData d= (CharacterData)linkedDatas.elementAt(i);
			other.addDataLinkage(d);
			d.addDataLinkage(other);
		}
		//mutually link other to this
		addDataLinkage(other);
		other.addDataLinkage(this);
	}
	/** Removes this Character data from its linkage group */
	public void resignFromLinkageGroup(){
		if (linkedDatas == null)
			return;
		//cut out of all linked ones
		for (int i=0; i<linkedDatas.size(); i++){
			CharacterData d= (CharacterData)linkedDatas.elementAt(i);
			d.removeDataLinkage(this);
		}
		linkedDatas.removeAllElements();
	}
	public Vector getDataLinkages(){
		return linkedDatas;
	}
	public boolean isLinked(CharacterData other){
		return (other!=null && linkedDatas.indexOf(other)>=0);
	}
	public boolean isLinked(){
		return (linkedDatas.size()>0);
	}
	
	public void setBasisTree(Tree tree){
		if (tree == null) {
			basisTree = null;
		}
		else {
			basisTree = tree.cloneTree();
		}
	}
	public Tree getBasisTree(){
		return basisTree;
	}
	/* ---------------- for HNode interface ----------------------*/
	public Image getHImage(){
		return null;
	}
	/* ---------------- for HNode interface ----------------------*/
	public Color getHColor(){
		MesquiteFile file = getFile();
		if (file!=null) {
			if (file.getProject()!=null)
			return ColorDistribution.project[file.getProject().getProjectColor()];
		}
		return ColorDistribution.light[0];  //project color
	}
	/* ---------------- for HNode interface ----------------------*/
	public boolean getHShow(){
		return true; 
	}
	/* ---------------- for use with touched from HNode interface ----------------------*/
	public void addToBrowserPopup(MesquitePopup popup){
		super.addToBrowserPopup(popup);
		ElementManager manager = getManager();
		if (manager!=null && manager instanceof MesquiteModule) {
			MesquiteFile file = getFile();
			if (file!=null && file.getProject()!=null) {
				popup.add(new MenuItem("-"));
				popup.add(new MesquiteMenuItem("Show List of Characters of \"" + getName() + "\"", MesquiteTrunk.mesquiteTrunk, MesquiteTrunk.mesquiteTrunk.makeCommand("showMe", this)));
				popup.add(new MesquiteMenuItem("Show Data Editor for \"" + getName() + "\"", MesquiteTrunk.mesquiteTrunk, MesquiteTrunk.mesquiteTrunk.makeCommand("showDataWindow", (MesquiteModule)manager), Integer.toString(file.getProject().getMatrixNumber(this))));
			}
		}
	}
}

