/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.lib.duties;

import java.awt.*;
import mesquite.lib.*;
import mesquite.lib.characters.*;


/* ======================================================================== */
/** Manages character data matrices of particular types.  Also reads and writes parts of CHARACTERS NEXUS block.  Different data
types have their own managers, so that the coordinating CharactersManager doesn't need to know about the details.
Example modules: "Manage Categorical char. matrices" (class ManageCategoricalChars); "Manage DNA/RNA matrices" (class ManageDNARNAChars) 
"Manage Continuous char. matrices" (class ManageContChars)*/

public abstract class CharMatrixManager extends MesquiteModule   {
	boolean logVerbose = true;

   	 public Class getDutyClass() {
   	 	return CharMatrixManager.class;
   	 }
 	public String getDutyName() {
 		return "Manager of Character data matrix, including read/write CHARACTERS block";
   	 }
	/** Process format command of NEXUS CHARACTERS block. Return the resulting CharacterData object*/
	public abstract CharacterData processFormat(MesquiteFile file, Taxa taxa, String dataType, String formatCommand, MesquiteInteger stringPos, int numChars);
	/** Process command in CHARACTERS block other than FORMAT or MATRIX command (e.g., CHARSTATELABELS command).  Should return true if processed
	successfully, false if not*/
	public abstract boolean processCommand(CharacterData data, String commandName, String commandString);
	/** return string consisting of CHARACTERS block for NEXUS file.*/
	public abstract void writeCharactersBlock(CharacterData data, CharactersBlock cB, MesquiteFile file, ProgressIndicator progIndicator);
	/** return true if can read the data type passed (token from NEXUS file is the string passed, e.g. "STANDARD", "CONTINUOUS") */
	public abstract boolean readsWritesDataType(String dataType);
	/** return true if can read the data class passed, e.g. CategoricalData.class) */
	public abstract boolean readsWritesDataType(Class dataClass);
	/** Return name of data class managed, for listing in menus and so on (e.g. "DNA data" .*/
	public abstract String getDataClassName();
	/** Returns data class managed, e.g. ContinuousData.*/
	public abstract Class getDataClass();
	/** return a new CharacterData object for the taxa indicated.*/
	public abstract CharacterData getNewData(Taxa taxa, int numChars);
	
//	public abstract void processMatrix(Taxa taxa, CharacterData data, String matrixString, boolean nameTaxa);
	
	/*.................................................................................................................*/
	public boolean isLogVerbose() {	
		return logVerbose;
	}
	/*.................................................................................................................*/
	public void setLogVerbose(boolean logVerbose) {	
		this.logVerbose = logVerbose;
	}
		/*.................................................................................................................*/
	/** Process the matrix, placing data into passed CharacterData object */
	public void processMatrix(Taxa taxa, CharacterData data, String matrixString, boolean nameTaxa) {
		String charstring;
		String taxonName;
		parser.getFirstToken(matrixString);
		if (data.interleaved) {
			int[] currentCharacter = new int[taxa.getNumTaxa()];
			for (int i=0; i<taxa.getNumTaxa(); i++) currentCharacter[i] =0;
			boolean done = false;
			parser.setLineEndingsDark(false);
			int it=0;
			while (!done && !(taxonName=parser.getNextToken()).equals(";")) {
				parser.setLineEndingsDark(true);
				if (nameTaxa && it<taxa.getNumTaxa())
					taxa.setTaxonName(it,taxonName);
				int whichTaxon = taxa.whichTaxonNumber(taxonName);
				if (whichTaxon == -1) {
					MesquiteMessage.warnUser("Unrecognized taxon name: " + taxonName + " in matrix ");
					whichTaxon = it % taxa.getNumTaxa();
				}
				boolean polymorphOn = false;
				boolean uncertainOn = false;
				int response = CharacterData.OK;
				while (currentCharacter[whichTaxon]<data.getNumChars() && response!=CharacterData.EOL) {
					response = data.setState(currentCharacter[whichTaxon], whichTaxon, parser, false, null);
					if (response !=CharacterData.EOL)
						currentCharacter[whichTaxon]++;
				}
				if (it==0) 
					data.interleavedLength=currentCharacter[0];
				done = true;
				for (int ic=0; ic<currentCharacter.length; ic++) 
					if (currentCharacter[ic] != data.getNumChars())
						done = false;
				parser.setLineEndingsDark(false);
				it++;
			}
			parser.setLineEndingsDark(false);
		}
		else {
			for (int it=0; it<taxa.getNumTaxa() && !(taxonName=parser.getNextToken()).equals(";"); it++) {
				if (nameTaxa)
					taxa.setTaxonName(it,taxonName);
				int whichTaxon = taxa.whichTaxonNumber(taxonName);
				if (whichTaxon == -1) {
					MesquiteMessage.warnUser("Unrecognized taxon name: " + taxonName + " in matrix ");
					whichTaxon = it;
				}
				int ic=0;
				boolean polymorphOn = false;
				boolean uncertainOn = false;
				while (ic<data.getNumChars()) {
					int response = data.setState(ic++, whichTaxon, parser, false, null);
				}
			}
		}
	}

}

