/* Mesquite source code, Treefarm package.  Copyright 1997-2004 W. Maddison, D. Maddison and P. Midford. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.treefarm.TerminalInstability;
/*~~  */

import java.util.*;
import java.awt.*;
import mesquite.lib.*;
import mesquite.lib.duties.*;

/** ======================================================================== */
/*
Terminal instability -- the difference in the position of a particular target taxon is measured by the 
sum of the weighted differences between the trees in path lengths between that taxon and each of the other taxa.  
The path length between taxa is the number of nodes intervening on the path between the one taxon and the other 
(branch lengths not considered, hence scaling not yet needed). 
Thus if on tree 1 the path length between the target taxon and another taxon i is pti1 and the length on tree 2 is pti2, 
then the weighted difference is Math.abs(pti1 - pti2)/sqr(pti1 + pti2).  The weighting by sqr(pti1 + pti2) 
means that the emphasis on another taxon i fades away as it is farther from the target taxon in on or the other tree.  
This weighted difference is summed over all other taxa i to get the difference in the target taxon's position in the two trees.

*/

public class TerminalInstability extends NumberForTaxon {
	Taxa currentTaxa;
	MesquiteString treeSourceName;
	MesquiteCommand cstC;
	TreeSource treeSourceTask;
	boolean needsRecalculation = true;
	NumberArray results;
	MesquiteString resultsString;
	MesquiteMenuItemSpec numTreesItem = null;
	boolean numTreesSet = false;
	int defNumTrees = 100;
	int skipTrees = 1;
	double exponent = 2.0;
	/*.................................................................................................................*/
	public boolean startJob(String arguments, Object condition, CommandRecord commandRec, boolean hiredByName) {
 		treeSourceTask = (TreeSource)hireEmployee(commandRec, TreeSource.class,  "Source of trees for comparison by for " + getName());
 		if (treeSourceTask == null) {
 			return sorry(commandRec, getName() + " couldn't start because no source of trees obtained");
 		}
		treeSourceName = new MesquiteString(treeSourceTask.getName());
		

		cstC =  makeCommand("setTreeSource",  this);
		treeSourceTask.setHiringCommand(cstC);
		if (numModulesAvailable(TreeSource.class)>1){ 
			MesquiteSubmenuSpec mss = addSubmenu(null, "Tree source",cstC, TreeSource.class);
			mss.setSelected(treeSourceName);
		}
		numTreesItem = addMenuItem("Number of Trees (for instability)...", makeCommand("setNumTrees", this));
		numTreesItem.setEnabled(false);
		addMenuItem("Skip trees (instability)...", makeCommand("setSkipping", this));
		addMenuItem("Exponent for Instability...", makeCommand("setExponent", this));
		results = new NumberArray(1);
		resultsString = new MesquiteString();
		needsRecalculation = commandRec.scripting();
		return true;  
 	}
 	
	/*.................................................................................................................*/
	/** returns whether this module is requesting to appear as a primary choice */
   	public boolean requestPrimaryChoice(){
   		return true;  
   	}
 	public void employeeQuit(MesquiteModule m){
 		if (m!=treeSourceTask)
 			iQuit();
 	}
	/*.................................................................................................................*/
   	 public void employeeParametersChanged(MesquiteModule employee, MesquiteModule source, Notification notification, CommandRecord commandRec) {
			needsRecalculation = true;
			if (!commandRec.scripting())
				super.employeeParametersChanged(employee, source, notification, commandRec);
   	 }
	/*.................................................................................................................*/
  	 public Snapshot getSnapshot(MesquiteFile file) {
   	 	Snapshot temp = new Snapshot();
  	 	temp.addLine("setTreeSource " , treeSourceTask);
  	 	temp.addLine("setNumTrees " + defNumTrees);
  	 	temp.addLine("setSkipping " + skipTrees);
  	 	temp.addLine("setExponent " + exponent);
  	 	temp.addLine("doCalc");
  	 	return temp;
  	 }
  	 MesquiteInteger pos = new MesquiteInteger(0);
	/*.................................................................................................................*/
    	 public Object doCommand(String commandName, String arguments, CommandRecord commandRec, CommandChecker checker) {
    	 	if (checker.compare(this.getClass(), "Sets the source of trees for comparison", "[name of module]", commandName, "setTreeSource")) {
    	 		TreeSource temp =  (TreeSource)replaceEmployee(commandRec, TreeSource.class, arguments, "Source of trees for comparison by for Terminal Instability", treeSourceTask);
 			if (temp!=null) {
	    	 		treeSourceTask = temp;
				treeSourceTask.setHiringCommand(cstC);
				treeSourceName.setValue(treeSourceTask.getName());
				treeSourceTask.initialize(currentTaxa, commandRec);
				numTreesSet = false;
				if (!commandRec.scripting()) {
					needsRecalculation = true;
					parametersChanged(null, commandRec);
				}
 			}
 			return temp;
    	 	}

    	 	else if (checker.compare(this.getClass(), "Sets the number of trees", "[number]", commandName, "setNumTrees")) {
			int newNum= MesquiteInteger.fromFirstToken(arguments, pos);
			if (!MesquiteInteger.isCombinable(newNum))
				newNum = MesquiteInteger.queryInteger(containerOfModule(), "Set Number of Trees", "Number of Trees (for " + getName() +  "):", defNumTrees, 0, MesquiteInteger.infinite);
    	 		if (newNum>0  && newNum!=defNumTrees) {
    	 			defNumTrees = newNum;
    	 			numTreesSet = true;
				if (!commandRec.scripting()) {
					needsRecalculation = true;
					parametersChanged(null, commandRec);
				}
    	 		}
    	 		else if (defNumTrees == newNum)
    	 			numTreesSet = true;
    	 	}
    	 	else if (checker.compare(this.getClass(), "Sets whether trees are skipped (every nth tree compared)", "[n]", commandName, "setSkipping")) {
			int newNum= MesquiteInteger.fromFirstToken(arguments, pos);
			if (!MesquiteInteger.isCombinable(newNum))
				newNum = MesquiteInteger.queryInteger(containerOfModule(), "Set Skip", "Enter n where every nth tree is compared (for " + getName() +  ").  By entering 2 for example, every other tree is skipped. Enter 1 to indicate that every tree is examined.", skipTrees, 1, MesquiteInteger.infinite);
    	 		if (newNum>0  && newNum!=skipTrees) {
    	 			skipTrees = newNum;
				if (!commandRec.scripting()) {
					needsRecalculation = true;
					parametersChanged(null, commandRec);
				}
    	 		}
    	 	}
    	 	else if (checker.compare(this.getClass(), "Sets the exponent used in the denominator of the instability calculations", "[number]", commandName, "setExponent")) {
			pos.setValue(0);
			double newNum= MesquiteDouble.fromString(arguments, pos);
			if (!MesquiteDouble.isCombinable(newNum))
				newNum = MesquiteDouble.queryDouble(containerOfModule(), "Set exponent", "Enter the exponent used in the denominatory of the instability calculations.  Default is 2.  Larger exponents emphasize near neighbors more strongly relative to distant taxa.", exponent, 0.9999, 4);
    	 		if (newNum>0  && newNum!=exponent) {
    	 			exponent = newNum;
				if (!commandRec.scripting()) {
					needsRecalculation = true;
					parametersChanged(null, commandRec);
				}
    	 		}
    	 	}
    	 	else if (checker.compare(this.getClass(), "Requests calculations", null, commandName, "doCalc")) {
			needsRecalculation = true;
			parametersChanged(null, commandRec);
    	 	}
    	 	else
    	 		return super.doCommand(commandName, arguments, commandRec, checker);
    	 	return null;
   	 }
   	 private int getNumTrees(CommandRecord commandRec){
   	 	if (commandRec.scripting() || numTreesSet)
   	 		return defNumTrees;
		int newNum = MesquiteInteger.queryInteger(containerOfModule(), "Set Number of Trees", "Number of Trees (for " + getName() + "):", defNumTrees, 0, MesquiteInteger.infinite);
 		if (newNum>0) {
 			defNumTrees = newNum;
 			numTreesSet = true;
 		}
 		return defNumTrees;
   	 }
	/*.................................................................................................................*/
   	/** Called to provoke any necessary initialization.  This helps prevent the module's intialization queries to the user from
   	happening at inopportune times (e.g., while a long chart calculation is in mid-progress)*/
   	public void initialize(Taxa taxa, CommandRecord commandRec){
   		currentTaxa = taxa;
   		needsRecalculation = true;
   		treeSourceTask.initialize(currentTaxa, commandRec);
   	}
	/*.................................................................................................................*/
	public void calculateNumber(Taxon taxon, MesquiteNumber result, MesquiteString resultString, CommandRecord commandRec){
		if (result==null|| taxon == null)
			return;
		result.setToUnassigned();
		if (taxon.getTaxa()!= currentTaxa) {
			initialize(taxon.getTaxa(), commandRec);
			needsRecalculation = true;
		}
		if (needsRecalculation)
			doCalcs(commandRec);
		int it = currentTaxa.whichTaxonNumber(taxon);
		results.placeValue(it, result);
		
		if (resultString !=null) {
			resultString.setValue(resultsString.getValue());
		}
	}
	/*.................................................................................................................*/
	public void doCalcs(CommandRecord commandRec){
		needsRecalculation = false;
		int numTaxa = currentTaxa.getNumTaxa();
		results.resetSize(numTaxa);
		results.zeroArray();
		int sourceNumTrees = treeSourceTask.getNumberOfTrees(currentTaxa, commandRec);
		numTreesItem.setEnabled(!MesquiteInteger.isCombinable(sourceNumTrees));
		MesquiteTrunk.resetMenuItemEnabling();
		int numTrees;
		if (MesquiteInteger.isCombinable(sourceNumTrees)) {
			numTrees = sourceNumTrees;
		}
		else
			numTrees = getNumTrees(commandRec);
		MesquiteInteger minDiscord = new MesquiteInteger();
		MesquiteNumber b = new MesquiteNumber();
		double[][] originalDistances = new double[numTaxa][numTaxa];
		double[][] comparisonDistances = new double[numTaxa][numTaxa];
		double[][] heights = new double[MesquiteTree.standardNumNodeSpaces(numTaxa)][numTaxa];
		int numTreesUsed = numTrees;
		if (skipTrees >=1)
			numTreesUsed = (numTrees+ (skipTrees-1)) / skipTrees; 
		int numComparisons = numTreesUsed*(numTreesUsed-1)/2;
		ProgressIndicator progIndicator = null;
		if (numTrees>10){
			progIndicator = new ProgressIndicator(getProject(),getName(), "Calculating Terminal Instability by comparing to other trees", numComparisons, true);
			progIndicator.start();
		}
		int num = 0;
		for (int it = 0; it< numTrees  && (progIndicator==null || !progIndicator.isAborted()); it++){ //get tree for comparison
			if (skipTrees<=1 || it%skipTrees ==0){
				Tree targetTree = treeSourceTask.getTree(currentTaxa, it, commandRec);
				if (heights == null || heights.length < targetTree.getNumNodeSpaces())
					heights = new double[targetTree.getNumNodeSpaces()][numTaxa];
				//calculatePatristic(targetTree, currentTaxa, originalDistances); //for this tree calculate patristic distances (number of nodes separating terminals; no branch lengths)
				if (progIndicator !=null) {
					progIndicator.setCurrentValue(num);
					progIndicator.setText("Comparing tree " + (it+1) + " (of " + numTrees + ") to other trees");
				}
				commandRec.tick("Comparing tree " + (it+1) + " (of " + numTrees + ") to other trees");
				calculatePatristic(targetTree, numTaxa, heights, originalDistances); //for this tree calculate patristic distances (number of nodes separating terminals; no branch lengths)
				for (int it2 = 0; it2< it && (progIndicator==null || !progIndicator.isAborted()); it2++){ //get tree for comparison
					if (skipTrees<=1 || it2%skipTrees ==0){
						num++;
						
						Tree comparison = treeSourceTask.getTree(currentTaxa, it2, commandRec);
						if (heights == null || heights.length < comparison.getNumNodeSpaces())
							heights = new double[comparison.getNumNodeSpaces()][numTaxa];
						//calculatePatristic(comparison, currentTaxa, comparisonDistances);
						calculatePatristic(comparison, numTaxa, heights, comparisonDistances);
						for (int i = 0; i < numTaxa; i++){ //For each of the taxa, add to its score how unstable it is on this tree
							for (int j = 0; j < numTaxa; j++)  //for each of the OTHER taxa, calculate how much original and comparison distances differ
								if (j!=i) {
									b.setValue(strain(originalDistances[i][j], comparisonDistances[i][j]));
									results.addValue(i, b); //add difference in distances to running total
								}
					}
					}
				}
			}
		}
		String s = getVeryShortName() + " calculated";
		if (skipTrees >1)
			s += " among " + numTreesUsed + " out of a total of " + numTrees + " trees. ";
		else
			s += " among " + numTrees + " trees. ";
		if (progIndicator!=null && progIndicator.isAborted()) 
			resultsString.setValue(s + " INCOMPLETE CALCULATION (only " + num + " of " + numComparisons + " counted).");
		else {
			resultsString.setValue(s);
		}
		if (progIndicator!=null) 
			progIndicator.goAway();
	}
	private double strain(double orig, double comp){
		double div = Math.abs(orig)+ Math.abs(comp);
		if (div == 0)
			return 0;
		if (exponent >= 1.9999999 && exponent < 2.0000001)
			return Math.abs(orig - comp)/(div*div); //downweights distant taxa
		else
			return Math.abs(orig - comp)/Math.pow(div, exponent); //downweights distant taxa
	}
	/*-----------------------------------------*/
	private void calculatePatristic(Tree tree, int numTaxa, double[][] heights, double[][] distances){
		for (int i = 0; i < numTaxa; i++){
			for (int j = 0; j < numTaxa; j++)
				if (i!=j)
					distances[i][j] = -1;
				else
					distances[i][j] = 0;
		}
		for (int i = 0; i < tree.getNumNodeSpaces(); i++){
			for (int j = 0; j < numTaxa; j++)
				heights[i][j] = -1;
		}
		getHeights(tree, tree.getRoot(), heights);
		getDistances(tree, tree.getRoot(), numTaxa, heights, distances);
  	 }
	/*-----------------------------------------*/
	/** Finds heights to taxa for each node.  Beforehand initialize heights to -1*/
	void getHeights(Tree tree, int node, double[][] heights) { 
  	 	if (tree.nodeIsTerminal(node)){
  	 		int taxon = tree.taxonNumberOfNode(node);
  	 		heights[node][taxon]=0;
  	 	}
  	 	else {
  	 		boolean firstDaughter = true;
	  	 	for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter)) {
	  			getHeights(tree, daughter, heights);
  	 			for (int i=0; i<heights[daughter].length; i++)
  	 				if (heights[daughter][i] >= 0){ //taxon i is in daughter clade
  	 					if (heights[node][i]<0)
  	 						heights[node][i] = 0;
  	 					if (tree.getRoot()==node && !firstDaughter && !tree.nodeIsPolytomous(node))
  	 						heights[node][i] += heights[daughter][i];
  	 					else
  	 						heights[node][i] += heights[daughter][i] + 1;
		    	 			firstDaughter = false;
	 				}
   	 		}
   	 	}
	}
	/** Finds distances among taxa.  Beforehand initialize distances to -1*/
	void getDistances(Tree tree, int node,  int numTaxa, double[][] heights, double[][] distances) { 
  	 	if (tree.nodeIsInternal(node)){
	  	 	for (int daughter = tree.firstDaughterOfNode(node); tree.nodeExists(daughter); daughter = tree.nextSisterOfNode(daughter)) {
	  			getDistances(tree, daughter, numTaxa, heights, distances);
  	 		}
  	  	 	for (int i= 0; i< numTaxa; i++)
	  	 		for (int j= 0; j< i; j++)
	  	 			if (distances[i][j]<=0) { //not yet assigned
	  	 				if (heights[node][i]>= 0 && heights[node][j] >=0){ //first common ancestor found
	  	 					distances[i][j] = heights[node][i] + heights[node][j];
	  	 					distances[j][i] = distances[i][j];
	  	 				}
	  	 			}
  	 	}
	}
	/*.................................................................................................................*/
    	 public String getVeryShortName() {
		return "Instability Among Trees";
   	 }
	/*.................................................................................................................*/
    	 public String getName() {
		return "Taxon Instability Among Trees";
   	 }
	/*.................................................................................................................*/
  	 public String getVersion() {
		return null;
   	 }
	/*.................................................................................................................*/
  	 public boolean isPrerelease() {
		return false;
   	 }
	/*.................................................................................................................*/
 	/** returns an explanation of what the module does.*/
 	public String getExplanation() {
 		return "Calculates for each taxon the degree to which its implied unweighted patristic distances between that taxon and others differs among trees. "
 		+ " For each taxon i this sums over all tree pairs x and y and over all other taxa j:  | Dijx - Dijy |/(Dijx + Dijy)^^z where Dijq is the distance between taxa i and j on tree q. "
 		+ "Close relationships are emphasized if the exponent z is higher (default is 2).";
   	 }
   	 
}

