/* Mesquite source code.  Copyright 1997-2004 W. Maddison and D. Maddison. 
Version 1.01, January 2004.
Disclaimer:  The Mesquite source code is lengthy and we are few.  There are no doubt inefficiencies and goofs in this code. 
The commenting leaves much to be desired. Please approach this source code with the spirit of helping out.
Perhaps with your help we can be more than a few, and make Mesquite better.

Mesquite is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY.
Mesquite's web site is http://mesquiteproject.org

This source code and its compiled class files are free and modifiable under the terms of 
GNU Lesser General Public License.  (http://www.gnu.org/copyleft/lesser.html)
*/
package mesquite.trees.ManyTreesFromFile;
/*~~  */

import java.util.*;
import java.awt.*;
import mesquite.lib.*;
import mesquite.lib.duties.*;

/** Supplies trees from tree blocks in a file.  Reads trees only when needed; hence suitable for files with too many trees to be held in memory at once, but slower than StoredTrees.*/
public class ManyTreesFromFile extends TreeSource implements MesquiteListener {
	int currentTree=0;
	Taxa preferredTaxa =null;
	Taxa taxaInBlock = null;
	TreesManager manager;
	MesquiteFile file = null;
	TreeVector trees = null;
	Vector filePosVector;
	int arraySize = 1000;
	int highestTreeMarked = -1;
	int lastTreeRead = -1;
	String currentTreeName = null;
	int numTrees = MesquiteInteger.finite;
	int highestSuccessfulTree = -1;
	/*.................................................................................................................*/
	public boolean startJob(String arguments, Object condition, CommandRecord commandRec, boolean hiredByName) {
		filePosVector = new Vector();
		if (!commandRec.scripting()){
			if (!obtainFile(arguments, commandRec)) {
				return sorry(commandRec, "No tree file was obtained.");
			}
			if (!processFile(commandRec))
				return sorry(commandRec, "The file could not be processed.");
		}
		manager = (TreesManager)findElementManager(TreeVector.class);
		if (manager == null)
			return sorry(commandRec, "Tree manager not found.");
		addMenuItem("File for Trees Directly...", makeCommand("setFilePath",  this));
		return true;
  	 }
  	 
	/*.................................................................................................................*/
  	 private boolean obtainFile(String arguments, CommandRecord commandRec){
 		MesquiteModule fCoord = getFileCoordinator();
 		Object obj = fCoord.doCommand("getNEXUSFileForReading", arguments, new CommandRecord(true), CommandChecker.defaultChecker);
  	 	if (obj != null && obj instanceof MesquiteFile) {
  	 		if (file !=null) {
		  	 	file.closeReading();
		  	 	file.dispose();
		  	 	if (trees !=null) {
		  	 		trees.dispose();
		  	 		trees = null;
		  	 	}
				for (int j = 0; j<filePosVector.size(); j++){
					long[] filePosTrees = (long[])filePosVector.elementAt(j);
					
					for (int i=0; i<arraySize; i++)
						filePosTrees[i] = MesquiteLong.unassigned;
				}
				currentTree=0;
				highestTreeMarked = -1;
				lastTreeRead = -1;
				taxaInBlock = null;
				numTrees = MesquiteInteger.finite;
				highestSuccessfulTree = -1;
	  	 	}
  	 		file = (MesquiteFile)obj;
  	 	}
  	 	else
  	 		return false;
  	 	return true;
  	 }
  	 
  	 private boolean getFileFromPath(String pathName, CommandRecord commandRec){
 	 	if (pathName.indexOf(MesquiteFile.fileSeparator)<0) {
	    	 	file =MesquiteFile.open(getProject().getHomeDirectoryName(), pathName);
	 	}
   	 	else {
	 		String dirName = StringUtil.getAllButLastItem(pathName, MesquiteFile.fileSeparator, "/") + MesquiteFile.fileSeparator;
	 		String fileName = StringUtil.getLastItem(pathName, MesquiteFile.fileSeparator, "/");
    	 		file =MesquiteFile.open(dirName, fileName);
    	 	}
    	 	return (file != null);
  	 }
  	 private boolean processFile(CommandRecord commandRec){
  	 	if (!goToTreeBlock(file, commandRec))
  	 		return false;
		processTreeBlock( commandRec);
  	 	return true;
  	 }
	/*.................................................................................................................*/
  	 public void endJob(){
  	 	if (file !=null){
	  	 	file.closeReading();
	  	 	file.dispose();
  	 	}
  	 	super.endJob();
  	 }
	/*.................................................................................................................*/
  	 public Snapshot getSnapshot(MesquiteFile file) {
   	 	Snapshot temp = new Snapshot();
   	 	temp.addLine("setFilePath " + StringUtil.tokenize(this.file.getName()));  
  	 	return temp;
  	 }
	/*.................................................................................................................*/
  	 MesquiteInteger pos = new MesquiteInteger(0);
	/*.................................................................................................................*/
    	 public Object doCommand(String commandName, String arguments, CommandRecord commandRec, CommandChecker checker) {
    	 	if (checker.compare(this.getClass(), "Specifies the tree file to use", "[path to file]", commandName, "setFilePath")) {
    	 		String path = parser.getFirstToken(arguments);
    	 		if (obtainFile(arguments,  commandRec)){
    	 			if (processFile(commandRec)) {
    	 				if (!commandRec.scripting())
    	 					parametersChanged(null, commandRec);
    	 				return null;
    	 			}
    	 			else
    	 				discreetAlert(commandRec, "File could not be processed for " + getName() + " (path " + path + ")");
    	 		}
    	 		else 
    	 			discreetAlert(commandRec, "File was not obtained for " + getName() + " (path " + path + ")");
    	 		iQuit();
		}
    	 	else
    	 		return super.doCommand(commandName, arguments, commandRec, checker);
		return null;
   	 }
	/*.................................................................................................................*/
	/** finds the ith block of a given type and returns it raw.*/
	public boolean goToTreeBlock(MesquiteFile mNF, CommandRecord commandRec){
		ProgressIndicator progIndicator =  new ProgressIndicator(getProject(),"Processing File "+ mNF.getName() + " to find tree block", mNF.existingLength());
		progIndicator.start();
		boolean found = false;
		if (mNF.openReading()) {
			try {
				long blockStart = 0;
				logln("Processing File "+ mNF.getName() + " to find tree block");
				String token= mNF.firstToken(null);
				if (token!=null) {
					if (!token.equalsIgnoreCase("#NEXUS")) {
						alert("Not a valid NEXUS file (first token is \"" + token + "\"");
					}
					else {
						String name = null;
						blockStart = -1;
						blockStart = file.getFilePosition();
						while (!found && (name = mNF.goToNextBlockStart( ))!=null) {
							if ("TREES".equalsIgnoreCase(name)){
									found = true;
									
									file.goToFilePosition(blockStart); //go back to start of trees block
							}
							else 
								blockStart = file.getFilePosition()+1;
						}
						progIndicator.goAway();
						if (found)
							logln("Tree block found");
					}
				}
			}
			catch (MesquiteException e){
				return false;
			}
		}
		else
			return false;
		return found;
	}
	/*.................................................................................................................*/
  	 private void processTreeBlock( CommandRecord commandRec){

		Parser commandParser = new Parser();
		
		MesquiteInteger cPos = new MesquiteInteger(0);
		MesquiteString comment = new MesquiteString();
		String s;
		int treeNum=-1;
		if (getProject().getNumberTaxas()==1)
			 taxaInBlock = getProject().getTaxa(0); //as default)
		trees = new TreeVector(taxaInBlock);
		trees.setTaxa(taxaInBlock);
		trees.setName("Trees from \"" + file.getName() + "\"");
		boolean nameSet = false;
		MesquiteInteger status = new MesquiteInteger(0);
		while (!StringUtil.blank(s=file.getNextCommand(status, null))) {
			if (status.getValue() == 2) { //end of block reached
				return;
			}
			String punc = ",";
			String commandName = parser.getFirstToken(s);
			if (commandName.equalsIgnoreCase("BEGIN") || commandName.equalsIgnoreCase("END")  || commandName.equalsIgnoreCase("ENDBLOCK")) {
				//ignoring these
			}
			else if (commandName.equalsIgnoreCase("TRANSLATE")) {
				Vector table = null;
				if (taxaInBlock == null)
					table = new Vector();
				String label =  parser.getNextToken();
				while (punc !=null && !punc.equalsIgnoreCase(";")) {
					String taxonName = parser.getNextToken();
					if (taxaInBlock==null) 
						table.addElement(StringUtil.tokenize(taxonName) + " " + StringUtil.tokenize(label));
					else
						trees.setTranslationLabel(label, taxonName, false);
					punc =  parser.getNextToken(); 
					if (punc !=null && !punc.equals(";")) {
						label =  StringUtil.deTokenize(parser.getNextToken());
						if (";".equalsIgnoreCase(label))
							punc = label;  //to pop out of loop
					}
				}
				
				if (taxaInBlock==null) {
					taxaInBlock = manager.findTaxaMatchingTable(trees, getProject(), file, table);
					if (taxaInBlock!=null) {
						trees.setTaxa(taxaInBlock);
						trees.setTranslationTable(table);
					}
					else  {
						String st = "FAILED.";
						discreetAlert(commandRec, st);
					}
					
				}
				else {
					if (table!=null)
						trees.setTranslationTable(table);
				}
				trees.checkTranslationTable();
				recordFilePos(0, file.getFilePosition()); 
			}
			else if (commandName.equalsIgnoreCase("TITLE")) {
				trees.setName(parser.getTokenNumber(2));
				nameSet = true;
				recordFilePos(0, file.getFilePosition()); 
			}
			else if (commandName.equalsIgnoreCase("LINK")) {
				if ("taxa".equalsIgnoreCase(parser.getTokenNumber(2))) {
					String taxaTitle = parser.getTokenNumber(4);
					taxaInBlock = getProject().getTaxa(file, taxaTitle);
					if (taxaInBlock == null)
						taxaInBlock = getProject().getTaxa(taxaTitle);
					if (taxaInBlock == null) {
						if (getProject().getNumberTaxas(file)==1) //if translation table should search for match
							taxaInBlock = getProject().getTaxa(file, 0);
						else if (getProject().getNumberTaxas(file)==0 && getProject().getNumberTaxas()==1) //if translation table should search for match
							taxaInBlock = getProject().getTaxa(0);
						else
							discreetAlert(commandRec, "Taxa block not found for tree block");
					}
					trees.setTaxa(taxaInBlock);
					if (!nameSet)
						trees.setName("Trees block from " + file.getName());
				}
				recordFilePos(0, file.getFilePosition()); 
			}
			else if (commandName.equalsIgnoreCase("TREE") || commandName.equalsIgnoreCase("UTREE") || commandName.equalsIgnoreCase("RTREE"))  {
				return;
			}
		}
		recordFilePos(0, file.getFilePosition()); 
	}
	/*.................................................................................................................*/
	long getFilePos(int iTree){
		if (!posExists(iTree))
			return MesquiteLong.unassigned;
		int vec = iTree / arraySize;
		int loc = iTree % arraySize;
		long[] filePosTrees = (long[])filePosVector.elementAt(vec);
		return filePosTrees[loc];
	}
	boolean posExists(int iTree){
			return (iTree < arraySize*filePosVector.size()) ;
	}
	/*.................................................................................................................*/
	void recordFilePos(int iTree, long pos){
		if (pos<0 || pos > 100000000000L) {
			MesquiteMessage.warnProgrammer("illegal file pos in " + getName());
			return;
		}
		long[] filePosTrees;
		while (!posExists(iTree)) {
			filePosTrees = new long[arraySize];
			for (int i=0; i<arraySize; i++)
				filePosTrees[i] = MesquiteLong.unassigned;
			filePosVector.addElement(filePosTrees);
		}
		int vec = iTree / arraySize;
		int loc = iTree % arraySize;
		filePosTrees = (long[])filePosVector.elementAt(vec);
		filePosTrees[loc] = pos-1; //-1 to ensure not too far
		if (iTree > highestTreeMarked)
			highestTreeMarked = iTree;
	}
	/*.................................................................................................................*/
   	 public boolean isSubstantive(){
   	 	return true;
   	 }
	/*.................................................................................................................*/
   	 public boolean isPrerelease(){
   	 	return false;
   	 }
   	 
	/*.................................................................................................................*/
  	public void setPreferredTaxa(Taxa taxa) {
  		if (taxa !=null && taxa.isDoomed())
  			return;
  		if (preferredTaxa!=taxa) {
  			preferredTaxa = taxa;
		}
  	}
	/*.................................................................................................................*/
  	public void initialize(Taxa taxa, CommandRecord commandRec) {
  	//	checkTreeBlock(taxa, commandRec);
  	}
  	
	/*.................................................................................................................*/
  	String getTreeDescription(int currentTree, StringBuffer comment, CommandRecord commandRec){
		String command  = null;
		boolean fileDone = false;
		boolean isTreeCommand = false;
		
		MesquiteInteger status = new MesquiteInteger(0);
		if (currentTree==0){  //first tree
			commandRec.tick("\"Trees Directly from File\": Going to tree " + (currentTree+1));
			file.goToFilePosition(getFilePos(0));
			while (!isTreeCommand && !fileDone){
				command = file.getNextCommand(status, comment); 
				if (StringUtil.blank(command))
					fileDone = true;
				else
					isTreeCommand = ParseUtil.darkBeginsWithIgnoreCase(command, "TREE") || ParseUtil.darkBeginsWithIgnoreCase(command, "UTREE") || ParseUtil.darkBeginsWithIgnoreCase(command, "RTREE");
				if (isTreeCommand && currentTree>highestSuccessfulTree)
					highestSuccessfulTree = currentTree;
			}
			recordFilePos(1, file.getFilePosition()-1); 
		}
		else if (currentTree == lastTreeRead+1){ //last tree read was one less than requested; just continue to next without resetting file position
			commandRec.tick("Going to tree " + (currentTree+1));
			while (!isTreeCommand && !fileDone){
				command = file.getNextCommand( status, comment); //this is highest tree read
				if (StringUtil.blank(command))
					fileDone = true;
				else
					isTreeCommand = ParseUtil.darkBeginsWithIgnoreCase(command, "TREE") || ParseUtil.darkBeginsWithIgnoreCase(command, "UTREE") || ParseUtil.darkBeginsWithIgnoreCase(command, "RTREE");
				if (isTreeCommand && currentTree>highestSuccessfulTree)
					highestSuccessfulTree = currentTree;
			}
			if (currentTree % 100 == 0) {
					MesquiteMessage.println("Tree " + (currentTree) + " found by \"Trees Directly From File\"");
				}
			recordFilePos(currentTree+1, file.getFilePosition()); 
		}
		else if (currentTree>highestTreeMarked){  // a tree not yet read & not next in line
			if (highestTreeMarked>0)
				file.goToFilePosition(getFilePos(highestTreeMarked));
			int timeout = 0;
			for (int i = highestTreeMarked; i<=currentTree && !fileDone; i++) {
				commandRec.tick("Going to tree " + (i+1));
				isTreeCommand = false;
				timeout = 0;
				while (!isTreeCommand && !fileDone && timeout < 10000){
					command = file.getNextCommand(status, comment); 
					if (StringUtil.blank(command))
						fileDone = true;
					else
						isTreeCommand = ParseUtil.darkBeginsWithIgnoreCase(command, "TREE") || ParseUtil.darkBeginsWithIgnoreCase(command, "UTREE") || ParseUtil.darkBeginsWithIgnoreCase(command, "RTREE");
					if (!isTreeCommand)
						timeout++;
					if (isTreeCommand && i>highestSuccessfulTree)
						highestSuccessfulTree = i;
				}
				if (i % 100 == 0 && !fileDone) {
					MesquiteMessage.println("Tree " + (i) + " found by \"Trees Directly From File\"");
				}
				if (timeout >= 10000)
					alert("Error in reading tree file; 10000 commands found other than tree commands [last command (" + command + ")]");
				recordFilePos(i+1, file.getFilePosition()); 
			}
		}
		else { //a tree that is at or before the lastTreeRead, thus position should be known
			commandRec.tick("Going to tree " + (currentTree+1));
			file.goToFilePosition(getFilePos(currentTree));
			while (!isTreeCommand && !fileDone){
				command = file.getNextCommand( status, comment); //this is highest tree read
				if (StringUtil.blank(command))
					fileDone = true;
				else
					isTreeCommand = ParseUtil.darkBeginsWithIgnoreCase(command, "TREE") || ParseUtil.darkBeginsWithIgnoreCase(command, "UTREE") || ParseUtil.darkBeginsWithIgnoreCase(command, "RTREE");
				if (isTreeCommand && currentTree>highestSuccessfulTree)
					highestSuccessfulTree = currentTree;
			}
		}
		lastTreeRead = currentTree;
		if (fileDone)
			numTrees = highestSuccessfulTree+1;
		return command;
  	}
  	
	/*.................................................................................................................*/
   	public Tree getCurrentTree(Taxa taxa, CommandRecord commandRec) {
		String treeDescription;
		
		MesquiteInteger cPos = new MesquiteInteger(0);
		if (taxa != null && taxa != taxaInBlock) {
			if (taxaInBlock !=null)
				discreetAlert(commandRec, "Sorry, the trees found in file are for a different set of taxa than that requested");
			return null;
		}
		if (taxa == null)
			taxa = taxaInBlock;
		
		MesquiteInteger status = new MesquiteInteger(0);
		StringBuffer comment = new StringBuffer();
		String treeCommand = getTreeDescription(currentTree, comment, commandRec);
		if (treeCommand == null)
			return null;
		String commandName = parser.getFirstToken(treeCommand);
		if (commandName == null)
			return null;
		int whichType = 1;
		if (commandName.equalsIgnoreCase("UTREE")) 
			whichType =2;
		currentTreeName=parser.getNextToken();
		if (currentTreeName.equals("*"))
			currentTreeName=parser.getNextToken();
		parser.getNextToken(); //eat up "equals"
		treeDescription=treeCommand.substring(parser.getPosition(), treeCommand.length());
		MesquiteTree thisTree =new MesquiteTree(taxa);
		String commentString = comment.toString();
		
		if (commentString!=null && commentString.length()>1){
			if (commentString.charAt(0)=='!')
				thisTree.setAnnotation(commentString.substring(1, commentString.length()), false);
			else {
				int wpos = commentString.indexOf("&W");
				if (wpos <0)
					wpos = commentString.indexOf("&w");
				if (wpos>=0) {
					cPos.setValue(wpos+2);
					String num = ParseUtil.getToken(commentString, cPos);
					String slash = ParseUtil.getToken(commentString, cPos);
					String denom = ParseUtil.getToken(commentString, cPos);
					double w = 0;
					if (slash !=null && "/".equals(slash))
						w = 1.0*(MesquiteInteger.fromString(num))/(MesquiteInteger.fromString(denom));
					else
						w = MesquiteDouble.fromString(num);
					if (MesquiteDouble.isCombinable(w)) {
						MesquiteDouble d = new MesquiteDouble(w);
						d.setName("Weight");
						thisTree.attach(d);
					}
				}
			}
		}
		thisTree.setTreeVector(trees);
		trees.addElement(thisTree, false);
		thisTree.readTree(treeDescription);
		thisTree.setName(currentTreeName);
		if (whichType ==2) 
			thisTree.setRooted(false, false);
		trees.removeElement(thisTree, false);

   		return thisTree;
   	}
	/*.................................................................................................................*/
   	public Tree getFirstTree(Taxa taxa, CommandRecord commandRec) {
   		setPreferredTaxa(taxa);
   		currentTree=0;
  		return getCurrentTree(taxa, commandRec);
   	}
	/*.................................................................................................................*/
   	public Tree getTree(Taxa taxa, int itree, CommandRecord commandRec) {
   		setPreferredTaxa(taxa);
   		currentTree=itree;
  		return getCurrentTree(taxa, commandRec);
   	}
	/*.................................................................................................................*/
   	public Tree getNextTree(Taxa taxa, CommandRecord commandRec) {
   		setPreferredTaxa(taxa);
   		currentTree++;
  		return getCurrentTree(taxa, commandRec);
   	}
	/*.................................................................................................................*/
   	public int getNumberOfTrees(Taxa taxa, CommandRecord commandRec) {
   		setPreferredTaxa(taxa);
		return numTrees; 
   	}
   
	/*.................................................................................................................*/
   	public String getTreeNameString(Taxa taxa, int itree, CommandRecord commandRec) {
   		if (itree == currentTree)
   			return getCurrentTreeNameString(taxa, commandRec);
   		setPreferredTaxa(taxa);
   		return "Tree " + (itree +1);
   	}
	/*.................................................................................................................*/
   	public String getCurrentTreeNameString(Taxa taxa, CommandRecord commandRec) {
   		setPreferredTaxa(taxa);
   		if (currentTreeName == null)
   			return "Tree " + (currentTree +1);
   		return currentTreeName;
   	}
	/*.................................................................................................................*/
    	 public String getName() {
		return "Trees Directly From File";
   	 }
 	/*.................................................................................................................*/
    	 public String getNameForMenuItem() {
		return "Trees Directly From File...";
   	 }
  	 
	/*.................................................................................................................*/
	/** returns whether this module is requesting to appear as a primary choice */
   	public boolean requestPrimaryChoice(){
   		return true;  
   	}
	/*.................................................................................................................*/
  	 public String getExplanation() {
		return "Supplies trees directly from a file, without bringing the contained tree block entirely into memory.  This allows much larger blocks of trees to be used within constraints of memory, but will make some calculations slower.  This module does not know how many trees are in the file, and hence may attempt to read files beyond the number in the file.";
   	 }
	/*.................................................................................................................*/
   	public String getParameters() {
		if (file == null)
			return null;
		return "Trees obtained from file " + file.getName();
   	}
}

